#ifndef _AUTOREF_H_
#define _AUTOREF_H_ 1



#include "utilities.h"



/// This class is to be used for reference counting and automatic deallocation.
/// Such a handle object behaves like a normal reference plus assignment operator.
/// Each copy operation from one AutoRef to another one increases the internal
/// counter and each AutoRef destruction decreases the internal counter. If the
/// counter reaches zero the memory pointed by the internal pointer is deallocated.
/// 
/// The array template parameter is needed to differ between object (de)allocation
/// and array (de)allocation.
template<class T, bool array=false>
class AutoRef
{
public:

    inline AutoRef();
    inline AutoRef(T* data);
    inline AutoRef(const AutoRef& src);
    inline ~AutoRef();

    bool allocate(SizeT nmb=1);
    void clear();

    inline bool bad() const;
    inline ULONG count() const;

    inline operator bool() const;

    inline const T& operator*() const;
    inline T&       operator*();

    inline const T* operator->() const;
    inline T*       operator->();

    inline const T* ptr() const;
    inline T*       ptr();

    AutoRef& operator=(T* data);
    AutoRef& operator=(const AutoRef& src);

    inline bool operator==(const AutoRef& other) const;

    template<class trgT>
    inline operator AutoRef<trgT,array>() const;

    /// NEVER EVER use this constructor directly!!!!!
    ///
    /// I had to define this function to allow type up-casting. The problem was,
    /// that different template classes are handled as different types that have
    /// no access to private/protected members of objects with other template
    /// parameters.
    inline AutoRef(T* data, SizeT* count);


private :

    T*     mData;
    SizeT* mRefCount;
};



//-----------------------------------------------------------------------------
// Inlined Functions of AutoRef


/// Constructs an empty AutoRef object. Bad() returns true afterwards.
template<class T, bool array>
inline AutoRef<T,array>::AutoRef()
:
    mData(0), mRefCount(0)
{}


/// Constructs a new AutoRef object and initialises it with a pointer.
///
/// @param[in]  data
///   Pointer to the data the AutoRef should be the reference of.
template<class T, bool array>
inline AutoRef<T,array>::AutoRef(T* data)
{
    if (data)
    {
        mRefCount = gNew SizeT(1);
        if (!mRefCount)
        {
            ERRLOG_RETURN("AutoRef::AutoRef(): Could not allocate reference counter.");
        }
        mData = data;
    }
    else
    {
        mData = 0;
        mRefCount = 0;
    }
}


/// Copy constructor. Can be used for copying from one AutoRef to another
/// one. The reference counter is increased, if the source AutoRef referenced
/// some data.
///
/// @param[in]  src
///   Reference of the other instance to copy from.
template<class T, bool array>
inline AutoRef<T,array>::AutoRef(const AutoRef& other)
:
    mData(other.mData), mRefCount(other.mRefCount)
{
    if (mRefCount) { ++(*mRefCount); }
}


/// Destructs the handle object. Before construction the reference counter
/// is decreased. If it reaches zero the internal referenced data is
/// deallocated.
template<class T, bool array>
inline AutoRef<T,array>::~AutoRef()
{
    clear();
}


/// Allocates new object(s) T on heap.
///
/// @param[in]  nmb
///   The number of objects to create. If the template parameter ARRAY is false,
///   the parameter NMB will be ignored.
/// @return
///   true if successful (bad() will return false), false otherwise (bad() will
///   return true).
template<class T, bool array>
bool AutoRef<T,array>::allocate(SizeT nmb)
{
    clear();
    if (array)
    {
        if (!(mData = gNew T[nmb]))
        {
            ERRLOG_RETURN_VALUE(false, "AutoRef::allocate(): Could not allocate data array.");
        }
    }
    else
    {
        if (!(mData = gNew T))
        {
            ERRLOG_RETURN_VALUE(false, "AutoRef::allocate(): Could not allocate data.");
        }
    }
    if (!(mRefCount = gNew SizeT(1)))
    {
        if (array)
        {
            bDelete(mData);
        }
        else
        {
            gDelete(mData);
        }
        ERRLOG_RETURN_VALUE(false, "AutoRef::allocate(): Could not allocate reference counter.");
    }
    return true;
}


/// Decreases the reference counter and then releases this AutoRef from the
/// referenced object. If the reference counter reaches zero the referenced
/// data is deallocated. The internal data pointer is always NULL afterwards.
template<class T, bool array>
void AutoRef<T,array>::clear()
{
    if (mRefCount)
    {
        if (--(*mRefCount) == 0)
        {
            gDelete(mRefCount);
            if (array)
            {
                bDelete(mData);
            }
            else
            {
                gDelete(mData);
            }
        }
    }
}


/// Returns true if the reference counter is not set and the handle is free.
template<class T, bool array>
inline bool AutoRef<T,array>::bad() const
{
    return (!mRefCount);
}


/// Returns the internal reference count.
template<class T, bool array>
inline ULONG AutoRef<T,array>::count() const
{
    return (mRefCount) ? (*mRefCount) : 0;
}



/// Returns true if if the internal data pointer is not NULL.
template<class T, bool array>
inline AutoRef<T,array>::operator bool() const
{
    return !bad();
}


/// Constant de-reference operator.
template<class T, bool array>
inline const T& AutoRef<T,array>::operator*() const
{
    GeAssert(mData);
    return *mData;
}


/// Non-constant de-reference operator.
template<class T, bool array>
inline T& AutoRef<T,array>::operator*()
{
    GeAssert(mData);
    return *mData;
}


/// Constant "pointer-to-structure-member" operator.
template<class T, bool array>
inline const T* AutoRef<T,array>::operator->() const
{
    GeAssert(mData);
    return mData;
}


/// Non-constant "pointer-to-structure-memeber" operator.
template<class T, bool array>
inline T* AutoRef<T,array>::operator->()
{
    GeAssert(mData);
    return mData;
}


/// Returns the internally stored pointer as constant.
template<class T, bool array>
inline const T* AutoRef<T,array>::ptr() const
{
    return mData;
}


/// Returns the internally stored pointer.
template<class T, bool array>
inline T* AutoRef<T,array>::ptr()
{
    return mData;
}


/// Copy operator. Behaves like the copy constructor.
template<class T, bool array>
AutoRef<T,array>& AutoRef<T,array>::operator=(const AutoRef& src)
{
    if (src.mData != mData)
    {
        clear();
        mData     = src.mData;
        mRefCount = src.mRefCount;
        if (mRefCount) { ++(*mRefCount); }
    }
    return *this;
}


/// Initialises the AutoRef object with a new reference. If the operation fails,
/// the AutoRef is not referencing the data. I.e. the data has to be deallocated
/// by the caller.
///
/// @param[in]  data
///   The pointer the handle shoud be assigned to.
/// @return
///   A reference of this AutoRef. If DATA is NULL or the assignment operation
///   failed (because the a new reference counter could not be allocated), the
///   AutoRef will be empty afterwards and Bad() will return true.
template<class T, bool array>
AutoRef<T,array>& AutoRef<T,array>::operator=(T* data)
{
    if (data != mData)
    {
        clear();
        if (!data)  { return *this; }
        mData = data;
        if (!(mRefCount = gNew SizeT(1)))
        {
            ERRLOG("AutoRef::operator=(): Could not allocate reference counter.");
        }
    }
    return *this;
}


/// Returns true if OTHER stores the same data pointer as this handle.
template<class T, bool array>
inline bool AutoRef<T,array>::operator==(const AutoRef& other) const
{
    return (mData == other.mData);
}


/// Cast operator. Can be used for up-casting.
template<class T, bool array>
template<class trgT>
inline AutoRef<T,array>::operator AutoRef<trgT,array>() const
{
    return AutoRef<trgT,array>(mData, mRefCount);
}


/// NEVER EVER use this constructor directly!!!!!
///
/// I had to define this function to allow type up-casting. The problem was,
/// that different template classes are handled as different types that have
/// no access to private/protected members of objects with other template
/// parameters.
template<class T, bool array>
inline AutoRef<T,array>::AutoRef(T* data, SizeT* count)
:
    mData(data), mRefCount(count)
{
    if (mRefCount) { ++(*mRefCount); }
}



#endif  // #ifndef _AUTOREF_H_
