#ifndef _DYNARRAY1D_H_
#define _DYNARRAY1D_H_  1



#include <c4d.h>

#include "utilities.h"



/// This class implements a dynamic array container, that can grow if it's internal
/// capacity is exhausted. Shrinking is not done autmatically but can be done
/// manually, if needed.
///
/// WARNING: During a grow operation the array data will be copied into a new
///          (larger array). This copy is done bitwise, i.e. not using the copy
///          operator. As a consequence you should not keep any pointers to the
///          entries after a append(), push(), reserve() or adaptCapacity(). The
///          same applies to linking entries against each other using memory
///          addresses (i.e. pointers).
///
/// The template type has to support the following functions:
/// - default constructor ( T::T() )
/// - default destructor ( T::~T() )
/// - copy operator ( T::operator=(const T&) )
template <class T>
class DynArray1D
{
  public:

    DynArray1D(const SizeT size=0,
               const SizeT capacity=0);
    ~DynArray1D();

    DynArray1D(const DynArray1D& other);
    DynArray1D& operator=(const DynArray1D& other);

    bool init(const SizeT size=0,
              SizeT       capacity=0);
    bool reserve(SizeT capacity);
    bool adaptCapacity();

    void fill(const T& value);
    void clear();
    void erase();

    inline SizeT size() const;
    inline SizeT capacity() const;

    inline const T& operator[](const SizeT pos) const;
    inline T&       operator[](const SizeT pos);

    inline const T& front() const;
    inline T&       front();
    inline const T& back() const;
    inline T&       back();

    inline bool append();
    inline bool push(const T& value);
    inline T    pop();

    void remove(SizeT pos);

    inline const T* arrayAddress() const;
    inline T*       arrayAddress();


  private:

    SizeT mSize;
    SizeT mCapacity;
    T*    mData;

    bool increaseCapacity();
    bool newCapacity(const SizeT size);
};

#include "dynarray1d_impl.h"



#endif  // #ifndef _DYNARRAY1D_H_
