#ifndef _DYNARRAY1D_IMPL_H_
#define _DYNARRAY1D_IMPL_H_ 1



#include <cstring>

#include "utilities.h"



//-----------------------------------------------------------------------------
// Implementation of template class DynArray1D


/// Constructs a new instance and allocates a memory block of size CAPACITY.
/// The accessible size will be of SIZE.
/// NOTE: Allocated memory won't get cleared (except T initialises itself in
///       the default constructor).
///
/// @param[in]  size
///   The size of the accessible range after construction.
/// @param[in]  capacity
///   The actual size of the allocated memory. If CAPACITY < SIZE, CAPACITY
///   will be set to SIZE.
template <class T>
DynArray1D<T>::DynArray1D(const SizeT size,
                          const SizeT capacity)
:
    mSize(0), mCapacity(0), mData(0)
{
    init(size, capacity);
}


/// Destroys a DynArray1D instance and frees all allocated resources.
template <class T>
DynArray1D<T>::~DynArray1D()
{
    erase();
}


/// Constructs a new instance with a bitwise copy of the content of
/// another instance.
///
/// @param[in]  other
///   The other array to copy the content from.
template <class T>
DynArray1D<T>::DynArray1D(const DynArray1D& other)
:
    mSize(0), mCapacity(0), mData(0)
{
    *this = other;
}


/// Deallocates current content and creates a bitwise copy of the content
/// of the other instance.
///
/// @param[in]  other
///   The other array to copy the content from.
/// @return
///   A reference to this instance.
template <class T>
DynArray1D<T>& DynArray1D<T>::operator=(const DynArray1D& other)
{
    if (init(other.mSize, other.mCapacity) && other.mSize)
    {
        for (SizeT i=0; i<other.mSize; ++i) { mData[i] = other.mData[i]; }
    }
    return *this;
}


/// Deallocates current content and allocates a new dynamic array. The
/// behaviour is the same as described for the constructor.
///
/// @param[in]  size
///   The size of the accessible range after initialisation.
/// @param[in]  capacity
///   The actual size of the allocated memory. If CAPACITY < SIZE, CAPACITY
///   will be set to SIZE.
/// @return
///   false if allocation failed, otherwise true. If this operation fails
///   the array will have size and capacity of 0.
template <class T>
bool DynArray1D<T>::init(const SizeT size,
                         SizeT       capacity)
{
    // first free old content
    erase();

    // make sure that capacity >= size
    if (capacity < size) { capacity = size; }

    // if the array shoulf not be empty, allocate new memory block and update
    // members
    if (capacity > 0)
    {
        if ((mData = bNew T[capacity]) == 0)
        {
            ERRLOG("DynArray1D::Init(): could not allocate array of size " + LLongToString(capacity));
            return false;
        }
        mSize     = size;
        mCapacity = capacity;
    }
    return true;
}


/// Makes sure that the array has a capacity of at least SIZE. If the capacity
/// is smaller a larger memory block will be allocated. Otherwise nothing
/// changes.
///
/// @param[in]  size
///   The minimum capacity the array should have.
/// @return
///   false if the capacity is not large enough, but no larger memory block
///   could be allocated - true otherwise. If this function fails, the old
///   content is still available.
template <class T>
bool DynArray1D<T>::reserve(SizeT capacity)
{
    // if we have enough memory allocated, return happily
    if (mCapacity >= capacity) { return true; }

    // allocate new memory block
    T* newData = bNew T[capacity];
    if (!newData)
    {
        ERRLOG("DynArray1D::reserve(): could not allocate new arrayof size " + LLongToString(capacity));
        return false;
    }

    // copy data into new memory block and deallocate the old one
    for (SizeT i=0; i<mSize; ++i) { newData[i] = mData[i]; }
    bDelete(mData);
    mData = newData;

    // update capacity and return
    mCapacity = capacity;
    return true;
}


/// Reduces the capacity of the array to the used size.
///
/// @return
///   true if successful, false if not. If this function fails, the old content
///   is still available.
template <class T>
bool DynArray1D<T>::adaptCapacity()
{
    // if array empty, deallocate memory block
    if (mSize <= 0)
    {
        erase();
        // if the array is not empty, but its size doesn't match it's capacity:
    }
    else if (mSize != mCapacity)
    {
        // allocate new memory block of correct size
        T* newData = bNew T[mSize];
        if (!newData)
        {
            ERRLOG("DynArray1D::adaptCapacity(): could not allocate new array of size " + LLongToString(mSize));
            return false;
        }
        // copy content from old to new arrat
        for (SizeT i=0; i<mSize; ++i) { newData[i] = mData[i]; }
        // deallocate old memory and update members
        bDelete(mData);
        mData = newData;
        mCapacity = mSize;
    }
    return true;
}


/// Sets all entries of the array to a specified value.
///
/// @param[in]  value
///   The value the entries will be set to.
template <class T>
void DynArray1D<T>::fill(const T& value)
{
    for (SizeT c=0; c<mSize; ++c) { mData[c] = value; }
}


/// Sets the array size to 0, but doesn't change its capacity.
template <class T>
void DynArray1D<T>::clear()
{
    mSize = 0;
}


/// Deallocates the internal memory. Size and(!) capacity will be 0 afterwards.
template <class T>
void DynArray1D<T>::erase()
{
    mSize = 0;
    mCapacity = 0;
    bDelete(mData);
    mData = 0;
}


/// Returns the size of the array.
template <class T>
inline SizeT DynArray1D<T>::size() const  
{ 
    return mSize; 
}


/// Returns the capacity of the array.
template <class T>
inline SizeT DynArray1D<T>::capacity() const  
{ 
    return mCapacity; 
}


/// Constant random access operator.
///
/// @param[in]  pos
///   The position of the value to retrieve. Must be >=0 and <size()!
/// @return
///   A const reference to the value.
template <class T>
inline const T& DynArray1D<T>::operator[](const SizeT pos) const
{
    GeAssert((pos >= 0) && (pos < mSize));
    return mData[pos];
}


/// Non-constant random access operator.
///
/// @param[in]  pos
///   The position of the value to retrieve. Must be >=0 and <size()!
/// @return
///   A reference to the value.
template <class T>
inline T& DynArray1D<T>::operator[](const SizeT pos)
{
    GeAssert((pos >= 0) && (pos < mSize));
    return mData[pos];
}


/// Returns a constant reference to the first entry of the array.
/// (size() must be !=0)
template <class T>
inline const T& DynArray1D<T>::front() const
{
    GeAssert(mSize > 0);
    return mData[0];
}


/// Returns a non-constant reference to the first entry of the array.
/// (size() must be !=0)
template <class T>
inline T& DynArray1D<T>::front()
{
    GeAssert(mSize > 0);
    return mData[0];
}


/// Returns a constant reference to the last entry of the array.
/// (size() must be !=0)
template <class T>
inline const T& DynArray1D<T>::back() const
{
    GeAssert(mSize > 0);
    return mData[mSize-1];
}


/// Returns a non-constant reference to the last entry of the array.
/// (size() must be !=0)
template <class T>
inline T& DynArray1D<T>::back()
{
    GeAssert(mSize > 0);
    return mData[mSize-1];
}


/// Increases the size of the array by 1.
///
/// @return
///   true if successful, false otherwise.
template <class T>
inline bool DynArray1D<T>::append()
{
    if (mSize == mCapacity)
    {
        if (!increaseCapacity()) { return false; }
    }
    ++mSize;
    return true;
}


/// Copies a value to the end of the array. The array size will be increased
/// by 1.
///
/// @param[in]  value
///   The value to copy.
/// @return
///   true if successful, false if not.
template <class T>
inline bool DynArray1D<T>::push(const T& value)
{
    if (mSize == mCapacity)
    {
        if (!increaseCapacity()) { return false; }
    }
    mData[mSize] = value;
    ++mSize;
    return true;
}


/// Removes the last entry from the array and returns it's value.
template <class T>
inline T DynArray1D<T>::pop()
{
    T value = back();
    if (mSize > 0) { --mSize; }
    return value;
}


/// Returns a constant pointer to the internal data.
template <class T>
inline const T* DynArray1D<T>::arrayAddress() const  
{ 
    return mData; 
}


/// Returns a pointer to the internal data.
template <class T>
inline T* DynArray1D<T>::arrayAddress()  
{ 
    return mData; 
}


/// Removes the entry at position POS. Size() will be one less afterwards.
template <class T>
void DynArray1D<T>::remove(const SizeT pos)
{
    GeAssert(mSize > 0);
    for (SizeT c=pos+1; c<mSize; ++c) { mData[c-1] = mData[c]; }
    if (mSize) { --mSize; }
}



//-----------------------------------------------------------------------------
// Private Functions of DynArray1D


/// Helper function that increases the capacity. It's is called, when the
/// current capacity is exhausted.
///
/// @return
///   true if successful, false if not.
template <class T>
bool DynArray1D<T>::increaseCapacity()
{
    // calculate and check new capacity
    SizeT newCapacity = (mCapacity < 1) ? 16 : mCapacity<<1;
    if (newCapacity <= mCapacity)
    {
        ERRLOG("DynArray1D::increaseCapacity(): could not increase capacity variable - end of value range reached");
        return false;
    }

    // allocate new (larger) memory block
    T* newData = bNew T[newCapacity];
    if (!newData)
    {
        ERRLOG("DynArray1D<T>::increaseCapacity(): could not allocate new array of size " + LLongToString(newCapacity));
        return false;
    }

    // copy from to new memory
    for (SizeT i=0; i<mSize; ++i) { newData[i] = mData[i]; }

    // deallocate old memory, update members and return
    bDelete(mData);
    mData = newData;
    mCapacity = newCapacity;
    return true;
}



#endif  // #ifndef _DYNARRAY1D_IMPL_H_
