#include "filepath.h"
#include "utilities.h"



//-----------------------------------------------------------------------------
// Implementation of public member functions of class FilePath.


FilePath::FilePath()
: 
    mDrive(0),
    mIsAbsolute(false)
{}


FilePath::FilePath(const Filename& filename)
{
    *this = filename.GetString();
}


FilePath::FilePath(const String& pathStr)
{
    *this = pathStr;
}


FilePath::FilePath(const FilePath& other)
{
    *this = other;
}


FilePath& FilePath::operator=(const FilePath& other)
{
    mDrive      = other.mDrive;
    mIsAbsolute = other.mIsAbsolute;
    mTokens     = other.mTokens;
    return *this;
}


FilePath& FilePath::operator=(const Filename& filename)
{
    *this = filename.GetString();
    return *this;
}


FilePath& FilePath::operator=(const String& pathStr)
{
    // clear current file path content
    clear();

    // do nothing of string is empty
    if (!pathStr.Content()) { return *this; }

    // initialise loop vars
    LONG pathStrLen = pathStr.GetLength();
    LONG pos=0, tokenStart=-1;

#ifdef __PC
    // on PC get drive, if specified
    if ((pathStrLen > 1) && (pathStr[1] == ':'))
    {
        mDrive = (char)pathStr[0];
        if (pathStrLen == 2) { return *this; }
        pos = 2;
        tokenStart = 1;
    }
#endif

    // check if string starts with delimiter -> path is absolute
    if ((pathStr[pos] == '\\') || (pathStr[pos] == '/') || (pathStr[pos] == ':'))
    {
        mIsAbsolute = true;
        ++pos;
        ++tokenStart;
    }

    // split string into tokens and fix "." as well as ".."
    String token;
    for (;; ++pos)
    {
        if ((pos >= pathStrLen) ||
            (pathStr[pos] == '\\') || (pathStr[pos] == '/') || (pathStr[pos] == ':'))
        {
            if (pos > tokenStart+1) {
                token = pathStr.SubStr(tokenStart+1, pos-tokenStart-1);
                if (token != ".")
                {
                    if ((token == "..") && mTokens.size() && (mTokens.back() != ".."))
                    {
                        mTokens.removeBack();
                    }
                    else
                    {
                        mTokens.pushBack(token);
                    }
                }
            }
            if (pos >= pathStrLen) { break; }
            tokenStart = pos;
        }
    }

    return *this;
}


void FilePath::clear()
{
    mDrive      = 0;
    mIsAbsolute = false;
    mTokens.erase();
}


FilePath FilePath::getDirectoryPath() const
{
    FilePath result(*this);
    result.mTokens.removeBack();
    return result;
}


String FilePath::getString() const
{
    return getDirectoryString() + getFileString();
}


String FilePath::getFileString() const
{
    if (mTokens.size())
    {
        return mTokens.back();
    }
    else
    {
        return String();
    }
}


String FilePath::getDirectoryString() const
{
    String result;

#ifdef __PC
    // on PC add drive, if available
    if (mDrive)
    {
        result += String(1, mDrive);
        result += ":";
    }
#endif

    // if path is absolute add a leading delimiter
    if (mIsAbsolute) { result += gPathDelimiterStr; }

    // build the rest of the string by concatenating tokens
    for (TokensT::ConstIteratorT iter=mTokens.begin(); !iter.isBack(); ++iter)
    {
        result += *iter;
        result += gPathDelimiterStr;
    }

    return result;
}


Filename FilePath::getFilename() const
{
    return Filename(getString());
}


void FilePath::makeRelativeTo(const FilePath& basePath)
{
    // we can't determine the relation between the two paths if they are not both
    // absolute or both relative or have different drives
    if ((mIsAbsolute != basePath.mIsAbsolute) || (mDrive != basePath.mDrive))
    {
        return;
    }

    // relative paths don't have drives and well are not absolute
    mDrive = 0;
    mIsAbsolute = false;

    // now determine how much of the start of the token chains are equal
    TokensT::IteratorT myToken;
    TokensT::ConstIteratorT otherToken;
    for (myToken=mTokens.begin(), otherToken=basePath.mTokens.begin();
         myToken.isValid() && otherToken.isValid();
         ++otherToken)
    {
        if (myToken->LexCompare(*otherToken) == 0)
        {
            mTokens.remove(myToken);
        }
        else
        {
            break;
        }
    }

    // now add for every left entry of the other token list an ".." at the start
    for (; otherToken.isValid(); ++otherToken)
    {
        mTokens.pushFront("..");
    }
}


FilePath FilePath::operator+(const FilePath& other) const
{
    if (other.isAbsolute()) { return other; }

    FilePath result(*this);
    for (TokensT::ConstIteratorT iter=other.mTokens.begin();
         iter.isValid();
         ++iter)
    {
        if ((*iter == "..") && result.mTokens.size() && (result.mTokens.back() != ".."))
        {
            result.mTokens.removeBack();
        }
        else
        {
            result.mTokens.pushBack(*iter);
        }
    }
    return result;
}


FilePath& FilePath::operator+=(const String& token)
{
    mTokens.pushBack(token);
    return *this;
}


String FilePath::getSuffix() const
{
    LONG pos;
    if (mTokens.size() && mTokens.back().FindLast('.', &pos))
    {
        return mTokens.back().SubStr(pos, mTokens.back().GetLength() - pos);
    }
    return String();
}


void FilePath::setSuffix(const String& suffix)
{
    if (!mTokens.size()) { return; }

    LONG pos;
    if (mTokens.back().FindLast('.', &pos))
    {
        mTokens.back().Delete(pos, mTokens.back().GetLength() - pos);
    }
    mTokens.back() += suffix;
}


void FilePath::replaceIndex(const LONG newIndex)
{
    // if path is empty, return early
    if (!mTokens.size()) { return; }

    // look for the digit sequence either in front of the suffix or at the end
    // of the last token
    String &lastToken = mTokens.back();
    LONG   endPos;
    if (!lastToken .FindLast('.', &endPos))
    {
        endPos = lastToken .GetLength();
    }

    // 
    LONG residual = newIndex;
    for (LONG pos = endPos - 1; pos >= 0; --pos)
    {
        const String::PChar ch = lastToken[pos];
        if ((ch < '0') || (ch > '9')) { break; }

        if (newIndex >= 0)
        {
            lastToken[pos] = '0' + (residual % 10);
            residual /= 10;
        }
        else
        {
            lastToken[pos] = 'x';
        }
    }
}


void FilePath::appendToFile(const String& text)
{
}
