#include "c4d_symbols.h"
#include "oe_about.h"



//-----------------------------------------------------------------------------
// BitmapWidget


/// Helper class for displaying a bitmap in a dialog.
class BitmapWidget : public GeUserArea
{
public:

    enum Alignment
    {
        ALIGN_LOW,
        ALIGN_CENTER,
        ALIGN_HIGH
    };


    BitmapWidget(const Filename& imageFilename,
                 const Vector&   bgColor = Vector(0.5),
                 const Alignment horizontalAlignment = ALIGN_CENTER,
                 const Alignment verticalAlignment = ALIGN_CENTER);

    virtual ~BitmapWidget();


    // --- Implementation(s) for GeUserArea ---

    virtual Bool GetMinSize(LONG& width,
                            LONG& height);

    virtual void DrawMsg(LONG                 x,
                         LONG                 y,
                         LONG                 x2,
                         LONG                 y2,
                         const BaseContainer& msg);


private:

    const Vector    cBGColor;
    const Alignment cHorizontalAlignment;
    const Alignment cVerticalAlignment;

    BaseBitmap* mBitmap;
    LONG        mWidth;
    LONG        mHeight;


    BitmapWidget(const BitmapWidget& other);
    BitmapWidget& operator=(const BitmapWidget& other);
};


BitmapWidget::BitmapWidget(const Filename& imageFilename,
                           const Vector&   bgColor,
                           const Alignment horizontalAlignment,
                           const Alignment verticalAlignment)
:
    cBGColor(bgColor),
    cHorizontalAlignment(horizontalAlignment),
    cVerticalAlignment(verticalAlignment),
    mWidth(0),
    mHeight(0)
{
    mBitmap = BaseBitmap::Alloc();
    if (mBitmap)
    {
        if (mBitmap->Init(imageFilename) != IMAGERESULT_OK)
        {
            BaseBitmap::Free(mBitmap);
        }
        else
        {
            mWidth  = mBitmap->GetBw();
            mHeight = mBitmap->GetBh();
        }
    }
}


BitmapWidget::~BitmapWidget()
{
    if (mBitmap)
    {
        BaseBitmap::Free(mBitmap);
    }
}


Bool BitmapWidget::GetMinSize(LONG& width,
                              LONG& height)
{
    width  = mWidth;
    height = mHeight;
    return TRUE;
}


void BitmapWidget::DrawMsg(LONG                 x,
                           LONG                 y,
                           LONG                 x2,
                           LONG                 y2,
                           const BaseContainer& msg)
{
    OffScreenOn();

    // render background
    LONG areaWidth  = GetWidth();
    LONG areaHeight = GetHeight();
    DrawSetPen(cBGColor);
    DrawRectangle(0, 0, areaWidth, areaHeight);

    // determine position and section of image to render
    LONG srcPosX, srcPosY;
    LONG srcWidth  = LMin(mWidth,  areaWidth);
    LONG srcHeight = LMin(mHeight, areaHeight);
    LONG dstPosX, dstPosY;
    LONG dstWidth  = srcWidth;
    LONG dstHeight = srcHeight;
    switch (cHorizontalAlignment)
    {
        case ALIGN_CENTER:
            srcPosX = (mWidth - srcWidth) / 2;
            dstPosX = (areaWidth - dstWidth) / 2;
            break;
        case ALIGN_HIGH:
            srcPosX = mWidth - srcWidth;
            dstPosX = areaWidth - dstWidth;
            break;
        default:
            srcPosX = 0;
            dstPosX = 0;
            break;
    }
    switch (cVerticalAlignment)
    {
        case ALIGN_CENTER:
            srcPosY = (mHeight - srcHeight) / 2;
            dstPosY = (areaHeight - dstHeight) / 2;
            break;
        case ALIGN_HIGH:
            srcPosY = mHeight - srcHeight;
            dstPosY = areaHeight - dstHeight;
            break;
        default:
            srcPosY = 0;
            dstPosY = 0;
            break;
    }

    // render bitmap
    if (mBitmap)
    {
        DrawBitmap(mBitmap,
                   dstPosX, dstPosY, dstWidth, dstHeight,
                   srcPosX, srcPosY, srcWidth, srcHeight,
                   BMP_NORMAL | BMP_ALLOWALPHA);
    }
}



//-----------------------------------------------------------------------------
// OEAboutDlg


/// A little about dialog for the exporter.
class OEAboutDlg : public GeModalDialog
{
public:

    enum
    {
        IDG_MAIN,
        IDI_DATA,
        IDS_ABOUT_TEXT,
        IDB_OK
    };


    OEAboutDlg();

    virtual ~OEAboutDlg();


    // --- Implementation(s) for GeModalDialog ---

    virtual Bool CreateLayout();

    virtual Bool Command(LONG                 id,
                         const BaseContainer& msg);


private:

    BitmapWidget *mBanner;
};


OEAboutDlg::OEAboutDlg()
{
    Filename bannerFilename = GeGetPluginPath();
    bannerFilename += "res";
    bannerFilename += "oe_dlg_title.tif";
    mBanner = gNew BitmapWidget(bannerFilename, Vector(0.9));
}


OEAboutDlg::~OEAboutDlg()
{
    gDelete(mBanner);
}


Bool OEAboutDlg::CreateLayout()
{
    GroupBegin(IDG_MAIN, 0, 1, 4, String(), 0);
    {
        GroupSpace(8, 8);
        GroupBorderSpace(8, 8, 8, 8);
        C4DGadget* bannerArea = AddUserArea(IDI_DATA, BFH_SCALEFIT|BFV_SCALEFIT, 0, 80);
        if (bannerArea)
        {
            AttachUserArea(*mBanner, bannerArea);
        }
        AddStaticText(IDS_ABOUT_TEXT, BFH_SCALE|BFH_CENTER, 0, 0,
            "OctaneRender exporter for CINEMA 4D", BORDER_NONE);
        AddStaticText(IDS_ABOUT_TEXT, BFH_SCALE|BFH_CENTER, 0, 0,
            "Version 0.13", BORDER_NONE);
        AddButton(IDB_OK, BFH_CENTER, 60, 0, "OK");
    } 
    GroupEnd();
    return TRUE;
}


Bool OEAboutDlg::Command(LONG                 id,
                         const BaseContainer& msg)
{
    if (id == IDB_OK)
    {
        Close(TRUE);
    }
    return true;
}


//-----------------------------------------------------------------------------
// Implementation of OEAbout


bool OEAbout::registerPlugin()
{
    return RegisterCommandPlugin(PID_OE_ABOUT,
                                 GeLoadString(IDS_OE_ABOUT),
                                 0,
                                 NULL,
                                 GeLoadString(IDS_OE_ABOUT_DESCR),
                                 this);
}


Bool OEAbout::Execute(BaseDocument* document)
{
    OEAboutDlg dialog;
    dialog.Open();
    return TRUE;
}
