#include <cstdio>
#include <ocamera.h>

#include "c4d_symbols.h"
#include "oe_export_dlg.h"
#include "oe_obj_converter.h"
#include "oe_preferences.h"
#include "oe_converter.h"



//-----------------------------------------------------------------------------
// Implementation of public member functions of class OEConverter.


OEConverter::OEConverter()
{
    clear();
}


void OEConverter::clear()
{
    mDocument           = 0;
    mProjectPath        = "";
    mProjectName        = "";
    mRenderTargets.clear();
    mRenderTargetIx     = 0;
    mExportScaleId      = ID_OE_DLG_NATIVE_UNIT_CM;
    mCollectAllImages   = false;
    mUseDocumentScale   = true;
    mSamples            = 0;
    mOutputFilename     = "";
    mImageFormatId      = 0;
    mExportResolution   = false;
    mWidth              = 0;
    mHeight             = 0;
    mResolutionScaleId  = ID_OE_DLG_SCALE_100;
    mExportCamera       = false;
    mCameraIx           = 0;
    mExportAperture     = false;
    mAperture           = 0.0;
    mExportFocalDepth   = false;
    mFocalDepth         = 0.0;
    mExportMotionBlur   = false;
    mMotionBlurToNext   = false;
    mExportSunDirection = false;
    mLightIx            = 0;

    mWasPreExport       = false;
    mTempRenames.erase();
    mObjFilePath.clear();
    mOcsFilePath.clear();
    mSunDir             = Vector(0.0);

    mArgCount           = 0;
    mOutputFilePath.clear();
    mOctaneFilePath.clear();
    mOctaneHandle       = gEmptyHandle;
}


CameraObject* OEConverter::getCamera()
{
    CameraObject* cam;

    // if a camera has been selected in the dialog, return that one
    if (mCameraIx != 0)
    {
        cam = (CameraObject*)findObject(*mDocument, mCameraIx-1, Ocamera);
        if (cam) { return cam; }
        ERRLOG("OEConverter::getCamera(): Could not find camera of index " + LongToString(mCameraIx));
    }

    // if no camera has been selected, use camera of stage object, if it exists
    BaseObject *object = mDocument->GetHighest(Ostage, false);
    if (object && (object->GetType() == Ostage) && object->GetDeformMode())
    {
        // ignore the stage object if it belongs to a layer that should not be
        // rendered
        const LayerData* layerData = object->GetLayerData(mDocument);
        if (!layerData || layerData->render)
        {
            cam = (CameraObject*)getParameterLink(*object, STAGEOBJECT_CLINK, Ocamera);
            if (cam) { return cam; }
        }
    }

    // if no camera was selected and no stage camera was available, return the
    // current render camera
    BaseDraw* view = mDocument->GetRenderBaseDraw();
    if (!view) { ERRLOG_RETURN_VALUE(0, "OEConverter::getCamera(): Could not obtain render base draw"); }
    object = view->GetSceneCamera(mDocument);
    if (!object) { object = view->GetEditorCamera(); }
    if (!object) { ERRLOG_RETURN_VALUE(0, "OEConverter::getCamera(): Could not obtain render camera"); }
    if (!object->IsInstanceOf(Ocamera)) { ERRLOG_RETURN_VALUE(0, "OEConverter::getCamera(): obtained camera object is no camera"); }
    return (CameraObject*)object;
}


Real OEConverter::getFocalDepth(
    CameraObject& cam) const
{
#if _C4D_VERSION>=130
    // also check if there is a focus object and calculate depth from that one
    BaseObject* focusObject = (BaseObject*)getParameterLink(cam, CAMERAOBJECT_TARGETOBJECT, Obase);
    const Matrix camMat = cam.GetMgn();
    return focusObject ? (focusObject->GetAbsPos()-camMat.off) * camMat.v3
                       : getParameterReal(cam, CAMERAOBJECT_TARGETDISTANCE);
#else
    return getParameterReal(cam, CAMERAOBJECT_TARGETDISTANCE);
#endif
}


Real OEConverter::getExportScale() const
{
#if _C4D_VERSION >= 120
    if (mUseDocumentScale)
    {
        const UnitScaleData* documentScale = (const UnitScaleData*)
            mDocument->GetDataInstance()->GetCustomDataType(DOCUMENT_DOCUNIT,
                                                            CUSTOMDATATYPE_UNITSCALE);
        if (documentScale)
        {
            AutoAlloc<UnitScaleData> octaneScale;
            if (octaneScale && octaneScale->SetUnitScale(1, DOCUMENT_UNIT_M))
            {
                return CalculateTranslationScale(documentScale, octaneScale);
            }
        }
    }
#endif
    return OEExportDlg::getExportScale(mExportScaleId);
}


bool OEConverter::exportToObj(
    const bool isPreExport)
{
    // check if a document has been set
    if (!mDocument)
    {
        ERRLOG_RETURN_VALUE(false, "OEConverter::exportToObj(): No document set -> no export.");
    }

    // if enabled, determine direction of sun light (i.e. Z axis of sun light)
    if (mExportSunDirection)
    {
        BaseObject* sunLight = findObject(*mDocument, mLightIx, Olight);
        if (sunLight)
        {
            Matrix sunMat = sunLight->GetMgn();
            mSunDir = -sunMat.v3;
        }
    }

    // convert scene and store it as OBJ/MTL files
    OEOBJConverter objConverter;
    if (!objConverter.convertScene(*mDocument,
                                   getExportScale(),
                                   true,
                                   mProjectPath,
                                   mProjectName,
                                   mCollectAllImages,
                                   isPreExport,
                                   mObjFilePath,
                                   mTempRenames))
    {
        GeOutString(GeLoadString(IDS_ERROR_OBJ_EXPORT_FAILED), GEMB_OK);
        return false;
    }
    mWasPreExport = isPreExport;

    // build .ocs file path and return
    mOcsFilePath = mProjectPath + mProjectName;
    mOcsFilePath.setSuffix(".ocs");
    return true;
}


bool OEConverter::launchOctane(
    const bool   isAnim,
    const LONG   frame,
    const Matrix *mbCamMat1,
    const Matrix *mbCamMat2)
{
    // return early, if Octane is still running
    if (octaneIsActive())
    {
        ERRLOG_RETURN_VALUE(false, "OEConverter::launchOctane(): Octane is still running. Can't launch second instance.");
    }

    // get Octane path
    const Filename octanePath = gPreferences->octanePath();
    if (!octanePath.Content())
    {
        GeOutString(GeLoadString(IDS_ERROR_NO_OCTANE_BINARY), GEMB_OK);
        return false;
    }
    Filename tempOctaneFilename(octanePath);
#ifdef __MAC
    // on MacOS, Octane is stored in a bundle 
    if (GeFExist(tempOctaneFilename, true))
    {
        tempOctaneFilename += "Contents";
        tempOctaneFilename += "MacOS";
        tempOctaneFilename += "octane";
    }
#endif

    // check if Octane path is valid
    if (!GeFExist(tempOctaneFilename, false))
    {
        GeOutString(GeLoadString(IDS_ERROR_PATH_DOESNT_EXIST, tempOctaneFilename.GetString()), GEMB_OK);
        return false;
    }

    // rename OBJ and MTL files, which have been stored under a temporary name
    if (mWasPreExport && !renameTemps())
    {
        GeOutString(GeLoadString(IDS_ERROR_OBJ_EXPORT_FAILED), GEMB_OK);
        ERRLOG_RETURN_VALUE(false, "OEConverter::launchOctane(): Could not rename pre-exported files.");
    }

    // reset argument count
    mArgCount = 0;

    // if this is an animation rendering, add animation specific arguments
    if (isAnim)
    {
        mArgs[mArgCount++] = "-e";
        mArgs[mArgCount++] = "-q";

        char buffer[16];
        sprintf(buffer, "%08d", (int)frame);
        buffer[ARRAY_LENGTH(buffer)-1] = '\0';
        mOutputFilePath = mOutputFilename.GetDirectory();
        switch (mImageFormatId)
        {
            default:
            case ID_OE_DLG_IMAGE_FORMAT_PNG8:
                mArgs[mArgCount++] = "-o";
                mOutputFilePath += mOutputFilename.GetFileString() + buffer + ".png";
                break;
            case ID_OE_DLG_IMAGE_FORMAT_PNG16:
                mArgs[mArgCount++] = "--output-png16";
                mOutputFilePath += mOutputFilename.GetFileString() + buffer + ".png";
                break;
            case ID_OE_DLG_IMAGE_FORMAT_EXR:
                mArgs[mArgCount++] = "--output-exr";
                mOutputFilePath += mOutputFilename.GetFileString() + buffer + ".exr";
                break;
            case ID_OE_DLG_IMAGE_FORMAT_EXR_TM:
                mArgs[mArgCount++] = "--output-exr-tm";
                mOutputFilePath += mOutputFilename.GetFileString() + buffer + ".exr";
                break;
        }
        mArgs[mArgCount++] = "\"" + mOutputFilePath.getString() + "\"";
        mArgs[mArgCount++] = "-s";
        mArgs[mArgCount++] = LongToString(mSamples>4 ? mSamples : 4);
        // remove output file
        GeFKill(mOutputFilePath.getFilename(), 0);
    }
    else
    {
        mOutputFilePath.clear();
    }

    // if the OCS file exists already, relink new OBJ file with the existing OCS
    bool newOCSFile;
    if (GeFExist(mOcsFilePath.getFilename()))
    {
        newOCSFile = false;
        // if a render traget was specified, that is not the preview render target, let Octane
        // render that target
        if ((mRenderTargetIx > 0) && ((size_t)mRenderTargetIx <= mRenderTargets.size()))
        {
            mArgs[mArgCount++] = "-t";
            mArgs[mArgCount++] = "\"" + mRenderTargets[mRenderTargetIx-1] + "\"";
        }
        // if the preview render target was specified, let Octane render the mesh node and relink
        // it with the exported OBJ
        else
        {
            mArgs[mArgCount++] = "-m";
            mArgs[mArgCount++] = "\"" + mProjectName + "\"";
            mArgs[mArgCount++] = "-r";
            mArgs[mArgCount++] = "\"" + mObjFilePath.getString() + "\"";
        }
    }
    else
    {
        newOCSFile = true;
        mArgs[mArgCount++] = "-n";
        mArgs[mArgCount++] = "\"" + mOcsFilePath.getString() + "\"";
        mArgs[mArgCount++] = "-m";
        mArgs[mArgCount++] = "\"" + mProjectName + "\"";
        mArgs[mArgCount++] = "-l";
        mArgs[mArgCount++] = "\"" + mObjFilePath.getString() + "\"";
    }

    // add film resolution if enabled
    if (mExportResolution && (mWidth > 4) && (mHeight > 4))
    {
        SReal resolutionScale = (SReal)OEExportDlg::getResolutionScale(mResolutionScaleId);
        LONG width  = (LONG)(((SReal)mWidth  * resolutionScale) + 0.5f);
        LONG height = (LONG)(((SReal)mHeight * resolutionScale) + 0.5f);
        mArgs[mArgCount++] = "--film-width";
        mArgs[mArgCount++] = LongToString(width);
        mArgs[mArgCount++] = "--film-height";
        mArgs[mArgCount++] = LongToString(height);
    }

    // find the camera object
    CameraObject* cam = getCamera();

    // add camera parameters
    if (mExportCamera && cam)
    {
        // obtain camera parameters
        const LONG perspective = getParameterLong(*cam, CAMERA_PROJECTION, Pperspective);
        const Real filmWidth   = getParameterReal(*cam, CAMERAOBJECT_APERTURE, 36);
        const Real focalLength = getParameterReal(*cam, CAMERA_FOCUS, 36);
        const Real zoom        = getParameterReal(*cam, CAMERA_ZOOM, 1);
        const Real filmOffsetX = getParameterReal(*cam, CAMERAOBJECT_FILM_OFFSET_X);
        const Real filmOffsetY = getParameterReal(*cam, CAMERAOBJECT_FILM_OFFSET_Y);
        // calculate/load Octane camera parameters
        const Real scale = getExportScale();
        const bool hasMBData = (mbCamMat1 && mbCamMat2);
        const Real focalDepth = (mFocalDepth == 0.0) ? getFocalDepth(*cam) * scale
                                                     : mFocalDepth;
        // get normalized camera matrix
        const Matrix camMat = hasMBData ? *mbCamMat1 : cam->GetMgn();
        // convert matrix to pos/trg/up
        Vector camPos, camTrg, camUp;
        calcOctaneCamera(camMat, focalDepth, scale, isAnim, camPos, camTrg, camUp);
        // add either FOV or scale parameter depending on whether cam is perspective or parallel
        if (perspective == Pparallel)
        {
            const Real camScale = 1024 / zoom * scale;
            mArgs[mArgCount++] = "--cam-scale";
            mArgs[mArgCount++] = toString(camScale);
        }
        else
        {
            const Real fov = 2.0 * atan(0.5*filmWidth / focalLength) / pi * 180.0;
            mArgs[mArgCount++] = "--cam-fov";
            mArgs[mArgCount++] = toString(fov);
        }
        // add parameters
        mArgs[mArgCount++] = "--cam-pos-x";
        mArgs[mArgCount++] = toString(camPos.x);
        mArgs[mArgCount++] = "--cam-pos-y";
        mArgs[mArgCount++] = toString(camPos.y);
        mArgs[mArgCount++] = "--cam-pos-z";
        mArgs[mArgCount++] = toString(-camPos.z);
        mArgs[mArgCount++] = "--cam-target-x";
        mArgs[mArgCount++] = toString(camTrg.x);
        mArgs[mArgCount++] = "--cam-target-y";
        mArgs[mArgCount++] = toString(camTrg.y);
        mArgs[mArgCount++] = "--cam-target-z";
        mArgs[mArgCount++] = toString(-camTrg.z);
        mArgs[mArgCount++] = "--cam-up-x";
        mArgs[mArgCount++] = toString(camUp.x);
        mArgs[mArgCount++] = "--cam-up-y";
        mArgs[mArgCount++] = toString(camUp.y);
        mArgs[mArgCount++] = "--cam-up-z";
        mArgs[mArgCount++] = toString(-camUp.z);
        if (mExportAperture)
        {
            mArgs[mArgCount++] = "--cam-aperture";
            mArgs[mArgCount++] = toString(mAperture);
        }
        if (mExportFocalDepth)
        {
            mArgs[mArgCount++] = "--cam-focaldepth";
            mArgs[mArgCount++] = toString(focalDepth);
        }
        mArgs[mArgCount++] = "--cam-lensshift-right";
        mArgs[mArgCount++] = toString(filmOffsetX);
        mArgs[mArgCount++] = "--cam-lensshift-up";
        mArgs[mArgCount++] = toString(-filmOffsetY);
        // export motion blur data, i.e. the camera one frame ahead or behind
        if (mExportMotionBlur && hasMBData)
        {
            // determine new position, target and up-vector
            calcOctaneCamera(*mbCamMat2, focalDepth, scale, isAnim, camPos, camTrg, camUp);
            // add parameters
            mArgs[mArgCount++] = "--cam-motion-pos-x";
            mArgs[mArgCount++] = toString(camPos.x);
            mArgs[mArgCount++] = "--cam-motion-pos-y";
            mArgs[mArgCount++] = toString(camPos.y);
            mArgs[mArgCount++] = "--cam-motion-pos-z";
            mArgs[mArgCount++] = toString(-camPos.z);
            mArgs[mArgCount++] = "--cam-motion-target-x";
            mArgs[mArgCount++] = toString(camTrg.x);
            mArgs[mArgCount++] = "--cam-motion-target-y";
            mArgs[mArgCount++] = toString(camTrg.y);
            mArgs[mArgCount++] = "--cam-motion-target-z";
            mArgs[mArgCount++] = toString(-camTrg.z);
            mArgs[mArgCount++] = "--cam-motion-up-x";
            mArgs[mArgCount++] = toString(camUp.x);
            mArgs[mArgCount++] = "--cam-motion-up-y";
            mArgs[mArgCount++] = toString(camUp.y);
            mArgs[mArgCount++] = "--cam-motion-up-z";
            mArgs[mArgCount++] = toString(-camUp.z);
        }
    }

    // add daylight settings
    if (mExportSunDirection && (mSunDir != Vector(0.0)))
    {
        mArgs[mArgCount++] = "--daylight-sundir-x";
        mArgs[mArgCount++] = toString(mSunDir.x);
        mArgs[mArgCount++] = "--daylight-sundir-y";
        mArgs[mArgCount++] = toString(mSunDir.y);
        mArgs[mArgCount++] = "--daylight-sundir-z";
        mArgs[mArgCount++] = toString(-mSunDir.z);
    }

    // add GPU settings
    if (gPreferences->hasSpecifiedGpus())
    {
        String gpuStr = gPreferences->specifiedGpuStr();
        String gpuID;
        for (LONG i=0; i<gpuStr.GetLength(); ++i)
        {
            String::PChar ch = gpuStr[i];
            if ((ch >= '0') && (ch <= '9'))
            {
                gpuID += String(1, ch);
            }
            else if (gpuID.Content())
            {
                mArgs[mArgCount++] = "-g";
                mArgs[mArgCount++] = gpuID;
                gpuID = "";
            }
        }
        if (gpuID.Content())
        {
            mArgs[mArgCount++] = "-g";
            mArgs[mArgCount++] = gpuID;
        }
    }

    // add OCS file at the end, if this is a relink
    if (!newOCSFile)
    {
        mArgs[mArgCount++] = "\"" + mOcsFilePath.getString() + "\"";
    }

    // launch Octane
    String workingDir(mOcsFilePath.getDirectoryString());
    mOctaneFilePath = tempOctaneFilename;
    mIsAnim = isAnim;
    mOctaneHandle = runProgram(mOctaneFilePath,
                               mArgs,
                               mArgCount,
                               &workingDir,
                               isAnim);
    if (!isValid(mOctaneHandle))
    {
        GeOutString(GeLoadString(IDS_ERROR_COULD_NOT_LAUNCH_OCTANE), GEMB_OK);
        return false;
    }

    return true;
}


bool OEConverter::relaunchOctane()
{
    String workingDir(mOcsFilePath.getDirectoryString());
    mOctaneHandle = runProgram(mOctaneFilePath,
                               mArgs,
                               mArgCount,
                               &workingDir,
                               mIsAnim);
    if (!isValid(mOctaneHandle))
    {
        GeOutString(GeLoadString(IDS_ERROR_COULD_NOT_LAUNCH_OCTANE), GEMB_OK);
        return false;
    }
    return true;
}


bool OEConverter::stopOctane()
{
    if (!octaneIsActive()) { return false; }
    stopProgram(mOctaneHandle);
    mOctaneHandle = gEmptyHandle;
    return true;
}


bool OEConverter::octaneIsActive()
{
    return (isValid(mOctaneHandle) && programIsRunning(mOctaneHandle));
}


bool OEConverter::imageRendered()
{
    return GeFExist(mOutputFilePath.getFilename());
}



//-----------------------------------------------------------------------------
// Implementation of private member functions of class OEConverter.


bool OEConverter::renameTemps()
{
    for (int tries = 0; mTempRenames.size() && (tries < 10); ++tries)
    {
        if (tries) { GeSleep(500); }

        for (OEOBJConverter::RenameListT::IteratorT iter = mTempRenames.begin();
             iter.isValid();)
        {
            const Filename dstFilename = iter->mToPath.getFilename();
            if (!GeFExist(dstFilename, FALSE) || GeFKill(dstFilename))
            {
                const Filename srcFilename = iter->mFromPath.getFilename();
                if (GeFRename(srcFilename, dstFilename))
                {
                    // file was successfully renamed -> remove from list
                    mTempRenames.remove(iter);
                    continue;
                }
            }
            // if we got here we couldn't rename the current entry, continue
            // with next one
            ++iter;
        }
    }
    return (mTempRenames.size() == 0);
}


void OEConverter::calcOctaneCamera(
    const Matrix &matrix,
    const Real   focalDepth,
    const Real   scale,
    const bool   anim,
    Vector       &pos,
    Vector       &target,
    Vector       &up)
{
    // get position
    pos = matrix.off * scale;

    // the target is the point along the Z axis that has distance focalDepth
    const Real targetDist = (focalDepth > 0.0001) ? focalDepth : 1;
    target = pos + matrix.v3 * targetDist;

    // calculate up-vector
    Real projY = matrix.v2 * Vector(0.0, 1.0, 0.0);
    // special case: if we are animating or the Y axis lies in the XZ plane,
    // we don't try to "upright" the Y axis, but use it as it is
    if (anim || (fabs(projY) < 0.01))
    {
        up = matrix.v2;
    }
    // if the Y axis points upwards, try to use that vector as up-vector that
    // is closest to (0,1,0) and lies in the YZ place of the camera
    else if (projY > 0.0)
    {
        Matrix upMat = matrix;
        upMat.off = Vector(0.0);
        up = !upMat * Vector(0.0, 1.0, 0.0);
        up.x = 0.0;
        up = upMat * up;
    }
    // if the Y axis points upwards, try to use that vector as up-vector that
    // is closest to (0,-1,0) and lies in the YZ place of the camera
    else
    {
        Matrix upMat = matrix;
        upMat.off = Vector(0.0);
        up = !upMat * Vector(0.0, -1.0, 0.0);
        up.x = 0.0;
        up = upMat * up;
    }
    normalize(up);
}


String OEConverter::toString(
    const Real value)
{
    char buffer[32];
    sprintf(buffer, "%.8g", value);
    buffer[31] = '\0';

    // handle case where sprintf spits out "-0"
    String retVal = buffer;
    if (retVal == "-0") { retVal = "0"; }

    return retVal;
}
