#include <algorithm>
#include <cstdio>

#include "c4d_symbols.h"
#include "stl_containers.h"
#include "dynarray1d.h"
#include "entry_dialog.h"
#include "filepath.h"
#include "oe_converter.h"
#include "oe_export.h"
#include "oe_export_dlg.h"
#include "oe_obj_converter.h"
#include "oe_scenehook.h"
#include "utilities.h"



//-----------------------------------------------------------------------------
// Implementation of public member functions of class OEExportDlg.


const Real OEExportDlg::cExportScales[ID_OE_DLG_NATIVE_UNIT_NUMBER] =
    {
        0.001,    // ID_OE_DLG_NATIVE_UNIT_MM
        0.01,     // ID_OE_DLG_NATIVE_UNIT_CM
        0.1,      // ID_OE_DLG_NATIVE_UNIT_DM
        1.0,      // ID_OE_DLG_NATIVE_UNIT_M
        10.0,     // ID_OE_DLG_NATIVE_UNIT_DECM
        100.0,    // ID_OE_DLG_NATIVE_UNIT_HM
        1000.0,   // ID_OE_DLG_NATIVE_UNIT_KM
        0.0254,   // ID_OE_DLG_NATIVE_UNIT_IN
        0.3048,   // ID_OE_DLG_NATIVE_UNIT_FT
        0.9144,   // ID_OE_DLG_NATIVE_UNIT_YD
        201.168,  // ID_OE_DLG_NATIVE_UNIT_FL
        1609.344  // ID_OE_DLG_NATIVE_UNIT_MI
    };

const Real OEExportDlg::cResolutionScales[ID_OE_DLG_SCALE_NUMBER] =
    {
        1.0, 0.75, 0.5, 0.25
    };


bool OEExportDlg::readExportSettings(
    BaseDocument &document,
    OEConverter  &converter)
{
    // at first, clear converter structure
    converter.clear();

    // store document in converter data
    converter.mDocument = &document;

    // get settings container from OESceneHook
    OESceneHook *sceneHook = OESceneHook::get(*converter.mDocument);
    if (!sceneHook)
    {
        ERRLOG_RETURN_VALUE(false, "OEExportDlg::readExportSettings(): Could not obtain scenehook from active document.");
    }
    BaseContainer *settings = &sceneHook->mExportDlgSettings;

    // get project path and name
    converter.mProjectPath = settings->GetString(IDD_OE_DLG_PROJECT_PATH,
                                                 document.GetDocumentPath().GetString());
    Filename defaultProjectName = document.GetDocumentName();
    defaultProjectName.ClearSuffix();
    converter.mProjectName = settings->GetString(IDD_OE_DLG_PROJECT_NAME,
                                                 defaultProjectName.GetFileString());
    cleanupFilename(converter.mProjectName);

    // set native unit
    BaseContainer worldData = GetWorldContainer();
    LONG defaultUnit = ID_OE_DLG_NATIVE_UNIT_CM;
    if (worldData.GetBool(WPREF_UNITS_USEUNITS))
    {
        switch (worldData.GetLong(WPREF_UNITS_BASIC))
        {
            case 1:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_KM;  break;
            case 2:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_M;   break;
            case 3:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_CM;  break;
            case 4:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_MM;  break;
            case 5:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_MM;  break;
            case 6:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_MM;  break;
            case 7:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_MI;  break;
            case 8:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_YD;  break;
            case 9:  defaultUnit = ID_OE_DLG_NATIVE_UNIT_FT;  break;
            case 10: defaultUnit = ID_OE_DLG_NATIVE_UNIT_IN;  break;
            default: break;
        }
    }
    converter.mExportScaleId = settings->GetLong(IDD_OE_DLG_NATIVE_UNIT, defaultUnit);
    if ((converter.mExportScaleId < 0) ||
        (converter.mExportScaleId >= ID_OE_DLG_NATIVE_UNIT_NUMBER))
    {
        converter.mExportScaleId = ID_OE_DLG_NATIVE_UNIT_CM;
    }
#if _C4D_VERSION >= 120
    converter.mUseDocumentScale = settings->GetBool(IDD_OE_DLG_USE_DOC_SCALE, TRUE);
#else
    converter.mUseDocumentScale = false;
#endif

    // get setting "collect all images"
    converter.mCollectAllImages = settings->GetBool(IDD_OE_DLG_COLLECT_IMAGES, FALSE);

    // get output image path from render settings and use project path as fallback
    RenderData    *c4dRenderData = document.GetActiveRenderData();
    BaseContainer *c4dRenderSettings = 0;
    if (!c4dRenderData)
    {
        ERRLOG("OEExportDlg::readExportSettings(): could not obtain render data");
    }
    else
    {
        c4dRenderSettings = c4dRenderData->GetDataInstance();
        if (!c4dRenderSettings)
        {
            ERRLOG("OEExportDlg::readExportSettings(): could not obtain render settings");
        }
    }
    Filename defaultOutputPath;
    if (c4dRenderSettings)
    {
        defaultOutputPath = c4dRenderSettings->GetFilename(RDATA_PATH);
    }
    if (!defaultOutputPath.Content())
    {
        defaultOutputPath = converter.mProjectPath + converter.mProjectName;
    }
    converter.mOutputFilename = settings->GetString(IDD_OE_DLG_IMAGE_PATH,
                                                    defaultOutputPath.GetString());
    converter.mImageFormatId = settings->GetLong(IDD_OE_DLG_IMAGE_FORMAT,
                                                 ID_OE_DLG_IMAGE_FORMAT_PNG8);

    // get maximum samples
    converter.mSamples = settings->GetLong(IDD_OE_DLG_PIXEL_SAMPLES);
    if (converter.mSamples <= 0)
    {
        converter.mSamples = settings->GetLong(IDD_OE_DLG_PIXEL_SAMPLES_CUSTOM);
        if (converter.mSamples <= 0)
        {
            converter.mSamples = ID_OE_DLG_PIXEL_SAMPLES_MEDIUM;
        }
    }

    // get render target list and the index of the currently selected render target
    const BaseContainer *renderTargetList =
        settings->GetContainerInstance(IDD_OE_DLG_RENDERTARGET_LIST);
    converter.mRenderTargets.clear();
    if (renderTargetList)
    {
        for (LONG id=1; ; ++id)
        {
            const String renderTarget = renderTargetList->GetString(id);
            if (!renderTarget.Content()) { break; }
            converter.mRenderTargets.push_back(renderTarget);
        }
    }
    converter.mRenderTargetIx = settings->GetLong(IDD_OE_DLG_RENDERTARGET);

    // get settings of the currently selected render target
    BaseContainer *rtSettings = getRenderTargetSettings(*settings);

    // obtain active render and setup film group
    converter.mExportResolution = rtSettings->GetBool(IDD_OE_DLG_EXPORT_RESOLUTION);
    LONG width  = 1024;
    LONG height = 512;
    if (c4dRenderSettings)
    {
        width  = c4dRenderSettings->GetLong(RDATA_XRES, width);
        height = c4dRenderSettings->GetLong(RDATA_YRES, height);
    }
    converter.mWidth             = rtSettings->GetLong(IDD_OE_DLG_RES_WIDTH,  width);
    converter.mHeight            = rtSettings->GetLong(IDD_OE_DLG_RES_HEIGHT, height);
    converter.mResolutionScaleId = rtSettings->GetLong(IDD_OE_DLG_RES_SCALE,  ID_OE_DLG_SCALE_100);
    if ((converter.mResolutionScaleId < 0) ||
        (converter.mResolutionScaleId >= ID_OE_DLG_SCALE_NUMBER))
    {
        converter.mResolutionScaleId = ID_OE_DLG_SCALE_100;
    }

    // obtain camera settings
    converter.mExportCamera = rtSettings->GetBool(IDD_OE_DLG_EXPORT_CAMERA);
    BaseObject *camera = rtSettings->GetObjectLink(IDD_OE_DLG_ACTIVE_CAMERA, &document);
    LONG cameraIx = getObjectIx(document, camera, Ocamera);
    if (!camera || (cameraIx < 0)) { cameraIx = 0; }
    else                           { ++cameraIx; }
    converter.mCameraIx         = cameraIx;
    converter.mExportAperture   = rtSettings->GetBool(IDD_OE_DLG_EXPORT_LENS_APERTURE);
    converter.mAperture         = rtSettings->GetReal(IDD_OE_DLG_LENS_APERTURE);
    converter.mExportFocalDepth = rtSettings->GetBool(IDD_OE_DLG_EXPORT_FOCAL_DEPTH);
    converter.mFocalDepth       = rtSettings->GetBool(IDD_OE_DLG_FOCUS_OF_CAM_OBJECT) ?
                                  0.0 : rtSettings->GetReal(IDD_OE_DLG_FOCAL_DEPTH);
    converter.mExportMotionBlur = rtSettings->GetBool(IDD_OE_DLG_EXPORT_MOTION_BLUR);
    converter.mMotionBlurToNext = (rtSettings->GetLong(IDD_OE_DLG_MOTION_BLUR) == ID_OE_DLG_MOTION_BLUR_NEXT) ?
                                  true : false;

    // daylight
    converter.mExportSunDirection = rtSettings->GetBool(IDD_OE_DLG_EXPORT_SUN_DIR);
    BaseObject *light = rtSettings->GetObjectLink(IDD_OE_DLG_SUN_LIGHT, &document);
    if (!light) { light = findObject(document, 0, Olight, Tsunexpression); }
    converter.mLightIx = light ? getObjectIx(document, light, Olight) : 0;

    return true;
}


OEExportDlg::OEExportDlg()
:
    mDocument(0),
    mSettings(0),
    mRenderTargetSettings(0),
    mAnimConverter(0)
{
    mCameras = AtomArray::Alloc();
    mLights  = AtomArray::Alloc();
}


OEExportDlg::~OEExportDlg()
{
    if (mCameras) { AtomArray::Free(mCameras); }
    if (mLights)  { AtomArray::Free(mLights); }
}


void OEExportDlg::run()
{
#if _C4D_VERSION >= 120
    Open(DLG_TYPE_ASYNC, PID_OE_EXPORT);
#else
    Open(TRUE, PID_OE_EXPORT);
#endif
}


Bool OEExportDlg::CreateLayout()
{
    if (LoadDialogResource(OE_EXPORT_DLG, 0, 0))
    {
#if _C4D_VERSION<120
        RemoveElement(IDD_OE_DLG_USE_DOC_SCALE);
#endif  
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}


Bool OEExportDlg::InitValues()
{
    // make sure that the atom arrays were allocated correctly
    if (!mCameras || !mLights)
    {
        ERRLOG_RETURN_VALUE(false, "OEExportDlg::InitValues(): Could not allocate internal arrays.");
    }

    // get active document
    mDocument = GetActiveDocument();
    if (!mDocument)
    {
        ERRLOG_RETURN_VALUE(FALSE, "OEExportDlg::InitValues(): Could not obtain active document.");
    }

    // read export settings
    OEConverter converter;
    if (!readExportSettings(*mDocument, converter))
    {
        ERRLOG_RETURN_VALUE(FALSE, "OEExportDlg::InitValues(): Could not read export settings.");
    }

    // get OESceneHook and its container, which contains the current Octane export settings
    OESceneHook *sceneHook = OESceneHook::get(*mDocument);
    if (!sceneHook)
    {
        Close();
        ERRLOG_RETURN_VALUE(FALSE, "OEExportDlg::InitValues(): Could not get scene hook.");
    }
    mSettings = &sceneHook->mExportDlgSettings;


    //--- initialise project configuration with the read settings
    SetFilename(IDD_OE_DLG_PROJECT_PATH,   converter.mProjectPath);
    SetString  (IDD_OE_DLG_PROJECT_NAME,   converter.mProjectName);
    SetLong    (IDD_OE_DLG_NATIVE_UNIT,    converter.mExportScaleId);
#if _C4D_VERSION >= 120
    SetBool    (IDD_OE_DLG_USE_DOC_SCALE,  converter.mUseDocumentScale);
#endif
    SetBool    (IDD_OE_DLG_COLLECT_IMAGES, converter.mCollectAllImages);

    // init output path
    SetFilename(IDD_OE_DLG_IMAGE_PATH,      converter.mOutputFilename);
    SetLong    (IDD_OE_DLG_IMAGE_FORMAT,    converter.mImageFormatId);

    // init animation settings
    LONG fps = mDocument->GetFps();
    copyLongFrom(IDD_OE_DLG_START_FRAME,           mSettings, mDocument->GetMinTime().GetFrame(fps));
    copyLongFrom(IDD_OE_DLG_END_FRAME,             mSettings, mDocument->GetMaxTime().GetFrame(fps));
    copyLongFrom(IDD_OE_DLG_STEP,                  mSettings, 1, 1);
    copyBoolFrom(IDD_OE_DLG_ONLY_CAMERA_ANIMATION, mSettings);
    copyLongFrom(IDD_OE_DLG_PIXEL_SAMPLES,         mSettings, ID_OE_DLG_PIXEL_SAMPLES_MEDIUM);
    copyLongFrom(IDD_OE_DLG_PIXEL_SAMPLES_CUSTOM,  mSettings, 64, 1);

    // disable all fields that will not be exported
    toggleNativeUnit();
    togglePixelSamples();

    //--- initialise render target settings

    mRenderTargetSettings = getRenderTargetSettings(*mSettings);

    // populate combo box with all render targets + an entry at the top for the
    // preview render target
    FreeChildren(IDD_OE_DLG_RENDERTARGET);
    AddChild(IDD_OE_DLG_RENDERTARGET, 0, GeLoadString(IDS_OE_DLG_RENDERTARGET_PREVIEW));
    for (size_t i=0; i<converter.mRenderTargets.size(); ++i)
    {
        AddChild(IDD_OE_DLG_RENDERTARGET, (LONG)i+1, converter.mRenderTargets[i]);
    }
    SetLong(IDD_OE_DLG_RENDERTARGET,   converter.mRenderTargetIx);

    // populate combo box with all camera objects + an entry at the top for the
    // current render camera and 
    FreeChildren(IDD_OE_DLG_ACTIVE_CAMERA);
    AddChild(IDD_OE_DLG_ACTIVE_CAMERA, 0, GeLoadString(IDS_OE_DLG_RENDER_CAMERA));
    collectAllObjects(*mDocument, Ocamera, *mCameras);
    for (LONG i=0; i<mCameras->GetCount(); ++i)
    {
        AddChild(IDD_OE_DLG_ACTIVE_CAMERA, i+1, ((BaseList2D*)mCameras->GetIndex(i))->GetName());
    }

    // populate combo box with all light objects
    FreeChildren(IDD_OE_DLG_SUN_LIGHT);
    collectAllObjects(*mDocument, Olight, *mLights);
    for (LONG i=0; i<mLights->GetCount(); ++i)
    {
        AddChild(IDD_OE_DLG_SUN_LIGHT, i, ((BaseObject*)mLights->GetIndex(i))->GetName());
    }

    // film configuration:
    SetBool    (IDD_OE_DLG_EXPORT_RESOLUTION, converter.mExportResolution);
    SetLong    (IDD_OE_DLG_RES_WIDTH,         converter.mWidth,  4, 16000);
    SetLong    (IDD_OE_DLG_RES_HEIGHT,        converter.mHeight, 4, 16000);
    SetLong    (IDD_OE_DLG_RES_SCALE,         converter.mResolutionScaleId);

    // camera settings:
    SetBool    (IDD_OE_DLG_EXPORT_CAMERA,        converter.mExportCamera);
    SetLong    (IDD_OE_DLG_ACTIVE_CAMERA,        converter.mCameraIx);
    SetBool    (IDD_OE_DLG_EXPORT_LENS_APERTURE, converter.mExportAperture);
    SetReal    (IDD_OE_DLG_LENS_APERTURE,        converter.mAperture, 0.0, 10.0, 0.1, FORMAT_REAL, 0.0, 20.0);
    SetBool    (IDD_OE_DLG_EXPORT_FOCAL_DEPTH,   converter.mExportFocalDepth);
    SetReal    (IDD_OE_DLG_FOCAL_DEPTH,          converter.mFocalDepth, 0.0, 100.0, 0.1, FORMAT_REAL, 0.0, MAXREALr);
    copyBoolFrom(IDD_OE_DLG_FOCUS_OF_CAM_OBJECT, mRenderTargetSettings);
    SetBool    (IDD_OE_DLG_EXPORT_MOTION_BLUR,   converter.mExportMotionBlur);
    SetLong    (IDD_OE_DLG_MOTION_BLUR,          converter.mMotionBlurToNext ? ID_OE_DLG_MOTION_BLUR_NEXT
                                                                             : ID_OE_DLG_MOTION_BLUR_PREVIOUS);

    // daylight configuration:
    SetBool    (IDD_OE_DLG_EXPORT_SUN_DIR, converter.mExportSunDirection);
    if (mLights->GetCount())
    {
        SetLong    (IDD_OE_DLG_SUN_LIGHT,  converter.mLightIx);
    }

    // disable all fields that will not be exported
    toggleRenderTargetButtons();
    toggleFilm();
    toggleCamera();
    toggleLight();

    return TRUE;
}


Bool OEExportDlg::CoreMessage(
    LONG  id,
    const BaseContainer& msg)
{
    if (id == EVMSG_CHANGE)
    {
        StopAllThreads();
        InitValues();
    }
    return TRUE;
}


Bool OEExportDlg::Command(
    LONG                id,
    const BaseContainer &msg)
{
    switch (id)
    {
        // project path
        case IDD_OE_DLG_PROJECT_PATH:
            copyStringTo(id, mSettings);
            break;
        case IDB_OE_DLG_PROJECT_PATH:
        {
            Filename projectPath;
            GetFilename(IDD_OE_DLG_PROJECT_PATH, projectPath);
            if (fileSelect(projectPath,
                           FILESELECTTYPE_ANYTHING,
                           FILESELECT_DIRECTORY,
                           GeLoadString(IDS_OE_DLG_PROJECT_PATH_FS_TITLE)))
            {
                SetFilename(IDD_OE_DLG_PROJECT_PATH, projectPath);
                copyStringTo(IDD_OE_DLG_PROJECT_PATH, mSettings);
            }
            break;
        }

        // project name
        case IDD_OE_DLG_PROJECT_NAME:
            cleanupFilename(IDD_OE_DLG_PROJECT_NAME);
            copyStringTo(id, mSettings);
            break;

        // "use existing project" allows the user to choose an already existing
        // OCS file, which will then fill in project path and name
        case IDB_OE_DLG_USE_EXISTING_PROJECT:
        {
            Filename projectPath;
            GetFilename(IDD_OE_DLG_PROJECT_PATH, projectPath);
            if (fileSelect(projectPath,
                           FILESELECTTYPE_ANYTHING,
                           FILESELECT_LOAD,
                           GeLoadString(IDB_OE_DLG_USE_EXISTING_PROJECT_FS_TITLE)))
            {
                projectPath.ClearSuffix();
                SetFilename(IDD_OE_DLG_PROJECT_PATH, projectPath.GetDirectory());
                copyStringTo(IDD_OE_DLG_PROJECT_PATH, mSettings);
                SetFilename(IDD_OE_DLG_PROJECT_NAME, projectPath.GetFile());
                cleanupFilename(IDD_OE_DLG_PROJECT_NAME);
                copyStringTo(IDD_OE_DLG_PROJECT_NAME, mSettings);
            }
            break;
        }

        // native unit
        case IDD_OE_DLG_NATIVE_UNIT:
            copyLongTo(id, mSettings);
            break;

        // enabler for native unit combo box
        case IDD_OE_DLG_USE_DOC_SCALE:
            copyBoolTo(id, mSettings);
            toggleNativeUnit();
            break;

        // collect all images
        case IDD_OE_DLG_COLLECT_IMAGES:
            copyBoolTo(id, mSettings);
            break;

        // animation image path
        case IDD_OE_DLG_IMAGE_PATH:
            copyStringTo(id, mSettings);
            break;
        case IDB_OE_DLG_IMAGE_PATH:
        {
            Filename imagePath;
            GetFilename(IDD_OE_DLG_IMAGE_PATH, imagePath);
            if (fileSelect(imagePath,
                           FILESELECTTYPE_ANYTHING,
                           FILESELECT_SAVE,
                           GeLoadString(IDS_OE_DLG_IMAGE_PATH_FS_TITLE)))
            {
                imagePath.ClearSuffix();
                SetFilename(IDD_OE_DLG_IMAGE_PATH, imagePath);
                copyStringTo(IDD_OE_DLG_IMAGE_PATH, mSettings);
            }
            break;
        }

        // animation image type
        case IDD_OE_DLG_IMAGE_FORMAT:
            copyLongTo(id, mSettings);
            break;

        // animation settings
        case IDD_OE_DLG_START_FRAME:
        case IDD_OE_DLG_END_FRAME:
        case IDD_OE_DLG_STEP:
            copyLongTo(id, mSettings);
            break;
        case IDD_OE_DLG_ONLY_CAMERA_ANIMATION:
            copyBoolTo(id, mSettings);
            break;
        case IDD_OE_DLG_PIXEL_SAMPLES:
            copyLongTo(id, mSettings);
            togglePixelSamples();
            break;
        case IDD_OE_DLG_PIXEL_SAMPLES_CUSTOM:
            copyLongTo(id, mSettings);
            break;

        // render target
        case IDD_OE_DLG_RENDERTARGET:
            copyLongTo(id, mSettings);
            InitValues();
            break;

        // add/rename/delete render target
        case IDB_OE_DLG_RENDERTARGET_ADD:
            addRenderTarget();
            InitValues();
            break;
        case IDB_OE_DLG_RENDERTARGET_REN:
            renameRenderTarget();
            InitValues();
            break;
        case IDB_OE_DLG_RENDERTARGET_DEL:
            removeRenderTarget();
            InitValues();
            break;

        // enabler for film export
        case IDD_OE_DLG_EXPORT_RESOLUTION:
            copyBoolTo(id, mRenderTargetSettings);
            toggleFilm();
            break;

        // film width/height/scale
        case IDD_OE_DLG_RES_WIDTH:
        case IDD_OE_DLG_RES_HEIGHT:
        case IDD_OE_DLG_RES_SCALE:
            copyLongTo(id, mRenderTargetSettings);
            break;

        // enabler in camera group
        case IDD_OE_DLG_EXPORT_CAMERA:
        case IDD_OE_DLG_EXPORT_LENS_APERTURE:
        case IDD_OE_DLG_EXPORT_FOCAL_DEPTH:
        case IDD_OE_DLG_FOCUS_OF_CAM_OBJECT:
        case IDD_OE_DLG_EXPORT_MOTION_BLUR:
            copyBoolTo(id, mRenderTargetSettings);
            toggleCamera();
            break;

        // a camera has been chosen
        case IDD_OE_DLG_ACTIVE_CAMERA:
        {
            LONG activeCamIx = getLong(IDD_OE_DLG_ACTIVE_CAMERA);
            if (activeCamIx == 0)
            {
                mRenderTargetSettings->SetVoid(IDD_OE_DLG_ACTIVE_CAMERA, 0);
            }
            else
            {
                mRenderTargetSettings->SetLink(IDD_OE_DLG_ACTIVE_CAMERA,
                                               (BaseObject*)mCameras->GetIndex(activeCamIx-1));
            }
            mDocument->SetChanged();
            break;
        }

        // aperture and focal length
        case IDD_OE_DLG_LENS_APERTURE:
        case IDD_OE_DLG_FOCAL_DEPTH:
            copyRealTo(id, mRenderTargetSettings);
            break;

        // motion blur setting
        case IDD_OE_DLG_MOTION_BLUR:
            copyLongTo(id, mRenderTargetSettings);
            break;

        // daylight environment
        case IDD_OE_DLG_EXPORT_SUN_DIR:
            copyBoolTo(id, mRenderTargetSettings);
            toggleLight();
            break;
        case IDD_OE_DLG_SUN_LIGHT:
        {
            LONG sunLightIx = getLong(IDD_OE_DLG_SUN_LIGHT);
            if (sunLightIx < mLights->GetCount())
            {
                mRenderTargetSettings->SetLink(IDD_OE_DLG_SUN_LIGHT,
                                               (BaseObject*)mLights->GetIndex(sunLightIx));
                mDocument->SetChanged();
            }
            break;
        }

        // export OBJ/MTL
        case IDB_OE_DLG_EXPORT_ONLY:
        {
            OEConverter converter;
            if (readExportSettings(*mDocument, converter))
            {
                converter.exportToObj(false);
            }
            break;
        }

        // export OBJ/MTL + render frame
        case IDB_OE_DLG_RENDER:
        {
            OEConverter converter;
            if (readExportSettings(*mDocument, converter))
            {
                if (converter.exportToObj(false))
                {
                    converter.launchOctane(false, 0);
                }
            }
            break;
        }

        // render animation
        case IDB_OE_DLG_RENDER_ANIM:
            exportAnimation();
            break;

        // cancel/stop animation export
        case IDB_OE_DLG_ANIM_CANCEL:
            stopAnimationExport();
            break;

        default:
            break;
    }
    return TRUE;
}


Bool OEExportDlg::AskClose()
{
    if (mAnimConverter && mAnimConverter->octaneIsActive())
    {
        return (GeOutString(GeLoadString(IDS_OE_DLG_ASK_CLOSE), GEMB_YESNO) != GEMB_R_YES);
    }
    return FALSE;
}


void OEExportDlg::Timer(
    const BaseContainer &msg)
{
    // if Octane is still rendering, export next frame after 10s, if not done already
    if (mAnimConverter->octaneIsActive())
    {
        mTimeSinceOctaneLaunch += cTimerInterval;
        if ((mExportedFrame != mCurrentFrame) &&
            (mCurrentFrame <= mEndFrame) &&
            (mTimeSinceOctaneLaunch >= cExportDelay))
        {
            if (!exportCurrentFrame()) { goto ERROR; }
        }
    }
    // if Octane stopped rendering and the image has been rendered
    else 
    {
        mTimeSinceOctaneExit += cTimerInterval;
        if (mAnimConverter->imageRendered())
        {
            mFailureCount = 0;
            // if not all frames have been rendered yet:
            if (mCurrentFrame <= mEndFrame)
            {
                // if next frame wasn't exported yet, do it now
                if (mExportedFrame != mCurrentFrame)
                {
                    if (!exportCurrentFrame()) { goto ERROR; }
                }
                // the next frame to render is now exported, let's render it if enough
                // time has passed
                if (mTimeSinceOctaneExit >= cNextFrameDelay)
                {
                    if (!renderCurrentFrame()) { goto ERROR; }
                }
            }
            // if all frames have been rendered, stop the animation export
            else
            {
                stopAnimationExport();
            }
        }
        // if the image wasn't rendered successfully, try again (up to 5x) if a
        // minimum time has passed, since sometimes the rendered image file doesn't
        // appear immediately
        else if (mTimeSinceOctaneExit >= cMaxRetryDelay)
        {
            ++mFailureCount;
            if (mFailureCount >= cMaxRetries) { goto ERROR; }
            setMessage(GeLoadString(IDS_OE_DLG_ANIM_MSG_FRAME_FAILED,
                                    LongToString(mCurrentFrame - mStartFrame + 1 - mStep),
                                    LongToString(mFailureCount+1),
                                    LongToString(cMaxRetries)));
            if (!mAnimConverter->relaunchOctane()) { goto ERROR; }
            mTimeSinceOctaneLaunch = 0;
            mTimeSinceOctaneExit   = 0;
        }
    }
    return;

ERROR:
    setMessage(GeLoadString(IDS_OE_DLG_ANIM_MSG_FAILED,
                            LongToString(mCurrentFrame - mStartFrame + 1 - mStep),
                            LongToString(mEndFrame - mStartFrame + 1)));
    SetTimer(0);
    mCurrentFrame = MAXLONGl;
}


void OEExportDlg::DestroyWindow()
{
    stopAnimationExport();
}



/*****************************************************************************
* Implementation of private member functions of class OEExportDlg.
*****************************************************************************/


const LONG  OEExportDlg::cTimerInterval(500);
const ULONG OEExportDlg::cMaxRetries(5);
const LONG  OEExportDlg::cExportDelay(10000);
const LONG  OEExportDlg::cMaxRetryDelay(5000);
const LONG  OEExportDlg::cNextFrameDelay(2000);


void OEExportDlg::copyBoolFrom(
    const LONG          id,
    const BaseContainer *container,
    const bool          preset)
{
    SetBool(id, container ? container->GetBool(id, preset) : preset);
}


void OEExportDlg::copyBoolTo(
    const LONG    id,
    BaseContainer *container)
{
    if (container)
    {
        GetBool(id, container, id);
        mDocument->SetChanged();
    }
}


void OEExportDlg::copyLongFrom(
    const LONG          id,
    const BaseContainer *container,
    const LONG          preset,
    const LONG          min,
    const LONG          max,
    const LONG          step)
{
    SetLong(id, container ? container->GetLong(id, preset) : preset, min, max, step);
}

void OEExportDlg::copyLongTo(
    const LONG    id,
    BaseContainer *container)
{
    if (container)
    {
        GetLong(id, container, id);
        mDocument->SetChanged();
    }
}


void OEExportDlg::copyRealFrom(
    const LONG          id,
    const BaseContainer *container,
    const Real          preset,
    const Real          min,
    const Real          max,
    const Real          min2,
    const Real          max2,
    const Real          step,
    const LONG          format)
{
    SetReal(id,
            container ? container->GetReal(id, preset) : preset,
            min,
            max,
            step,
            format,
            min2,
            max2);
}


void OEExportDlg::copyRealTo(
    const LONG    id,
    BaseContainer *container)
{
    if (container)
    {
        GetReal(id, container, id);
        mDocument->SetChanged();
    }
}


void OEExportDlg::copyStringFrom(
    const LONG          id,
    const BaseContainer *container,
    const String        &preset)
{
    SetString(id, container ? container->GetString(id, preset) : preset);
}

void OEExportDlg::copyStringTo(
    const LONG    id,
    BaseContainer *container)
{
    if (container)
    {
        GetString(id, container, id);
        mDocument->SetChanged();
    }
}


bool OEExportDlg::getBool(
    const LONG id,
    const bool preset)
{
    Bool retVal;
    if (!GetBool(id, retVal)) { retVal = preset; }
    return retVal;
}

LONG OEExportDlg::getLong(
    const LONG id,
    const LONG preset)
{
    LONG retVal;
    if (!GetLong(id, retVal)) { retVal = preset; }
    return retVal;
}

Real OEExportDlg::getReal(
    const LONG id,
    const Real preset)
{
    Real retVal;
    if (!GetReal(id, retVal)) { retVal = preset; }
    return retVal;
}

String OEExportDlg::getString(
    const LONG   id,
    const String preset)
{
    String retVal;
    if (!GetString(id, retVal)) { retVal = preset; }
    return retVal;
}


void OEExportDlg::cleanupFilename(
    String& str)
{
    LONG pos = 0;
    while (str.FindFirst(' ', &pos, pos))
    {
        str[pos] = '_';
    }
}


void OEExportDlg::cleanupFilename(
    const LONG id)
{
    String str;
    if (GetString(id, str))
    {
        cleanupFilename(str);
        SetString(id, str);
    }
}


void OEExportDlg::toggleRenderTargetButtons()
{
    const bool isPreview = (getLong(IDD_OE_DLG_RENDERTARGET) == 0);
    Enable(IDB_OE_DLG_RENDERTARGET_REN, !isPreview);
    Enable(IDB_OE_DLG_RENDERTARGET_DEL, !isPreview);
}


void OEExportDlg::toggleNativeUnit()
{
    bool enabled = getBool(IDD_OE_DLG_USE_DOC_SCALE);
    Enable(IDD_OE_DLG_NATIVE_UNIT, !enabled);
}


void OEExportDlg::toggleFilm()
{
    bool enabled = getBool(IDD_OE_DLG_EXPORT_RESOLUTION);
    Enable(IDS_OE_DLG_RES_WIDTH,  enabled);
    Enable(IDD_OE_DLG_RES_WIDTH,  enabled);
    Enable(IDS_OE_DLG_RES_HEIGHT, enabled);
    Enable(IDD_OE_DLG_RES_HEIGHT, enabled);
    Enable(IDD_OE_DLG_RES_SCALE,  enabled);
}


void OEExportDlg::toggleCamera()
{
    bool enabled = getBool(IDD_OE_DLG_EXPORT_CAMERA);
    Enable(IDS_OE_DLG_ACTIVE_CAMERA,        enabled);
    Enable(IDD_OE_DLG_ACTIVE_CAMERA,        enabled);
    Enable(IDD_OE_DLG_EXPORT_LENS_APERTURE, enabled);
    Enable(IDD_OE_DLG_LENS_APERTURE,        enabled && getBool(IDD_OE_DLG_EXPORT_LENS_APERTURE));
    Enable(IDD_OE_DLG_EXPORT_FOCAL_DEPTH,   enabled);
    Enable(IDD_OE_DLG_FOCAL_DEPTH,          enabled && getBool(IDD_OE_DLG_EXPORT_FOCAL_DEPTH)
                                            && !getBool(IDD_OE_DLG_FOCUS_OF_CAM_OBJECT));
    Enable(IDD_OE_DLG_FOCUS_OF_CAM_OBJECT,  enabled && getBool(IDD_OE_DLG_EXPORT_FOCAL_DEPTH));
    Enable(IDD_OE_DLG_EXPORT_MOTION_BLUR,   enabled);
    Enable(IDS_OE_DLG_MOTION_BLUR,          enabled && getBool(IDD_OE_DLG_EXPORT_MOTION_BLUR));
    Enable(IDD_OE_DLG_MOTION_BLUR,          enabled && getBool(IDD_OE_DLG_EXPORT_MOTION_BLUR));
}


void OEExportDlg::toggleLight()
{
    bool enabled = getBool(IDD_OE_DLG_EXPORT_SUN_DIR);
    Enable(IDS_OE_DLG_SUN_LIGHT, enabled);
    Enable(IDD_OE_DLG_SUN_LIGHT, enabled);
}


void OEExportDlg::togglePixelSamples()
{
    bool enabled = (getLong(IDD_OE_DLG_PIXEL_SAMPLES) == ID_OE_DLG_PIXEL_SAMPLES_CUSTOM);
    Enable(IDD_OE_DLG_PIXEL_SAMPLES_CUSTOM, enabled);
}


String OEExportDlg::toString(
    const Real value)
{
    char buffer[32];
    sprintf(buffer, "%.8g", value);
    buffer[31] = '\0';
    return buffer;
}


BaseContainer* OEExportDlg::getRenderTargetSettings(
    BaseContainer &settings)
{
    const LONG    renderTargetIx = settings.GetLong(IDD_OE_DLG_RENDERTARGET, 0);

    BaseContainer *allRTSettings =
        settings.GetContainerInstance(IDD_OE_DLG_RENDERTARGET_SETTINGS);
    if (!allRTSettings)
    {
        settings.SetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS, BaseContainer());
        allRTSettings = settings.GetContainerInstance(IDD_OE_DLG_RENDERTARGET_SETTINGS);
    }

    BaseContainer *rtSettings = allRTSettings->GetContainerInstance(renderTargetIx);
    if (!rtSettings)
    {
        allRTSettings->SetContainer(renderTargetIx, allRTSettings->GetContainer(0));
        rtSettings = allRTSettings->GetContainerInstance(renderTargetIx);
    }

    return rtSettings;
}

    
class RenderTargetSorter
{
public:

    RenderTargetSorter(
        const BaseContainer &renderTargets,
        const String        &newRenderTarget)
    :
        mRenderTargets(&renderTargets),
        mNewRenderTarget(&newRenderTarget)
    {}

    bool operator()(
        const LONG a,
        const LONG b)
    {
        if (a == 0) { return true; }
        if (b == 0) { return false; }

        const String aName((a > 0) ? mRenderTargets->GetString(a) : *mNewRenderTarget);
        const String bName((b > 0) ? mRenderTargets->GetString(b) : *mNewRenderTarget);

        return (aName.LexCompare(bName) < 0);
    }

private:

    const BaseContainer *mRenderTargets;
    const String        *mNewRenderTarget;
};


void OEExportDlg::addRenderTarget()
{
    // let user enter a new name
    EntryDialog dlg;
    if ((dlg.openModal("Add Render Target...",
                       "Please enter the name of the new render target:",
                       "",
                       EntryDialog::BTN_OK_CANCEL) == EntryDialog::RET_CANCEL)
        || !dlg.lastValue().Content())
    {
        return;
    }
    const String& newRenderTarget = dlg.lastValue();

    // create list of all render targets and add the new one to the end
    const BaseContainer oldRenderTargets        = mSettings->GetContainer(IDD_OE_DLG_RENDERTARGET_LIST);
    const BaseContainer oldRenderTargetSettings = mSettings->GetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS);
    StdVector<LONG>     sortedRenderTargets(1, 0);
    for (LONG id = 1; ; ++id)
    {
        if (!oldRenderTargets.GetString(id).Content()) { break; }
        sortedRenderTargets.push_back(id);
    }
    sortedRenderTargets.push_back(-1);

    //
    std::sort(sortedRenderTargets.begin(),
              sortedRenderTargets.end(),
              RenderTargetSorter(oldRenderTargets, newRenderTarget));

    //
    LONG          newRenderTargetId = 0;
    BaseContainer newRenderTargets;
    BaseContainer newRenderTargetSettings;
    for (size_t ix = 0; ix < sortedRenderTargets.size(); ++ix)
    {
        const LONG srcId = sortedRenderTargets[ix];
        const LONG dstId = (LONG)ix;
        if (srcId >= 0)
        {
            newRenderTargets.SetString(dstId, oldRenderTargets.GetString(srcId));
            newRenderTargetSettings.SetContainer(dstId, oldRenderTargetSettings.GetContainer(srcId));
        }
        else
        {
            newRenderTargetId = dstId;
            newRenderTargets.SetString(dstId, newRenderTarget);
            newRenderTargetSettings.SetContainer(dstId, oldRenderTargetSettings.GetContainer(0));
        }
    }

    // store new render target list and render target selection in settings
    mSettings->SetContainer(IDD_OE_DLG_RENDERTARGET_LIST,     newRenderTargets);
    mSettings->SetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS, newRenderTargetSettings);
    mSettings->SetLong(IDD_OE_DLG_RENDERTARGET, newRenderTargetId);
}


void OEExportDlg::renameRenderTarget()
{
    // get name of current render target
    LONG renameId = mSettings->GetLong(IDD_OE_DLG_RENDERTARGET);
    if (renameId <= 0) { return; }
    const BaseContainer oldRenderTargets = mSettings->GetContainer(IDD_OE_DLG_RENDERTARGET_LIST);
    String renderTarget = oldRenderTargets.GetString(renameId);
    if (!renderTarget.Content()) { return; }

    // let user enter a new name
    EntryDialog dlg;
    if ((dlg.openModal("Rename Render Target...",
                       "Please enter the name of the new render target:",
                       renderTarget,
                       EntryDialog::BTN_OK_CANCEL) == EntryDialog::RET_CANCEL)
        || !dlg.lastValue().Content())
    {
        return;
    }
    if (!dlg.lastValue().Content()) { return; }
    const String newName = dlg.lastValue();

    // create list of all render targets and set id of render target to rename to -1
    const BaseContainer oldRenderTargetSettings = mSettings->GetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS);
    StdVector<LONG>     sortedRenderTargets(1, 0);
    for (LONG id = 1; ; ++id)
    {
        if (!oldRenderTargets.GetString(id).Content()) { break; }
        sortedRenderTargets.push_back((id != renameId) ? id : -1);
    }

    //
    std::sort(sortedRenderTargets.begin(),
              sortedRenderTargets.end(),
              RenderTargetSorter(oldRenderTargets, newName));

    //
    BaseContainer newRenderTargets;
    BaseContainer newRenderTargetSettings;
    for (size_t ix = 0; ix < sortedRenderTargets.size(); ++ix)
    {
        const LONG srcId = sortedRenderTargets[ix];
        const LONG dstId = (LONG)ix;
        if (srcId >= 0)
        {
            newRenderTargets.SetString(dstId, oldRenderTargets.GetString(srcId));
        }
        else
        {
            renameId = dstId;
            newRenderTargets.SetString(dstId, newName);
        }
        newRenderTargetSettings.SetContainer(dstId, oldRenderTargetSettings.GetContainer(srcId));
    }

    // store new render target list and render target selection in settings
    mSettings->SetContainer(IDD_OE_DLG_RENDERTARGET_LIST,     newRenderTargets);
    mSettings->SetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS, newRenderTargetSettings);
    mSettings->SetLong(IDD_OE_DLG_RENDERTARGET, renameId);
}


void OEExportDlg::removeRenderTarget()
{
    LONG deleteId = mSettings->GetLong(IDD_OE_DLG_RENDERTARGET);
    if (deleteId <= 0) { return; }

    // copy render target names and render target settings into new containers, omitting the
    // render target we want to delete
    const BaseContainer oldRenderTargets        = mSettings->GetContainer(IDD_OE_DLG_RENDERTARGET_LIST);
    const BaseContainer oldRenderTargetSettings = mSettings->GetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS);
    BaseContainer       newRenderTargets, newRenderTargetSettings;
    for (LONG srcId = 0, dstId = srcId; ; ++srcId, ++dstId)
    {
        if (srcId > 0)
        {
            const String renderTarget = oldRenderTargets.GetString(srcId);
            if (!renderTarget.Content())
            {
                if (deleteId >= srcId) { deleteId = srcId - 1; }
                break;
            }
            if (srcId == deleteId) { --dstId; continue; }
            newRenderTargets.SetString(dstId, renderTarget);
        }
        newRenderTargetSettings.SetContainer(dstId, oldRenderTargetSettings.GetContainer(srcId));
    }

    // store new render target list and render target selection in settings
    mSettings->SetContainer(IDD_OE_DLG_RENDERTARGET_LIST,     newRenderTargets);
    mSettings->SetContainer(IDD_OE_DLG_RENDERTARGET_SETTINGS, newRenderTargetSettings);
    mSettings->SetLong(IDD_OE_DLG_RENDERTARGET, deleteId);
}


void OEExportDlg::exportAnimation()
{
    GeAssert(mAnimConverter == NULL);

    AutoAlloc<AliasTrans> aliasTrans;
    BaseDocument*         clonedDoc(0);

    // switch to animation mode - disable everything and enable animation
    // export state
    Enable(IDT_EXPORT_DLG_MAIN_TABS, false);
    SetLong(IDT_EXPORT_DLG_MODE_TABS, IDG_OE_DLG_EXPORT_STATE);

    // allocate new converter and store its pointer in member, as will need
    // it for a while
    mAnimConverter = gNew OEConverter;
    if (!mAnimConverter)
    {
        ERRLOG("OEExportDlg::exportAnimation(): Could not allocate Octane converter.");
        goto ERROR;
    }

    // read export settings from dialog
    if (!readExportSettings(*mDocument, *mAnimConverter))
    {
        ERRLOG("OEExportDlg::exportAnimation(): Could not open animation export dialog.");
        goto ERROR;
    }

    // make sure that we've got a document
    if (!mAnimConverter->mDocument)
    {
        ERRLOG("OEExportDlg::exportAnimation(): No document specified in converter struct!");
        goto ERROR;
    }

    // setup alias translator
    if (!aliasTrans || !aliasTrans->Init(mAnimConverter->mDocument))
    {
        ERRLOG("OEExportDlg::exportAnimation(): Could not initialise alias translator.");
        goto ERROR;
    }

    // clone document and translate base links
    clonedDoc = (BaseDocument*)mAnimConverter->mDocument->GetClone(COPYFLAGS_DOCUMENT,
                                                                   aliasTrans);
    if (!clonedDoc)
    {
        ERRLOG("OEExportDlg::exportAnimation(): Could not clone document.");
        goto ERROR;
    }
    aliasTrans->Translate(true);

    // replace document pointer in converter and store converter handle for later use
    mAnimConverter->mDocument = clonedDoc;

    // store frame range
    mStartFrame = getLong(IDD_OE_DLG_START_FRAME);
    mEndFrame   = getLong(IDD_OE_DLG_END_FRAME);
    if (mStartFrame > mEndFrame)
    {
        LONG tmp = mStartFrame;
        mStartFrame = mEndFrame;
        mEndFrame   = tmp;
    }
    mStep                         = LMax(1, getLong(IDD_OE_DLG_STEP));
    mCurrentFrame                 = mStartFrame;
    mOnlyCameraAnimation          = getBool(IDD_OE_DLG_ONLY_CAMERA_ANIMATION);
    mCameraAnimationSceneExported = false;
    mFailureCount                 = 0;

    // export first frame and render it
    if (!exportCurrentFrame() || !renderCurrentFrame())
    {
        goto ERROR;
    }

    // start timer for exporting the next frame and watching Octane
    SetTimer(cTimerInterval);

    return;

ERROR:
    stopAnimationExport();
    return;
}


void OEExportDlg::stopAnimationExport()
{
    // cleanup
    if (mAnimConverter)
    {
        SetTimer(0);
        StopAllThreads();
        BaseDocument::Free(mAnimConverter->mDocument);
        gDelete(mAnimConverter);
    }

    // switch to animation mode - disable everything and enable animation
    // export state
    Enable(IDT_EXPORT_DLG_MAIN_TABS, true);
    SetLong(IDT_EXPORT_DLG_MODE_TABS, IDG_OE_DLG_EXPORT_BUTTONS);
}


///
void OEExportDlg::setMessage(
    const String& message)
{
    // clear group and add new static text with message
    if (LayoutFlushGroup(IDG_OE_DLG_ANIM_MESSAGE))
    {
        AddStaticText(IDD_OE_DLG_ANIM_MESSAGE,
                      BFH_LEFT | BFH_SCALE,
                      0, 0,
                      message,
                      BORDER_NONE);
        LayoutChanged(IDG_OE_DLG_ANIM_MESSAGE);
    }
}


bool OEExportDlg::exportCurrentFrame()
{
    // ok, this is a bit more complicated, we basically have these three
    // possible scenarios
    //
    // without camera motion blur:
    //
    //  ----------------|----------------|----------------|----------------
    //     ...   mCurrentFrame-1    mCurrentFrame    mCurrentFrame+1  ...
    //                               (exported)
    //
    // camera motion blur with interpolation to next frame:
    //
    //                              mMBCamMat1  mMBCamMat2
    //  ----------------|-----------------|--------+--------|----------------
    //     ...   mCurrentFrame-1    mCurrentFrame    mCurrentFrame+1  ...
    //                               (exported)
    //
    // camera motion blur with interpolation from previous frame:
    //
    //                     mMBCamMat1  mMBCamMat2
    //  ----------------|--------+--------|-----------------|----------------
    //     ...   mCurrentFrame-1    mCurrentFrame    mCurrentFrame+1  ...
    //                               (exported)


    // if this the first frame to render, write rendering message
    if (mCurrentFrame == mStartFrame)
    {
        setMessage(GeLoadString(IDS_OE_DLG_ANIM_MSG_RENDERING,
                                LongToString(mCurrentFrame - mStartFrame + 1),
                                LongToString(mEndFrame - mStartFrame + 1)));
    }

    // determine if the current frame represents the beginning or the end of the
    // MB interval
    const bool currentIsFrame1 = (!mAnimConverter->mExportMotionBlur ||
                                  mAnimConverter->mMotionBlurToNext);

    // set document to frame 1
    const LONG fps = mAnimConverter->mDocument->GetFps();
    const LONG frame1 = currentIsFrame1 ? mCurrentFrame * 2
                                        : mCurrentFrame * 2 - 1;
    mAnimConverter->mDocument->SetTime(BaseTime(frame1, 2*fps));
#if _C4D_VERSION>=120
    mAnimConverter->mDocument->ExecutePasses(NULL,
                                             TRUE,
                                             TRUE,
                                             TRUE,
                                             BUILDFLAGS_EXTERNALRENDERER);
#elif _C4D_VERSION>=115
    mAnimConverter->mDocument->ExecutePasses(NULL, TRUE, TRUE, TRUE);
#else
    mAnimConverter->mDocument->AnimateDocument(NULL, TRUE, TRUE);
#endif

    // export geometry if the current frame is the first frame
    bool retVal = true;
    if (currentIsFrame1) { retVal = exportCurrentState(); }

    // get camera matrix and focal depth (for current frame only)
    CameraObject* cam = mAnimConverter->getCamera();
    if (cam)
    {
        mMBCamMat1     = cam->GetMgn();
    }

    // if motion blur is enabled get camera matrix of other frame
    mMBCamMatsValid = false;
    if (cam && mAnimConverter->mExportMotionBlur)
    {
        // do a 50% motion blur
        const LONG frame2 = frame1 + 1;
        mAnimConverter->mDocument->SetTime(BaseTime(frame2, 2*fps));
#if _C4D_VERSION>=120
        mAnimConverter->mDocument->ExecutePasses(NULL,
                                                 TRUE,
                                                 TRUE,
                                                 TRUE,
                                                 BUILDFLAGS_EXTERNALRENDERER);
#elif _C4D_VERSION>=115
        mAnimConverter->mDocument->ExecutePasses(NULL, TRUE, TRUE, TRUE);
#else
        mAnimConverter->mDocument->AnimateDocument(NULL, TRUE, TRUE);
#endif

        // if interpolate from previous frame, export the current state now
        if (!mAnimConverter->mMotionBlurToNext)
        {
            retVal = exportCurrentState();
        }

        // store second camera matrix
        mMBCamMat2      = cam->GetMgn();
        mMBCamMatsValid = true;
    }

    return retVal;
}


bool OEExportDlg::exportCurrentState()
{
    bool retVal = true;

    if (!mOnlyCameraAnimation)
    {
        retVal = mAnimConverter->exportToObj(true);
    }
    else if (!mCameraAnimationSceneExported)
    {
        retVal = mAnimConverter->exportToObj(false);
        mCameraAnimationSceneExported = retVal;
    }
    mExportedFrame = mCurrentFrame;

    return retVal;
}


bool OEExportDlg::renderCurrentFrame()
{
    // set message string
    setMessage(GeLoadString(IDS_OE_DLG_ANIM_MSG_RENDERING,
                            LongToString(mCurrentFrame - mStartFrame + 1),
                            LongToString(mEndFrame - mStartFrame + 1)));

    // launch Octane
    if (!mAnimConverter->launchOctane(true,
                                      mCurrentFrame,
                                      mMBCamMatsValid ? &mMBCamMat1 : 0,
                                      mMBCamMatsValid ? &mMBCamMat2 : 0))
    {
        return false;
    }

    // reset timer
    mTimeSinceOctaneLaunch = 0;
    mTimeSinceOctaneExit   = 0;

    // switch to next frame
    mCurrentFrame += mStep;
    return true;
}
