#include <algorithm>
#include <cstdio>

#include <customgui_datetime.h>
#include <mmaterial.h>

#include "filepath.h"
#include "oe_obj_converter.h"
#include "oe_obj_material.h"
#include "utilities.h"



static const Real cDefaultGamma = 2.2;



//-----------------------------------------------------------------------------
// Helper classes for storing geometric stuff in maps.


/// Helper class that stores a normal and provides a less operator, needed for
/// the rb-trees.
struct NormalHelper
{
    SReal c[3];

    NormalHelper() {}
    NormalHelper(const SVector& v) { c[0] = v.x;  c[1] = v.y;  c[2] = v.z; }

    bool operator<(const NormalHelper& other) const
    {
        static const SReal sEpsilon((SReal)0.00001);

        for (ULONG i=0; i<3; ++i)
        {
            if (fabs(c[i] - other.c[i]) > sEpsilon) { return c[i] < other.c[i]; }
        }
        return false;
    }
};

typedef RBTreeMap<NormalHelper, ULONG, true>  NormalCountsT;


/// Helper class that stores a UV coordinate and provides a less operator,
/// needed for the rb-trees.
struct UVHelper
{
    Real c[2];

    UVHelper() {}
    UVHelper(const Vector& v) { c[0] = v.x;  c[1] = v.y; }

    bool operator<(const UVHelper& other) const
    {
        static const Real sEpsilon = 0.00001;

        for (ULONG i=0; i<2; ++i)
        {
            if (fabs(c[i] - other.c[i]) > sEpsilon) { return c[i] < other.c[i]; }
        }
        return false;
    }
};

typedef RBTreeMap<UVHelper, ULONG, true>  UVCountsT;



//-----------------------------------------------------------------------------
// Nested classes of OEOBJConverter


OEOBJConverter::ToRename::ToRename()
{}


OEOBJConverter::ToRename::ToRename(
    const FilePath& fromPath,
    const FilePath& toPath)
:
    mFromPath(fromPath),
    mToPath(toPath)
{}


struct OEOBJConverter::Texture
{
    BaseMaterial* mMaterial;
    OEUVWTexMapH  mTexMap;

    Texture()
    :   mMaterial(0)
    {}

    Texture(BaseMaterial* material,
            OEUVWTexMapH  texMap)
    :   mMaterial(material), mTexMap(texMap)
    {}

    Texture(const Texture& other)
    {
        *this = other;
    }

    Texture& operator=(const Texture& other)
    {
        mMaterial = other.mMaterial;
        mTexMap   = other.mTexMap;
        return *this;
    }
};


struct OEOBJConverter::HierarchyData
{
    bool    mVisible;
    Texture mTexture;

    HierarchyData(bool visible=true)
    :   mVisible(visible)
    {}
};

/*
struct OEOBJConverter::MaterialInfo
{
    String        mName;
    bool          mConverted;
    OEOBJMaterial mConvertedMaterial;

    MaterialInfo()
    :
        mConverted(false)
    {}
};
*/

struct OEOBJConverter::ImageInfo
{
    String   mFilename;
    LONG     mPlaybackMode;
    LONG     mTimingMode;
    LONG     mStartFrame;
    LONG     mEndFrame;
    Real     mFps;
    BaseTime mRangeFrom;
    BaseTime mRangeTo;
    LONG     mRangeLoops;

    ImageInfo()
    :
        mPlaybackMode(0),
        mTimingMode(0),
        mStartFrame(0),
        mEndFrame(0),
        mFps(0),
        mRangeLoops(0)
    {}

    bool operator<(
        const ImageInfo& other) const
    {
        if (mFilename     != other.mFilename)     { return (mFilename     < other.mFilename); }
        if (mPlaybackMode != other.mPlaybackMode) { return (mPlaybackMode < other.mPlaybackMode); }
        if (mTimingMode   != other.mTimingMode)   { return (mTimingMode   < other.mTimingMode); }
        if (mStartFrame   != other.mStartFrame)   { return (mStartFrame   < other.mStartFrame); }
        if (mEndFrame     != other.mEndFrame)     { return (mEndFrame     < other.mEndFrame); }
        if (mFps          != other.mFps)          { return (mFps          < other.mFps); }
        if (mRangeFrom    != other.mRangeFrom)    { return (mRangeFrom    < other.mRangeFrom); }
        if (mRangeTo      != other.mRangeTo)      { return (mRangeTo      < other.mRangeTo); }
        if (mRangeLoops   != other.mRangeLoops)   { return (mRangeLoops   < other.mRangeLoops); }
        return false;
    }
};


struct OEOBJConverter::Face
{
    ULONG mTextureIx;
    ULONG mVertexIxs[4];
    ULONG mUVWIxs[4];
    ULONG mNormalIxs[4];
};


struct OEOBJConverter::SimpleFace
{
    ULONG mVertexIxs[4];
    ULONG mNormalIxs[4];

    SimpleFace() {}

    SimpleFace(
        ULONG v1, ULONG v2, ULONG v3,
        ULONG n1, ULONG n2, ULONG n3)
    {
        mVertexIxs[0] = v1;
        mVertexIxs[1] = v2;
        mVertexIxs[2] = v3;
        mVertexIxs[3] = v3;
        mNormalIxs[0] = n1;
        mNormalIxs[1] = n2;
        mNormalIxs[2] = n3;
        mNormalIxs[3] = n3;
    }

    SimpleFace(
        ULONG v1, ULONG v2, ULONG v3, ULONG v4,
        ULONG n1, ULONG n2, ULONG n3, ULONG n4)
    {
        mVertexIxs[0] = v1;
        mVertexIxs[1] = v2;
        mVertexIxs[2] = v3;
        mVertexIxs[3] = v4;
        mNormalIxs[0] = n1;
        mNormalIxs[1] = n2;
        mNormalIxs[2] = n3;
        mNormalIxs[3] = n4;
    }
};



//-----------------------------------------------------------------------------
// Implementation of OEOBJConverter


OEOBJConverter::OEOBJConverter()
{}


OEOBJConverter::~OEOBJConverter()
{}


bool OEOBJConverter::convertScene(
    BaseDocument&   document,
    const Real      scale,
    const bool      useRelativePaths,
    const Filename& projectDirectory,
    const String&   projectName,
    const bool      collectImages,
    const bool      writeTempFiles,
    FilePath        &objFilePath,
    RenameListT     &renameList)
{
    char     buffer[256];
    DateTime time;
    bool     returnValue = false;

    // show hourglass as mouse pointer
    SetMousePointer(MOUSE_BUSY);

    // init internal data
    clearTemporaryData();
    mDocument        = &document;
    mExportScale     = scale;
    mView            = mDocument->GetActiveBaseDraw();
    mProjectDir      = projectDirectory;
    mProjectFilesDir = mProjectDir + (projectName + "_files");
    mCollectImages   = collectImages;
    
    // determine file paths for OBJ and MTL files
    objFilePath = mProjectFilesDir + projectName;
    FilePath mtlFilePath     = objFilePath;
    FilePath convObjFilePath = objFilePath;
    FilePath convMtlFilePath = objFilePath;
    objFilePath.setSuffix(".obj");
    mtlFilePath.setSuffix(".mtl");
    convObjFilePath.setSuffix(".obj");
    convMtlFilePath.setSuffix(".mtl");

    // for safety, always clear rename list
    renameList.erase();

    // if everything should be exported into temporary files:
    if (writeTempFiles)
    {
        // crate rename entry in rename list for the obj file and append a temp suffix
        ToRename rename;
        rename.mToPath   = convObjFilePath;
        convObjFilePath.setSuffix(String(".obj") + "__temp__");
        rename.mFromPath = convObjFilePath;
        renameList.pushBack(rename);
        // crate rename entry in rename list for the mtl file and append a temp suffix
        rename.mToPath   = convMtlFilePath;
        convMtlFilePath.setSuffix(String(".mtl") + "__temp__");
        rename.mFromPath = convMtlFilePath;
        renameList.pushBack(rename);
    }

    // if it doesn't exist yet, create sub-folder where all files for the
    // project will go
    if (!GeFExist(mProjectFilesDir.getFilename(), TRUE))
    {
        if (!GeFCreateDir(mProjectFilesDir.getFilename()))
        {
            ERRLOG("Failed to create directory for project files '" + mProjectFilesDir.getString() + "'");
            goto CLEANUP_AND_RETURN;
        }
    }

    // open main OBJ/MTL files
    if (!mOBJFile.open(convObjFilePath.getFilename(), FILEOPEN_WRITE, FILEDIALOG_ANY))
    {
        ERRLOG("Could not open OBJ file '" + convObjFilePath.getString() + "'");
        goto CLEANUP_AND_RETURN;
    }
    if (!mMTLFile.open(convMtlFilePath.getFilename(), FILEOPEN_WRITE, FILEDIALOG_ANY))
    {
        ERRLOG("Could not open MTL file '" + convMtlFilePath.getString() + "'");
        goto CLEANUP_AND_RETURN;
    }

    // write OBJ/MTL headers
    GetDateTimeNow(time);
    sprintf(buffer, "# Octane Exporter for CINEMA 4D\n# Exported on %d.%d.%d\n",
#if _C4D_VERSION < 120
            (int)time.lDay, (int)time.lMonth, (int)time.lYear);
#else
            (int)time.day, (int)time.month, (int)time.year);
#endif
    mOBJFile.write(buffer);
    mMTLFile.write(buffer);
    mOBJFile.write("\nmtllib " + mtlFilePath.getFileString() + "\n");

    // export OBJ scene
    collectLightObjects();
    if (!collectMaterials() ||
        !exportObjects() ||
        !exportMaterials())
    {
        goto CLEANUP_AND_RETURN;
    }

    // add collected rename information to rename list we return
    renameList.adoptBack(mRenameList);

    // everything was fine ...
    returnValue = true;

CLEANUP_AND_RETURN:

    clearTemporaryData();
    SetMousePointer(MOUSE_NORMAL);
    return returnValue;
}


void* OEOBJConverter::Alloc()
{
    return gNew HierarchyData();
}


void OEOBJConverter::Free(
    void *data)
{
    HierarchyData* hierarchyData = (HierarchyData*)data;
    gDelete(hierarchyData);
}


void OEOBJConverter::CopyTo(
    void *src,
    void *dst)
{
    if (src && dst && (src != dst))
    {
        ((HierarchyData*)dst)->mVisible = ((HierarchyData*)src)->mVisible;
        ((HierarchyData*)dst)->mTexture = ((HierarchyData*)src)->mTexture;
    }
}


Bool OEOBJConverter::Do(
    void         *data,
    BaseObject   *object,
    const Matrix &globalMatrix,
    Bool         controlObject)
{
    // for convenience cast void* to HierarchyData*
    HierarchyData& hierarchyData(*((HierarchyData*)data));

    // get render visibility and store it in HierarchyData, if it's explicit
    // (otherwise the implicit visibility from the parent object will be used)
    LONG mode = object->GetRenderMode();
    if (mode != MODE_UNDEF) { hierarchyData.mVisible = (mode == MODE_ON); }

    // skip generator objects and invisible objects
    bool exportable = true;
    if (controlObject || !hierarchyData.mVisible)
    {
        exportable = false;
    }

    // skip objects that belong to a layer that should not be rendered
    const LayerData* layerData = object->GetLayerData(mDocument);
    if (layerData && !layerData->render)
    {
        exportable = false;
    }

    // if the object is used by an area light, it will not be exported as normal
    // geometry, but as an area light -> no explicit export necessary
    if (mAreaLightObjects.get(object))
    {
        exportable = false;
    }

    // if object can be exported call corresponding export function
    bool texturesHandled = false;
    if (exportable)
    {
        if (object->GetType() == Opolygon)
        {
            if (!exportPolygonObject((PolygonObject&)*object,
                                     globalMatrix,
                                     hierarchyData))
            {
                return FALSE;
            }
            texturesHandled = true;
        }
        else if ((object->GetType() == Olight) && object->GetDeformMode())
        {
            if (!exportLightObject(*object, globalMatrix))
            {
                return FALSE;
            }
        }
    }

    // only if this object was exported as polygon object, the texture tags
    // have been handled. in all other cases if we still need to record its
    // material as it gets passed to its children.
    if (!texturesHandled)
    {
        // find last texture tag with valid material link that is not restricted to
        // a selection.
        BaseMaterial* material;
        BaseMaterial* lastMaterial = 0;
        TextureTag*   lastTextureTag = 0;
        for (BaseTag* tag=object->GetFirstTag(); tag; tag=tag->GetNext())
        {
            if (tag->IsInstanceOf(Ttexture))
            {
                material = (BaseMaterial*)getParameterLink(*tag, TEXTURETAG_MATERIAL, Mbase);
                if (material)
                {
                    String restriction = getParameterString(*tag, TEXTURETAG_RESTRICTION);
                    if (!getParameterString(*tag, TEXTURETAG_RESTRICTION).Content())
                    {
                        lastMaterial   = material;
                        lastTextureTag = (TextureTag*)tag;
                    }
                }
            }
        }
        // if we could find a texture tag, we store it in the hierarchy data so that
        // it can be used as default material by the children.
        if (lastMaterial && lastTextureTag)
        {
            OEUVWTexMapH texMap = gNew OEUVWTexMap(lastTextureTag, globalMatrix);
            if (!texMap) { return false; }
            hierarchyData.mTexture.mMaterial = lastMaterial;
            hierarchyData.mTexture.mTexMap   = texMap;
        }
    }

    return TRUE;
}



//-----------------------------------------------------------------------------
// Implementation of private member functions of class OEOBJConverter.


void OEOBJConverter::clearTemporaryData()
{
    mDocument       = 0;
    mProjectDir.clear();
    mProjectFilesDir.clear();
    mCollectImages  = false;
    mWriteTempFiles = false;
    mRenameList.erase();
    mOBJFile.close();
    mMTLFile.close();
    mExportScale    = 1.0;
    mView           = NULL;
    mAreaLightObjects.erase();
    mUsedMaterials.clear();
    mUsedLightMaterials.clear();
    mMaterialNames.erase();
    mMaterialNameCounts.erase();
    mObjectNameCounts.erase();
    mTextures.erase();
    mImageInfos.erase();
    mImageNameCounts.erase();
    mFaces.erase();
    mVertexIxOffset = 0;
    mUVWIxOffset    = 0;
    mNormalIxOffset = 0;
}


void OEOBJConverter::findUniqueName(
    const String &name,
    NameCountsT  &nameCountMap,
    String       &uniqueName)
{
    uniqueName = name;
    mangleName(uniqueName);
    ULONG* usageCount = nameCountMap.get(uniqueName);
    if (!usageCount)
    {
        nameCountMap.set(name, 1);
    }
    else
    {
        do
        {
            uniqueName = name + "_" + LongToString(++(*usageCount));
        }
        while (nameCountMap.get(uniqueName));
    }
}


void OEOBJConverter::collectLightObjects()
{
    // collect all light objects
    AutoAlloc<AtomArray> lights;
    collectAllObjects(*mDocument, Olight, lights);
    LONG lightCount = lights->GetCount();

    // loop over light objects and get the referenced objects for the object area lights
    for (LONG i=0; i<lightCount; ++i)
    {
        BaseObject* light = (BaseObject*)lights->GetIndex(i);
        if ((getParameterLong(*light, LIGHT_TYPE, -1) == LIGHT_TYPE_AREA) &&
            (getParameterLong(*light, LIGHT_AREADETAILS_SHAPE, -1) == LIGHT_AREADETAILS_SHAPE_OBJECT))
        {
            BaseObject* polyObject = (BaseObject*)getParameterLink(*light,
                                                                   LIGHT_AREADETAILS_OBJECT,
                                                                   Opolygon);
            if (polyObject) { mAreaLightObjects.add(polyObject, false); }
        }
    }
}


bool OEOBJConverter::collectMaterials()
{
    static const CHAR *sDefaultMaterialName = "__default";

    // first, add the default material under the NULL pointer. i.e. if we don't have a material,
    // use the default material
    mMaterialNameCounts.set(sDefaultMaterialName, 1);
    mMaterialNames.set(NULL, sDefaultMaterialName);

    // loop over all materials in scene, map each material to a unique name and create a map
    // [material -> unique material name]
    String matName, uniqueName;
    bool   renamedMaterials = false;
    for (BaseMaterial *mat = mDocument->GetFirstMaterial(); mat; mat = mat->GetNext())
    {
        // get name and find a unique name
        matName = mat->GetName();
        findUniqueName(mat->GetName(), mMaterialNameCounts, uniqueName);
        // store new material info
        mMaterialNames.set(mat, uniqueName);
        // if the unique name is different than the material name, rename material.
        // we do that to avoid different unique names if you do a new export with a different
        // order. basically the idea is to solve the material name uniqueness in C4D.
        if (matName != uniqueName)
        {
            mat->SetName(uniqueName);
            renamedMaterials = true;
        }
    }

    // send event and set dirty flag in document
    if (renamedMaterials)
    {
        mDocument->SetChanged();
        EventAdd();
    }
    return true;
}


bool OEOBJConverter::exportObjects()
{
    HierarchyData data;
    data.mVisible           = true;
    data.mTexture.mMaterial = 0;
    data.mTexture.mTexMap   = gNew OEUVWTexMap(0, Matrix());
    if (!data.mTexture.mTexMap) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportObjects(): Could not allocate initial texture mapping class."); }
#if _C4D_VERSION >= 120
    return Run(mDocument,
               false,
               1.0,
               false,
               BUILDFLAGS_EXTERNALRENDERER,
               &data,
               0);
#else
    return Run(mDocument,
               false,
               1.0,
               VFLAG_EXTERNALRENDERER | VFLAG_POLYGONAL,
               &data,
               0);
#endif
}


bool OEOBJConverter::exportMaterials()
{
    typedef StdMap<String, BaseMaterial*> SortedMatsT;

    // create a list of all used non-default materials sorted by name
    SortedMatsT sorted;
    bool        defaultHasBeenUsed = false;
    for (MaterialsT::const_iterator usedMatIt=mUsedMaterials.begin(), endIt=mUsedMaterials.end();
         usedMatIt != endIt;
         ++usedMatIt)
    {
        BaseMaterial *material = *usedMatIt;
        if (material)
        {
            const String *name = mMaterialNames.get(material);
            if (!name) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportMaterials(): Could not find name for used material."); }
            sorted[*name] = material;
        }
        else
        {
            defaultHasBeenUsed = true;
        }
    }

    // export these materials
    bool success = true;
    for (SortedMatsT::const_iterator sortedIt=sorted.begin(), endIt=sorted.end();
         sortedIt != endIt;
         ++sortedIt)
    {
        success &= exportMaterial(sortedIt->second);
    }

    // export the light materials separately
    for (LightMaterialsT::const_iterator lightIt = mUsedLightMaterials.begin(),
                                         endIt = mUsedLightMaterials.end();
         lightIt != endIt;
         ++lightIt)
    {
        if (!lightIt->second.write(mMTLFile, lightIt->first))
        {
            return false;
        }
    }

    // if the default material was used, export it, too
    if (defaultHasBeenUsed)
    {
        success &= exportMaterial(NULL);
    }
    
    return success;
}


bool OEOBJConverter::exportLightObject(BaseObject&   object,
                                       const Matrix& globalMatrix)
{
    // get light type and export geometry using corresponding function
    switch (getParameterLong(object, LIGHT_TYPE, -1))
    {
        case LIGHT_TYPE_AREA:
            switch (getParameterLong(object, LIGHT_AREADETAILS_SHAPE))
            {
                case LIGHT_AREADETAILS_SHAPE_DISC:
                    return exportDiscAreaLight(object, globalMatrix);
                case LIGHT_AREADETAILS_SHAPE_RECTANGLE:
                    return exportRectangleAreaLight(object, globalMatrix);
                case LIGHT_AREADETAILS_SHAPE_SPHERE:
                    return exportSphereAreaLight(object, globalMatrix);
                case LIGHT_AREADETAILS_SHAPE_CYLINDER:
                    return exportCylinderAreaLight(object, globalMatrix, false);
#if _C4D_VERSION >= 120
                case LIGHT_AREADETAILS_SHAPE_PCYLINDER:
                    return exportCylinderAreaLight(object, globalMatrix, true);
#endif
                case LIGHT_AREADETAILS_SHAPE_CUBE:
                    return exportCubeAreaLight(object, globalMatrix);
                case LIGHT_AREADETAILS_SHAPE_HEMISPHERE:
                    return exportHemisphereAreaLight(object, globalMatrix);
                case LIGHT_AREADETAILS_SHAPE_OBJECT:
                    return exportObjectAreaLight(object);
                default:
                    break;
            }
            break;
        default:
            break;
    }

    return true;
}


bool OEOBJConverter::exportDiscAreaLight(BaseObject&   object,
                                         const Matrix& globalMatrix)
{
    static const ULONG cMaxSegments = 32;

    // create vertices: the first one as centre and the other on a circle around
    // the center
    VerticesT vertices(cMaxSegments + 1);
    vertices[0] = Vector(0.0);
    Real step = pi2 / (Real)cMaxSegments;
    for (ULONG i=0; i<cMaxSegments; ++i)
    {
        Real angle((Real)i * step);
        vertices[i+1] = Vector(cos(angle), sin(angle), 0.0);
    }

    // there is only one normal for all faces
    NormalsT normals(1);
    normals[0] = SVector(0.0, 0.0, 1.0);

    // create faces - the last one, connects the last vertex with the second
    // vertex
    SimpleFacesT faces(cMaxSegments);
    for (ULONG i=0; i+1<cMaxSegments; ++i)
    {
        faces[i] = SimpleFace(i+1, i+2, 0, 0, 0, 0);
    }
    faces[cMaxSegments-1] = SimpleFace(cMaxSegments, 1, 0, 0, 0, 0);

    // export geometry
    return exportLightGeometry(object, globalMatrix, vertices, normals, faces);
}


bool OEOBJConverter::exportRectangleAreaLight(BaseObject&   object,
                                              const Matrix& globalMatrix)
{
    VerticesT vertices(4);
    vertices[0] = Vector(-1.0,  1.0, 0.0);
    vertices[1] = Vector(-1.0, -1.0, 0.0);
    vertices[2] = Vector( 1.0, -1.0, 0.0);
    vertices[3] = Vector( 1.0,  1.0, 0.0);

    NormalsT normals(1);
    normals[0] = SVector(0.0, 0.0, 1.0);

    SimpleFacesT faces(1);
    faces[0] = SimpleFace(0, 1, 2, 3, 0, 0, 0, 0);

    return exportLightGeometry(object, globalMatrix, vertices, normals, faces);
}


bool OEOBJConverter::exportSphereAreaLight(BaseObject&   object,
                                           const Matrix& globalMatrix)
{
    /*  static const ULONG cMaxSubDivision = 0;

    VerticesT vertices(12 + cMaxSubDivision * (20 + 10*cMaxSubDivision));
    Real yRing = 0.2 * sqrt(5.0);
    Real rRing = 0.4 * sqrt(5.0);
    vertices[0]  = Vector(0.0,                    1.0,   0.0);
    vertices[1]  = Vector(0.0,                    yRing, rRing);
    vertices[2]  = Vector(sin(Rad(72.0)) *rRing,  yRing, cos(Rad(72.0)) *rRing);
    vertices[3]  = Vector(sin(Rad(144.0))*rRing,  yRing, cos(Rad(144.0))*rRing);
    vertices[4]  = Vector(sin(Rad(216.0))*rRing,  yRing, cos(Rad(216.0))*rRing);
    vertices[5]  = Vector(sin(Rad(288.0))*rRing,  yRing, cos(Rad(288.0))*rRing);
    vertices[6]  = Vector(sin(Rad(36.0)) *rRing, -yRing, cos(Rad(36.0)) *rRing);
    vertices[7]  = Vector(sin(Rad(108.0))*rRing, -yRing, cos(Rad(108.0))*rRing);
    vertices[8]  = Vector(sin(Rad(180.0))*rRing, -yRing, cos(Rad(180.0))*rRing);
    vertices[9]  = Vector(sin(Rad(252.0))*rRing, -yRing, cos(Rad(252.0))*rRing);
    vertices[10] = Vector(sin(Rad(324.0))*rRing, -yRing, cos(Rad(324.0))*rRing);
    vertices[11] = Vector(0.0,                   -1.0,   0.0);
    FacesT faces(20 * cMaxSubDivision * cMaxSubDivision);*/

    static const ULONG cMaxSegments = 16;

    // calculate the number of latitude/longitude bands
    ULONG thetaSteps = cMaxSegments;
    ULONG phiSteps = cMaxSegments * 2;

    // initialise vertex array - the first two vertices are at the poles
    VerticesT vertices(2 + (thetaSteps-1) * phiSteps);
    vertices[0] = Vector(0.0,  1.0, 0.0);
    vertices[1] = Vector(0.0, -1.0, 0.0);
    ULONG nextVertexIx = 2;

    // initialise normal array - all normals have the same vector as the vertices
    NormalsT normals(vertices.size());
    normals[0] = toSV(vertices[0]);
    normals[1] = toSV(vertices[1]);

    // initialise faces
    SimpleFacesT faces(thetaSteps * phiSteps);
    ULONG nextFaceIx = 0;

    // loop over residual vertices and create faces along the way
    Real theta, phi, r, y;
    ULONG offset, a, b, c, d;

    // North pole band:
    theta = pi * 1 / thetaSteps;
    y = cos(theta);
    r = sin(theta);
    for (ULONG phiStep=0; phiStep<phiSteps; ++phiStep)
    {
        // calculate vertex coordinate (and with that the normal)
        phi = pi2 * (phiStep+1) / phiSteps;
        vertices[nextVertexIx] = Vector(r * sin(phi), y, r * cos(phi));
        normals[nextVertexIx]  = toSV(vertices[nextVertexIx]);
        ++nextVertexIx;
        // create new polygon (in the same order we create the vertices)
        a = 2 + phiStep;
        b = 2 + (phiStep+1) % phiSteps;
        faces[nextFaceIx] = SimpleFace(a, b, 0, a, b, 0);
        ++nextFaceIx;
    }

    // middle latitude bands:
    for (ULONG thetaStep=1; thetaStep<thetaSteps-1; ++thetaStep)
    {
        // calculate y coordinate of circle where the vertices lie on
        theta = pi * (thetaStep+1) / thetaSteps;
        y = cos(theta);
        r = sin(theta);
        offset = 2 + thetaStep*phiSteps;
        // for each longitude band:
        for (ULONG phiStep=0; phiStep<phiSteps; ++phiStep)
        {
            // calculate vertex coordinate (and with that the normal)
            phi = pi2 * (phiStep+1) / phiSteps;
            vertices[nextVertexIx] = Vector(r * sin(phi), y, r * cos(phi));
            normals[nextVertexIx]  = toSV(vertices[nextVertexIx]);
            ++nextVertexIx;
            // create new polygon (in the same order we create the vertices)
            a = offset + phiStep;
            b = offset + (1+phiStep) % phiSteps;
            c = b - phiSteps;
            d = a - phiSteps;
            faces[nextFaceIx] = SimpleFace(a, b, c, d, a, b, c, d);
            ++nextFaceIx;
        }
    }

    // South pole band:
    offset = 2 + (thetaSteps-2)*phiSteps;
    for (ULONG phiStep=0; phiStep<phiSteps; ++phiStep)
    {
        a = offset + phiStep;
        b = offset + (1+phiStep) % phiSteps;
        faces[nextFaceIx] = SimpleFace(1, b, a, 1, b, a);
        ++nextFaceIx;
    }

    return exportLightGeometry(object, globalMatrix, vertices, normals, faces);
}


bool OEOBJConverter::exportCylinderAreaLight(BaseObject&   object,
                                             const Matrix& globalMatrix,
                                             bool          perpendicular)
{
    static const ULONG cMaxSegments = 32;

    // create vertices and normals: the first one as centre and the other on a circle around
    // the center
    VerticesT vertices(2*cMaxSegments + 2);
    NormalsT normals(cMaxSegments + 2);
    Real step = pi2 / (Real)cMaxSegments;
    if (!perpendicular)
    {
        vertices[0] = Vector(0.0, 0.0, -1.0);
        vertices[1] = Vector(0.0, 0.0,  1.0);
        normals[0] = SVector(0.0, 0.0, -1.0);
        normals[1] = SVector(0.0, 0.0,  1.0);
        for (ULONG i=0; i<cMaxSegments; ++i)
        {
            Real angle = (Real)i * step;
            Real cosAngle = cos(angle);
            Real sinAngle = sin(angle);
            vertices[2+i*2] = Vector(cosAngle, sinAngle, -1);
            vertices[3+i*2] = Vector(cosAngle, sinAngle,  1);
            normals[2+i] = SVector((SReal)cosAngle, (SReal)sinAngle, 0);
        }
    }
    else
    {
        vertices[0] = Vector(-1.0, 0.0, 0.0);
        vertices[1] = Vector( 1.0, 0.0, 0.0);
        normals[0] = SVector(-1.0, 0.0, 0.0);
        normals[1] = SVector( 1.0, 0.0, 0.0);
        for (ULONG i=0; i<cMaxSegments; ++i)
        {
            Real angle = (Real)i * step;
            Real cosAngle = cos(angle);
            Real sinAngle = sin(angle);
            vertices[2+i*2] = Vector(-1, cosAngle, sinAngle);
            vertices[3+i*2] = Vector( 1, cosAngle, sinAngle);
            normals[2+i] = SVector(0, (SReal)cosAngle, (SReal)sinAngle);
        }
    }

    // create faces, one cylinder segment consists of 2 triangles and 1 quad:
    //
    //              n
    //         *---------*
    //        /| b     c |\
    //       / |         | \
    //    0 *  |         |  * 1
    //       \ |         | /
    //        \| a     d |/
    //         *---------*
    //              m
    //
    // m, n are the normals of the lower and upper edge
    SimpleFacesT faces(cMaxSegments * 3);
    for (ULONG i=0; i<cMaxSegments; ++i)
    {
        ULONG a, b, c, d, m, n;
        a = 2 + 2*i;
        d = 3 + 2*i;
        m = 2 + i;
        if (i+1 < cMaxSegments)
        {
            b = 4 + 2*i;
            c = 5 + 2*i;
            n = 3 + i;
        }
        else
        {
            b = 2;
            c = 3;
            n = 2;
        }
        faces[i*3]   = SimpleFace(0, b, a, 0, 0, 0);
        faces[i*3+2] = SimpleFace(1, d, c, 1, 1, 1);
        faces[i*3+1] = SimpleFace(a, b, c, d, m, n, n, m);
    }

    // export geometry
    return exportLightGeometry(object, globalMatrix, vertices, normals, faces);
}


bool OEOBJConverter::exportCubeAreaLight(BaseObject&   object,
                                         const Matrix& globalMatrix)
{
    VerticesT vertices(8);
    vertices[0] = Vector(-1.0,  1.0,  1.0);
    vertices[1] = Vector(-1.0, -1.0,  1.0);
    vertices[2] = Vector( 1.0, -1.0,  1.0);
    vertices[3] = Vector( 1.0,  1.0,  1.0);
    vertices[4] = Vector( 1.0,  1.0, -1.0);
    vertices[5] = Vector( 1.0, -1.0, -1.0);
    vertices[6] = Vector(-1.0, -1.0, -1.0);
    vertices[7] = Vector(-1.0,  1.0, -1.0);

    NormalsT normals(6);
    normals[0] = SVector( 0.0,  0.0,  1.0);
    normals[1] = SVector( 0.0,  0.0, -1.0);
    normals[2] = SVector( 0.0,  1.0,  0.0);
    normals[3] = SVector( 0.0, -1.0,  0.0);
    normals[4] = SVector( 1.0,  0.0,  0.0);
    normals[5] = SVector(-1.0,  0.0,  0.0);

    SimpleFacesT faces(6);
    faces[0] = SimpleFace(0, 1, 2, 3, 0, 0, 0, 0);
    faces[1] = SimpleFace(4, 5, 6, 7, 1, 1, 1, 1);
    faces[2] = SimpleFace(0, 3, 4, 7, 2, 2, 2, 2);
    faces[3] = SimpleFace(1, 6, 5, 2, 3, 3, 3, 3);
    faces[4] = SimpleFace(2, 5, 4, 3, 4, 4, 4, 4);
    faces[5] = SimpleFace(0, 7, 6, 1, 5, 5, 5, 5);

    return exportLightGeometry(object, globalMatrix, vertices, normals, faces);
}


bool OEOBJConverter::exportHemisphereAreaLight(BaseObject&   object,
                                               const Matrix& globalMatrix)
{
    static const ULONG cMaxSegments = 8;

    // calculate the number of latitude/longitude bands
    ULONG thetaSteps = cMaxSegments;
    ULONG phiSteps = cMaxSegments * 4;

    // initialise vertex array - the first vertex is at the pole
    VerticesT vertices(1 + thetaSteps * phiSteps);
    vertices[0] = Vector(0.0, 1.0, 0.0);
    ULONG nextVertexIx = 1;

    // initialise normal array - all normals have the same vector as the vertices
    NormalsT normals(vertices.size());
    normals[0] = toSV(vertices[0]);

    // initialise faces
    SimpleFacesT faces(thetaSteps * phiSteps);
    ULONG nextFaceIx = 0;

    // loop over residual vertices and create faces along the way
    Real theta, phi, r, y;
    ULONG offset, a, b, c, d;

    // North pole band:
    theta = pi05 * 1 / thetaSteps;
    y = cos(theta);
    r = sin(theta);
    for (ULONG phiStep=0; phiStep<phiSteps; ++phiStep)
    {
        // calculate vertex coordinate (and with that the normal)
        phi = pi2 * (phiStep+1) / phiSteps;
        vertices[nextVertexIx] = Vector(r * sin(phi), y, r * cos(phi));
        normals[nextVertexIx]  = toSV(vertices[nextVertexIx]);
        ++nextVertexIx;
        // create new polygon (in the same order we create the vertices)
        a = 1 + phiStep;
        b = 1 + (phiStep+1) % phiSteps;
        faces[nextFaceIx] = SimpleFace(a, b, 0, a, b, 0);
        ++nextFaceIx;
    }

    // for all latitude bands (theta 0 at the pole):
    for (ULONG thetaStep=1; thetaStep<thetaSteps; ++thetaStep)
    {
        // calculate y coordinate of circle where the vertices lie on
        theta = pi05 * (thetaStep+1) / thetaSteps;
        if (thetaStep+1 < thetaSteps)
        {
            y = cos(theta);
            r = sin(theta);
        }
        else
        {
            y = 0.0;
            r = 1.0;
        }
        offset = 1 + thetaStep*phiSteps;
        // for each longitude band:
        for (ULONG phiStep=0; phiStep<phiSteps; ++phiStep)
        {
            // calculate vertex coordinate (and with that the normal)
            phi = pi2 * (phiStep+1) / phiSteps;
            vertices[nextVertexIx] = Vector(r * sin(phi), y, r * cos(phi));
            normals[nextVertexIx]  = toSV(vertices[nextVertexIx]);
            ++nextVertexIx;
            // create new polygon (in the same order we create the vertices)
            a = offset + phiStep;
            b = offset + (phiStep+1) % phiSteps;
            c = b - phiSteps;
            d = a - phiSteps;
            faces[nextFaceIx] = SimpleFace(a, b, c, d, a, b, c, d);
            ++nextFaceIx;
        }
    }

    return exportLightGeometry(object, globalMatrix, vertices, normals, faces);
}


bool OEOBJConverter::exportObjectAreaLight(BaseObject& object)
{
    // get object from link and return without error if there isn't one
    PolygonObject* polyObject = (PolygonObject*)getParameterLink(object,
                                                                 LIGHT_AREADETAILS_OBJECT,
                                                                 Opolygon);
    if (!polyObject) { return true; }
    Matrix globalMatrix = polyObject->GetMg();

    // get polygons and vertices
    const CPolygon* polygons = getPolygons(*polyObject);
    ULONG polygonCount       = polyObject->GetPolygonCount();
    const Vector* vertices   = getPoints(*polyObject);
    ULONG vertexCount        = polyObject->GetPointCount();

    // return without an error if object is empty
    if (!polygons || !polygonCount || !vertices || !vertexCount )
    {
        return true;
    }

    // helper variables
    char lineBuffer[1024];

    // write comment with object name
    mOBJFile.write("\n# " + object.GetName() + "\n");

    // write vertices to file
    for (ULONG i=0; i<vertexCount; ++i)
    {
        Vector v(globalMatrix * vertices[i]);
        v *= mExportScale;
        sprintf(lineBuffer, "v %.8g %.8g %.8g\n", v.x, v.y, -v.z);
        mOBJFile.write(lineBuffer);
    }

    // write vertex IDs into face structures
    SimpleFacesT faces(polygonCount);
    if (faces.size() != polygonCount)
    {
        ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportObjectAreaLight(): Could not allocate enough memory for face cache.");
    }
    for (ULONG i=0; i<polygonCount; ++i)
    {
        const CPolygon &p(polygons[i]);
        SimpleFace& f(faces[i]);
        f.mVertexIxs[0] = p.a + 1;
        f.mVertexIxs[1] = p.b + 1;
        f.mVertexIxs[2] = p.c + 1;
        f.mVertexIxs[3] = (p.c == p.d) ? 0 : p.d + 1;
    }

    // export normals
    ULONG normalCount = 0;
    SVector* c4dNormals = polyObject->CreatePhongNormals();
    // if phong normals are available
    if (c4dNormals)
    {
        // loop over all faces and determine the set of unique normals. while doing
        // that we export the unique normals and store their indices in the faces.
        NormalCountsT normalCounts;
        ULONG         faceVertexCount, c, *normalIx;
        SVector       n, tn;
        SMatrix       invMatrix = toSM(!globalMatrix);
        for (ULONG i=0; i<polygonCount; ++i)
        {
            SimpleFace& f(faces[i]);
            faceVertexCount = f.mVertexIxs[3] ? 4 : 3;
            for (c=0; c<faceVertexCount; ++c)
            {
                n = toSV(c4dNormals[i*4 + c]);
                tn.x = invMatrix.v1.x*n.x + invMatrix.v1.y*n.y + invMatrix.v1.z*n.z;
                tn.y = invMatrix.v2.x*n.x + invMatrix.v2.y*n.y + invMatrix.v2.z*n.z;
                tn.z = invMatrix.v3.x*n.x + invMatrix.v3.y*n.y + invMatrix.v3.z*n.z;
                normalize(tn);
                normalIx = normalCounts.set(tn, 0, false);
                if (!normalIx) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Ran out of memory."); }
                if (!*normalIx)
                {
                    sprintf(lineBuffer, "vn %.8g %.8g %.8g\n", tn.x, tn.y, -tn.z);
                    mOBJFile.write(lineBuffer);
                    *normalIx = (ULONG)normalCounts.size();
                }
                f.mNormalIxs[c] = *normalIx;
            }
        }
        normalCount = (ULONG)normalCounts.size();
        bDelete(c4dNormals);
    }
    // if no phong normals available:
    else
    {
        // loop over all faces and determine the set of unique normals. while doing
        // that we export the unique normals and store their indices in the faces.
        NormalCountsT normalCounts;
        ULONG         *normalIx;
        SVector       n;
        for (ULONG i=0; i<polygonCount; ++i)
        {
            SimpleFace& f(faces[i]);
            if (f.mVertexIxs[3])
            {
                n = toSV(calculateNormal(vertices[f.mVertexIxs[0]-1],
                                         vertices[f.mVertexIxs[1]-1],
                                         vertices[f.mVertexIxs[2]-1],
                                         vertices[f.mVertexIxs[3]-1]));
            }
            else
            {
                n = toSV(calculateNormal(vertices[f.mVertexIxs[0]-1],
                                         vertices[f.mVertexIxs[1]-1],
                                         vertices[f.mVertexIxs[2]-1]));
            }
            normalIx = normalCounts.set(n, 0, false);
            if (!normalIx) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Ran out of memory."); }
            if (!*normalIx)
            {
                sprintf(lineBuffer, "vn %.8g %.8g %.8g\n", n.x, n.y, -n.z);
                mOBJFile.write(lineBuffer);
                *normalIx = (ULONG)normalCounts.size();
            }
            f.mNormalIxs[0] = *normalIx;
            f.mNormalIxs[1] = *normalIx;
            f.mNormalIxs[2] = *normalIx;
            if (f.mVertexIxs[3])
            {
                f.mNormalIxs[3] = *normalIx;
            }
        }
        normalCount = (ULONG)normalCounts.size();
    }

    // find unique object name
    String uniqueObjectName;
    findUniqueName(object.GetName(), mObjectNameCounts, uniqueObjectName);
    String materialName;
    findUniqueName(uniqueObjectName + "_mat", mMaterialNameCounts, materialName);

    // create material and store it in the light material map
    OEOBJMaterial material;
    material.mKd.mEnabled = true;
    material.mKd.mColor   = getParameterVector(object, LIGHT_COLOR, Vector(1.0));
    material.mKd.mColor.x = pow(material.mKd.mColor.x, cDefaultGamma);
    material.mKd.mColor.y = pow(material.mKd.mColor.y, cDefaultGamma);
    material.mKd.mColor.z = pow(material.mKd.mColor.z, cDefaultGamma);
    mUsedLightMaterials[materialName] = material;

    // start new group with just exported material
    mOBJFile.write("g " + uniqueObjectName + "\n");
    mOBJFile.write("usemtl " + materialName + "\n");
    mOBJFile.write("s 1\n");

    // export polygons
    for (ULONG i=0; i<polygonCount; ++i)
    {
        const SimpleFace& f(faces[i]);
        // write quad
        if (f.mVertexIxs[3])
        {
            sprintf(lineBuffer,
                    "f %u//%u %u//%u %u//%u %u//%u\n",
                    (unsigned int)(f.mVertexIxs[3] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[3] + mNormalIxOffset),
                    (unsigned int)(f.mVertexIxs[2] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[2] + mNormalIxOffset),
                    (unsigned int)(f.mVertexIxs[1] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[1] + mNormalIxOffset),
                    (unsigned int)(f.mVertexIxs[0] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[0] + mNormalIxOffset));
        }
        // write tri
        else
        {
            sprintf(lineBuffer,
                    "f %u//%u %u//%u %u//%u\n",
                    (unsigned int)(f.mVertexIxs[2] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[2] + mNormalIxOffset),
                    (unsigned int)(f.mVertexIxs[1] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[1] + mNormalIxOffset),
                    (unsigned int)(f.mVertexIxs[0] + mVertexIxOffset),
                    (unsigned int)(f.mNormalIxs[0] + mNormalIxOffset));
        }
        mOBJFile.write(lineBuffer);
    }

    // update offsets
    mVertexIxOffset += vertexCount;
    mNormalIxOffset += normalCount;

    return true;
}


bool OEOBJConverter::exportLightGeometry(BaseObject&         object,
                                         const Matrix&       globalMatrix,
                                         const VerticesT&    vertices,
                                         const NormalsT&     normals,
                                         const SimpleFacesT& faces)
{
    // write comment with object name
    mOBJFile.write("\n# " + object.GetName() + "\n");

    // get X, Y and Z size
    Vector scale(getParameterReal(object, LIGHT_AREADETAILS_SIZEX) / 2.0,
                 getParameterReal(object, LIGHT_AREADETAILS_SIZEY) / 2.0,
                 getParameterReal(object, LIGHT_AREADETAILS_SIZEZ) / 2.0);
    Matrix scaledMatrix(globalMatrix.off,
                        globalMatrix.v1 * scale.x,
                        globalMatrix.v2 * scale.y,
                        globalMatrix.v3 * scale.z);

    // helper variables
    char lineBuffer[1024];

    // write vertices to file
    for (ULONG i=0; i<vertices.size(); ++i)
    {
        Vector v(scaledMatrix * vertices[i]);
        v *= mExportScale;
        sprintf(lineBuffer, "v %.8g %.8g %.8g\n", v.x, v.y, -v.z);
        mOBJFile.write(lineBuffer);
    }

    // export normals
    SMatrix invMatrix = toSM(!scaledMatrix);
    SVector n, tn;
    for (ULONG i=0; i<normals.size(); ++i)
    {
        n = normals[i];
        tn.x = invMatrix.v1.x*n.x + invMatrix.v1.y*n.y + invMatrix.v1.z*n.z;
        tn.y = invMatrix.v2.x*n.x + invMatrix.v2.y*n.y + invMatrix.v2.z*n.z;
        tn.z = invMatrix.v3.x*n.x + invMatrix.v3.y*n.y + invMatrix.v3.z*n.z;
        normalize(tn);
        sprintf(lineBuffer, "vn %.8g %.8g %.8g\n", tn.x, tn.y, -tn.z);
        mOBJFile.write(lineBuffer);
    }

    // find unique object and material name
    String uniqueObjectName;
    findUniqueName(object.GetName(), mObjectNameCounts, uniqueObjectName);
    String materialName;
    findUniqueName(uniqueObjectName + "_mat", mMaterialNameCounts, materialName);

    // create material and store it in the light material map
    OEOBJMaterial material;
    material.mKd.mEnabled = true;
    material.mKd.mColor   = getParameterVector(object, LIGHT_COLOR, Vector(1.0));
    material.mKd.mColor.x = pow(material.mKd.mColor.x, cDefaultGamma);
    material.mKd.mColor.y = pow(material.mKd.mColor.y, cDefaultGamma);
    material.mKd.mColor.z = pow(material.mKd.mColor.z, cDefaultGamma);
    mUsedLightMaterials[materialName] = material;

    // start new group with just exported material
    mOBJFile.write("g " + uniqueObjectName + "\n");
    mOBJFile.write("usemtl " + materialName + "\n");
    mOBJFile.write("s 1\n");

    // export polygons
    for (ULONG i=0; i<faces.size(); ++i)
    {
        const SimpleFace& face(faces[i]);
        // write quad
        if (face.mVertexIxs[3] != face.mVertexIxs[2])
        {
            sprintf(lineBuffer,
                    "f %u//%u %u//%u %u//%u %u//%u\n",
                    (unsigned int)(face.mVertexIxs[3] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[3] + 1 + mNormalIxOffset),
                    (unsigned int)(face.mVertexIxs[2] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[2] + 1 + mNormalIxOffset),
                    (unsigned int)(face.mVertexIxs[1] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[1] + 1 + mNormalIxOffset),
                    (unsigned int)(face.mVertexIxs[0] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[0] + 1 + mNormalIxOffset));
        }
        // write tri
        else
        {
            sprintf(lineBuffer,
                    "f %u//%u %u//%u %u//%u\n",
                    (unsigned int)(face.mVertexIxs[2] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[2] + 1 + mNormalIxOffset),
                    (unsigned int)(face.mVertexIxs[1] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[1] + 1 + mNormalIxOffset),
                    (unsigned int)(face.mVertexIxs[0] + 1 + mVertexIxOffset),
                    (unsigned int)(face.mNormalIxs[0] + 1 + mNormalIxOffset));
        }
        mOBJFile.write(lineBuffer);
    }

    // update offsets
    mVertexIxOffset += (ULONG)vertices.size();
    mNormalIxOffset += (ULONG)normals.size();

    return true;
}


bool OEOBJConverter::exportPolygonObject(PolygonObject& object,
                                         const Matrix&  globalMatrix,
                                         HierarchyData& hierarchyData)
{
    // get polygons and vertices
    const CPolygon* polygons = getPolygons(object);
    ULONG polygonCount       = object.GetPolygonCount();
    const Vector* vertices   = getPoints(object);
    ULONG vertexCount        = object.GetPointCount();

    // return without an error if object is empty, but setup materials to determine
    // the material that should be used by children
    if (!polygons || !polygonCount || !vertices || !vertexCount)
    {
        return setupTextures(object, globalMatrix, hierarchyData);
    }

    // helper variables
    char   lineBuffer[1024];
    Vector globalVector;

    // initialise face array
    mFaces.clear();
    if (!mFaces.reserve(polygonCount))
    {
        ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Could not allocate enough memory for face cache.");
    }

    // write comment with object name
    mOBJFile.write("\n# " + object.GetName() + "\n");

    // write vertices to file
    for (const Vector* v=vertices, *end=v+vertexCount; v!=end; ++v)
    {
        globalVector = globalMatrix * *v;
        globalVector *= mExportScale;
        sprintf(lineBuffer,
                "v %.8g %.8g %.8g\n",
                globalVector.x,
                globalVector.y,
                -globalVector.z);
        mOBJFile.write(lineBuffer);
    }

    // reserve space in face cache and write vertex IDs into face structures
    for (const CPolygon* p=polygons, *end=p+polygonCount; p!=end; ++p)
    {
        mFaces.append();
        Face& face(mFaces.back());
        face.mVertexIxs[0] = p->a + 1;
        face.mVertexIxs[1] = p->b + 1;
        face.mVertexIxs[2] = p->c + 1;
        face.mVertexIxs[3] = (p->c == p->d) ? 0 : p->d + 1;
    }

    // export normals
    ULONG normalCount = 0;
    SVector* c4dNormals = object.CreatePhongNormals();
    // if phon normals are available
    if (c4dNormals)
    {
        // loop over all faces and determine the set of unique normals. while doing
        // that we export the unique normals and store their indices in the faces.
        NormalCountsT normalCounts;
        ULONG         faceVertexCount, c, *normalIx;
        SVector       n, tn;
        SMatrix       invMatrix = toSM(!globalMatrix);
        for (ULONG i=0; i<polygonCount; ++i)
        {
            Face& f(mFaces[i]);
            faceVertexCount = f.mVertexIxs[3] ? 4 : 3;
            for (c=0; c<faceVertexCount; ++c)
            {
                n = toSV(c4dNormals[i*4 + c]);
                tn.x = invMatrix.v1.x*n.x + invMatrix.v1.y*n.y + invMatrix.v1.z*n.z;
                tn.y = invMatrix.v2.x*n.x + invMatrix.v2.y*n.y + invMatrix.v2.z*n.z;
                tn.z = invMatrix.v3.x*n.x + invMatrix.v3.y*n.y + invMatrix.v3.z*n.z;
                normalize(tn);
                normalIx = normalCounts.set(tn, 0, false);
                if (!normalIx) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Ran out of memory."); }
                if (!*normalIx)
                {
                    sprintf(lineBuffer, "vn %.8g %.8g %.8g\n", tn.x, tn.y, -tn.z);
                    mOBJFile.write(lineBuffer);
                    *normalIx = (ULONG)normalCounts.size();
                }
                f.mNormalIxs[c] = *normalIx;
            }
        }
        normalCount = (ULONG)normalCounts.size();
        bDelete(c4dNormals);
    }
    // if no phong normals available:
    else
    {
        // loop over all faces and determine the set of unique normals. while doing
        // that we export the unique normals and store their indices in the faces.
        NormalCountsT normalCounts;
        ULONG         *normalIx;
        SVector       normal;
        for (ULONG i=0; i<polygonCount; ++i)
        {
            Face& f(mFaces[i]);
            if (f.mVertexIxs[3])
            {
                normal = toSV(calculateNormal(vertices[f.mVertexIxs[0]-1],
                                              vertices[f.mVertexIxs[1]-1],
                                              vertices[f.mVertexIxs[2]-1],
                                              vertices[f.mVertexIxs[3]-1]));
            }
            else
            {
                normal = toSV(calculateNormal(vertices[f.mVertexIxs[0]-1],
                                              vertices[f.mVertexIxs[1]-1],
                                              vertices[f.mVertexIxs[2]-1]));
            }
            normalIx = normalCounts.set(normal, 0, false);
            if (!normalIx) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Ran out of memory."); }
            if (!*normalIx)
            {
                sprintf(lineBuffer,
                        "vn %.8g %.8g %.8g\n",
                        normal.x,
                        normal.y,
                        -normal.z);
                mOBJFile.write(lineBuffer);
                *normalIx = (ULONG)normalCounts.size();
            }
            f.mNormalIxs[0] = *normalIx;
            f.mNormalIxs[1] = *normalIx;
            f.mNormalIxs[2] = *normalIx;
            if (f.mVertexIxs[3])
            {
                f.mNormalIxs[3] = *normalIx;
            }
        }
        normalCount = (ULONG)normalCounts.size();
    }

    // setup materials for object (this will also export used materials and update
    // hierarchy data)
    if (!setupTextures(object, globalMatrix, hierarchyData)) { return false; }

    // loop over all faces and determine the set of unique UVWs. while doing
    // that we export the unique UVWs and store their indices in the faces.
    ULONG uvwCount = 0;
    {
        UVCountsT uvCounts;
        Vector    uvws[4];
        ULONG     faceVertexCount, v, *uvIx;
        for (ULONG i=0; i<polygonCount; ++i)
        {
            Face& f(mFaces[i]);
            mTextures[f.mTextureIx].mTexMap->getUVWs(i, uvws);
            faceVertexCount = f.mVertexIxs[3] ? 4 : 3;
            for (v=0; v<faceVertexCount; ++v)
            {
                uvws[v].y = 1.0 - uvws[v].y;
                uvIx = uvCounts.set(uvws[v], 0, false);
                if (!uvIx) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Ran out of memory."); }
                if (!*uvIx)
                {
                    sprintf(lineBuffer, "vt %.8g %.8g\n", uvws[v].x, uvws[v].y);
                    mOBJFile.write(lineBuffer);
                    *uvIx = (ULONG)uvCounts.size();
                }
                f.mUVWIxs[v] = *uvIx;
            }
        }
        uvwCount = (ULONG)uvCounts.size();
    }

    // find unique object name
    String uniqueObjectName;
    findUniqueName(object.GetName(), mObjectNameCounts, uniqueObjectName);

    // now loop over each texture and export their polygon set
    for (SizeT t=0; t<mTextures.size(); ++t)
    {
        // get name we should use for this material
        const String *matName = mMaterialNames.get(mTextures[t].mMaterial);
        if (!matName) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportPolygonObject(): Could not find material export name. Bailing out."); }

        // export polygons
        bool started = false;
        for (ULONG i=0; i<polygonCount; ++i)
        {
            Face& f(mFaces[i]);
            if (f.mTextureIx == t)
            {
                // if not done yet, write material group for this object
                if (!started)
                {
                    mOBJFile.write("g " + uniqueObjectName + " " + *matName + "\n");
                    mOBJFile.write("usemtl " + *matName + "\n");
                    mOBJFile.write("s 1\n");
                    started = true;
                }
                // write quad
                if (f.mVertexIxs[3])
                {
                    sprintf(lineBuffer,
                            "f %u/%u/%u %u/%u/%u %u/%u/%u %u/%u/%u\n",
                            (unsigned int)(f.mVertexIxs[3] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[3]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[3] + mNormalIxOffset),
                            (unsigned int)(f.mVertexIxs[2] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[2]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[2] + mNormalIxOffset),
                            (unsigned int)(f.mVertexIxs[1] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[1]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[1] + mNormalIxOffset),
                            (unsigned int)(f.mVertexIxs[0] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[0]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[0] + mNormalIxOffset));
                }
                // write tri
                else
                {
                    sprintf(lineBuffer,
                            "f %u/%u/%u %u/%u/%u %u/%u/%u\n",
                            (unsigned int)(f.mVertexIxs[2] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[2]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[2] + mNormalIxOffset),
                            (unsigned int)(f.mVertexIxs[1] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[1]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[1] + mNormalIxOffset),
                            (unsigned int)(f.mVertexIxs[0] + mVertexIxOffset),
                            (unsigned int)(f.mUVWIxs[0]    + mUVWIxOffset),
                            (unsigned int)(f.mNormalIxs[0] + mNormalIxOffset));
                }
                mOBJFile.write(lineBuffer);
            }
        }
    }

    // update offsets
    mVertexIxOffset += vertexCount;
    mUVWIxOffset    += uvwCount;
    mNormalIxOffset += normalCount;

    return true;
}


bool OEOBJConverter::setupTextures(PolygonObject& object,
                                   const Matrix&  globalMatrix,
                                   HierarchyData& hierarchyData)
{
    Texture texture;

    // clear textures and add texture coming from parent object as first entry
    // to texture stack (i.e. at the bottom of the stack)
    mTextures.clear();
    mTextures.push(hierarchyData.mTexture);

    // by default we use the inherited texture, which is the first one in the stack,
    // i.e. with index==0
    for (SizeT i=0; i<mFaces.size(); ++i)
    {
        mFaces[i].mTextureIx = 0;
    }

    // loop over all texture tags and add them to texture stack. also mark all
    // polygons that use it.
    BaseMaterial* material;
    for (BaseTag* tag=object.GetFirstTag(); tag; tag=tag->GetNext())
    {
        // continue if it's not a texture tag
        if (!tag->IsInstanceOf(Ttexture)) { continue; }
        // continue if the texture tag doesn't have a valid material link
        material = (BaseMaterial*)getParameterLink(*tag, TEXTURETAG_MATERIAL, Mbase);
        if (!material) { continue; }
        // check polygon restriction
        String restriction = getParameterString(*tag, TEXTURETAG_RESTRICTION);
        // if we have a valid polygon selection, mark all selected polygons and put
        // texture on stack
        if (restriction.Content())
        {
            BaseSelect* select = getPolygonSelection(object, restriction);
            if (!select || !select->GetSegments()) { continue; }
            for (LONG start, end, segment=0;
#if _C4D_VERSION>=130
                 select->GetRange(segment, MAXLONGl, &start, &end);
#else
                 select->GetRange(segment, &start, &end);
#endif
                 ++segment)
            {
                if (end >= (LONG)mFaces.size()) { end = (LONG)mFaces.size() - 1; }
                ULONG textureIx = (ULONG)mTextures.size();
                for (LONG i=start; i<=end; ++i) { mFaces[i].mTextureIx = textureIx; }
            }
        }
        // if the texture tag applies to the complete object, clear texture stack
        // and set texture index for all faces to 0
        else
        {
            mTextures.clear();
            for (SizeT i=0; i<mFaces.size(); ++i) { mFaces[i].mTextureIx = 0; }
        }
        // add texture to stack
        OEUVWTexMapH texMap = gNew OEUVWTexMap((TextureTag*)tag, globalMatrix);
        if (!texMap) { return false; }
        mTextures.push(Texture(material, texMap));
    }

    // intialise all texture mappings with polygon object and record used materials for later export
    for (SizeT i=0; i<mTextures.size(); ++i)
    {
        // look up material and its unique name in collection
        const String *found = mMaterialNames.get(mTextures[i].mMaterial);
        if (!found) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::setupTextures(): Could not find material in collection of all materials?!? Bailing out."); }
        // add material to used set
        mUsedMaterials.insert(mTextures[i].mMaterial);
        // determine UVW coordinates for this particular texture
        mTextures[i].mTexMap->init(object, globalMatrix, mView);
    }

    // update hierarchy data with the texture at the bottom of the stack which is
    // the last material that is applied to the whole object
    hierarchyData.mTexture = mTextures.front();

    return true;
}


bool OEOBJConverter::exportMaterial(
    BaseMaterial *material)
{
    // get export name for material
    const String *matName = mMaterialNames.get(material);
    if (!matName) { ERRLOG_RETURN_VALUE(false, "OEOBJConverter::exportMaterial(): Material not found in collection of all materials?!? Bailing out."); }

    // convert C4D material to OBJ material
    OEOBJMaterial objMaterial;
    // if a NULL pointer was passed through we need to export the default material
    if (!material)
    {
        // do nothing, the default constructor of OEOBJMaterial is already setting it up
        // as the default material
    }
    // if it's a standard C4D material:
    else if (material->IsInstanceOf(Mmaterial))
    {
        // determine diffuse channel
        if (getParameterLong(*material, MATERIAL_USE_COLOR))
        {
            // get texture strength and base colour
            Vector color    = getParameterVector(*material, MATERIAL_COLOR_COLOR);
            Real   strength = getParameterReal  (*material, MATERIAL_COLOR_BRIGHTNESS, 1.0);
            color = CutColor(color) * Clamp(0.0, 1.0, strength);
            color.x = pow(color.x, cDefaultGamma);
            color.y = pow(color.y, cDefaultGamma);
            color.z = pow(color.z, cDefaultGamma);
            objMaterial.mKd.mColor = color;
            objMaterial.mKd.mEnabled = true;
            // get image from shader if possible
            getTextureFilename(*material,
                               MATERIAL_COLOR_SHADER,
                               objMaterial.mKd.mFilename);
        }
        else
        {
            objMaterial.mKd.mEnabled = false;
        }

        // determine specular channel, prefer C4D reflection channel, if available,
        // but use C4D specular channel as a fallback
        if (getParameterLong(*material, MATERIAL_USE_REFLECTION))
        {
            // get texture strength and colour from reflection channel
            Vector  color    = getParameterVector(*material, MATERIAL_REFLECTION_COLOR);
            Real    strength = getParameterReal  (*material, MATERIAL_REFLECTION_BRIGHTNESS, 1.0);
            color = CutColor(color) * Clamp(0.0, 1.0, strength);
            color.x = pow(color.x, cDefaultGamma);
            color.y = pow(color.y, cDefaultGamma);
            color.z = pow(color.z, cDefaultGamma);
            objMaterial.mKs.mColor = color * 2.0;   // KLUDGE: for some reason Octane halves the specular color
            objMaterial.mKs.mEnabled = true;
            // get image from shader if possible
            getTextureFilename(*material,
                               MATERIAL_REFLECTION_SHADER,
                               objMaterial.mKs.mFilename);
            // determine glossyness from C4D roughness
            Real dispersion = getParameterReal(*material, MATERIAL_REFLECTION_DISPERSION, 1.0);
            dispersion = pow(dispersion, 3) * 0.2;
            objMaterial.mNs.mValue = (dispersion < 0.00001) ? 100000.0 : 1.0 / dispersion;
        }
        else if (getParameterLong(*material, MATERIAL_USE_SPECULAR))
        {
            // determine specular colour and assume white as default
            Vector specColor(1.0);
            if (getParameterLong(*material, MATERIAL_USE_SPECULARCOLOR))
            {
                specColor     = getParameterVector(*material, MATERIAL_SPECULAR_COLOR);
                Real strength = getParameterReal  (*material, MATERIAL_SPECULAR_BRIGHTNESS, 1.0);
                specColor   = CutColor(specColor) * Clamp(0.0, 1.0, strength);
                specColor.x = pow(specColor.x, cDefaultGamma);
                specColor.y = pow(specColor.y, cDefaultGamma);
                specColor.z = pow(specColor.z, cDefaultGamma);
                // get image from shader if possible
                getTextureFilename(*material,
                                   MATERIAL_SPECULAR_SHADER,
                                   objMaterial.mKs.mFilename);
            }
            // derive reflectivity from specular "height"
            Real height = Clamp(0.0, 1.0, getParameterReal(*material, MATERIAL_SPECULAR_HEIGHT, 1.0));
            objMaterial.mKs.mColor = Vector(specColor * height);
            objMaterial.mKs.mEnabled = true;
            // derive glossyness from specular "width"
            Real width  = Clamp(0.0, 1.0, getParameterReal(*material, MATERIAL_SPECULAR_WIDTH));
            objMaterial.mNs.mValue = (width < 0.00001) ? 100000.0 : 1.0 / width;
        }
        else
        {
            objMaterial.mKs.mEnabled = false;
            objMaterial.mNs.mValue   = 0.0;
        }

        // determine transmission
        if (getParameterLong(*material, MATERIAL_USE_TRANSPARENCY))
        {
            Vector color    = getParameterVector(*material, MATERIAL_TRANSPARENCY_COLOR);
            Real   strength = getParameterReal(*material, MATERIAL_TRANSPARENCY_BRIGHTNESS, 1.0);
            color = CutColor(color) * Clamp(0.0, 1.0, strength);
            color.x = pow(color.x, cDefaultGamma);
            color.y = pow(color.y, cDefaultGamma);
            color.z = pow(color.z, cDefaultGamma);
            objMaterial.mTf = color;
            objMaterial.mNi = getParameterReal(*material, MATERIAL_TRANSPARENCY_REFRACTION, 1.0);
            objMaterial.mHasTransparency = true;
        }
        else
        {
            objMaterial.mTf = Vector(1.0);
            objMaterial.mNi = 1.0;
        }

        // determine alpha
        objMaterial.mD.mValue = 1.0;
        if (getParameterLong(*material, MATERIAL_USE_ALPHA))
        {
            // get image from shader if possible
            getTextureFilename(*material,
                               MATERIAL_ALPHA_SHADER,
                               objMaterial.mD.mFilename);
        }

        // determine bump channel
        objMaterial.mBump.mEnabled = false;
        if (getParameterLong(*material, MATERIAL_USE_BUMP))
        {
            // get image from shader if possible
            if (getTextureFilename(*material,
                                   MATERIAL_BUMP_SHADER,
                                   objMaterial.mBump.mFilename))
            {
                objMaterial.mBump.mEnabled = true;
                objMaterial.mBump.mStrength =
                    getParameterReal(*material, MATERIAL_BUMP_STRENGTH, 1.0) *
                    mExportScale;
            }
        }
    }
    // fallback, if it's not a standard C4D material: just create a diffuse
    // material with the average colour as diffuse colour
    else
    {
        objMaterial.mKd.mEnabled = true;
        objMaterial.mKd.mColor   = material->GetAverageColor();
        objMaterial.mKd.mColor.x = pow(objMaterial.mKd.mColor.x, cDefaultGamma);
        objMaterial.mKd.mColor.y = pow(objMaterial.mKd.mColor.y, cDefaultGamma);
        objMaterial.mKd.mColor.z = pow(objMaterial.mKd.mColor.z, cDefaultGamma);
    }

    // write material into file
    return objMaterial.write(mMTLFile, *matName);
}


bool OEOBJConverter::getTextureFilename(BaseMaterial& material,
                                        LONG          linkId,
                                        Filename&     texFilename)
{
    BaseList2D* bitmapLink = getParameterLink(material, linkId, Xbitmap);
    if (bitmapLink)
    {
        // get file path
        Filename filename = getParameterFilename(*bitmapLink, BITMAPSHADER_FILENAME);
        Filename fullFilename;
        GenerateTexturePath(mDocument->GetDocumentPath(),
                            filename,
                            Filename(),
                            &fullFilename);

        // handle animated textures (i.e. write them into separate folder using
        // a different filename)
        prepareImageForOctane(*bitmapLink, fullFilename);

        // make texture filename relative to project path
        FilePath fullPath(fullFilename);
        fullPath.makeRelativeTo(mProjectFilesDir);

        // convert FilePath back into Filename
        texFilename = fullPath.getFilename();
        return true;
    }
    else
    {
        return false;
    }
}


bool OEOBJConverter::prepareImageForOctane(BaseList2D& bitmap,
                                           Filename&   texFilename)
{
    // gather all information about the image and store it in the ImageInfo
    // structure
    ImageInfo img;
    img.mFilename = texFilename.GetString();
    const LONG startFrame = getParameterLong(bitmap, BITMAPSHADER_TIMING_FROM);
    const LONG endFrame   = getParameterLong(bitmap, BITMAPSHADER_TIMING_TO);
    const bool isAnimated = (endFrame > startFrame);
    if (isAnimated)
    {
        img.mFilename     = texFilename.GetString();
        img.mPlaybackMode = getParameterLong(bitmap,
                                             BITMAPSHADER_TIMING_MODE,
                                             BITMAPSHADER_TIMING_MODE_SIMPLE);
        img.mTimingMode   = getParameterLong(bitmap,
                                             BITMAPSHADER_TIMING_TIMING,
                                             BITMAPSHADER_TIMING_TIMING_FRAME);
        img.mStartFrame   = startFrame;
        img.mEndFrame     = endFrame;
        if (img.mTimingMode == BITMAPSHADER_TIMING_TIMING_SECOND)
        {
            img.mFps        = getParameterReal(bitmap,
                                               BITMAPSHADER_TIMING_FPS,
                                               mDocument->GetFps());
        }
        else if (img.mTimingMode == BITMAPSHADER_TIMING_TIMING_AREA)
        {
            img.mRangeFrom  = getParameterTime(bitmap, BITMAPSHADER_TIMING_RANGEFROM);
            img.mRangeTo    = getParameterTime(bitmap, BITMAPSHADER_TIMING_RANGETO);
            img.mRangeLoops = getParameterLong(bitmap, BITMAPSHADER_TIMING_LOOPS) + 1;
        }
    }

    // check if the image was already handled (i.e. is in the image info list)
    const Filename* prepFilename = mImageInfos.get(img);
    if (prepFilename)
    {
        // we have dealed with it before -> reuse the filename from previous run
        texFilename = *prepFilename;
        return true;
    }

    // check if file type is supported
    FilePath     srcFilePath = texFilename;
    const String srcSuffix = srcFilePath.getSuffix();
    const bool   isSupported = supportedByOctane(srcSuffix);

    // if we don't want to collect images and the image is supported and not
    // animated, leave it where it is and return
    if (!mCollectImages && isSupported && !isAnimated)
    {
        mImageInfos.set(img, texFilename);
        return true;
    }

    // ok, when we get here we need to do some more work. the texture is either
    // - unsupported and needs to be converted to a supported format
    // - an animation and we have to pull out the current frame
    // - an image and all images should be copied to the files sub-folder
    // - any combination of the above
    // -> we need to create a copy of the image/frame in the sub-folder

    // if it's an animated image, get the current frame
    const ULONG frame = isAnimated ? determineCurrentFrame(img) : 0;

    // determine the destination file string without suffix
    AutoAlloc<BaseBitmap> unsupportedFrame;
    FilePath              dstFilePath = srcFilePath.getFileString();
    if (isSupported)
    {
        // if it's supported, it must be an image, if it's animated, we need
        // to set the source path to the actual frame (in the image sequence)
        if (isAnimated)
        {
            srcFilePath.replaceIndex(frame);
            dstFilePath.replaceIndex(-1);
        }
    }
    else
    {
        // if it's not supported it can be a movie or an image. to find out we
        // try to the calculated frame into a bitmap
        Bool isMovie = FALSE;
        if (unsupportedFrame->Init(texFilename, frame, &isMovie) != IMAGERESULT_OK)
        {
            ERRLOG_RETURN_VALUE(false, String("OEOBJConverter::prepareImageForOctane: Failed to load frame of '" + srcFilePath.getString() + "'"));
        }
        if (isAnimated && !isMovie)
        {
            // if it's an animation from an image sequence, we have to reload the
            // image now from the correct image in the sequence
            srcFilePath.replaceIndex(frame);
            dstFilePath.replaceIndex(-1);
            if (unsupportedFrame->Init(srcFilePath.getFilename(), frame) != IMAGERESULT_OK)
            {
                ERRLOG_RETURN_VALUE(false, String("OEOBJConverter::prepareImageForOctane: Failed to load frame of '" + srcFilePath.getString() + "'"));
            }
        }
    }

    // create unique name of the destination image and make it a full path
    String unique;
    findUniqueName(dstFilePath.getFileString(), mImageNameCounts, unique);
    dstFilePath = mProjectFilesDir + unique;
#if _C4D_VERSION >= 120
    dstFilePath.setSuffix(isSupported ? srcSuffix : ".png");
#else
    dstFilePath.setSuffix(isSupported ? srcSuffix : ".tif");
#endif

    // animated images will always be copied, but not animated ones only if
    // stuff has changed
    if (!isAnimated && GeFExist(dstFilePath.getFilename()))
    {
        LocalFileTime srcTime, dstTime;
        if (!GeGetFileTime(srcFilePath.getFilename(), GE_FILETIME_MODIFIED, &srcTime) ||
            !GeGetFileTime(dstFilePath.getFilename(), GE_FILETIME_MODIFIED, &dstTime))
        {
            ERRLOG_RETURN_VALUE(false, String("OEOBJConverter::prepareImageForOctane: Failed to determine file time stamps"));
        }
        // if the time stamps are exactly the same we assume, that nothing changed
        if (srcTime == dstTime)
        {
            texFilename = dstFilePath.getFilename();
            mImageInfos.set(img, texFilename);
            return true;
        }
    }

    // now finally copy the image over...

    // determine file path we will actually copy to - this can be either
    // dstFilePath or a temporary file name
    FilePath copyFilePath = dstFilePath;
    if (mWriteTempFiles)
    {
        copyFilePath.setSuffix(copyFilePath.getSuffix() + "__temp__");
    }

    // if it's a supported file format, we just copy the file as it is
    if (isSupported)
    {
        if (!GeFCopyFile(srcFilePath.getFilename(),
                         copyFilePath.getFilename(),
                         GE_FCOPY_OVERWRITE))
        {
            ERRLOG_RETURN_VALUE(false, String("OEOBJConverter::prepareImageForOctane: Failed to copy file '" + srcFilePath.getString() + "' to '" + copyFilePath.getString() + "'"));
        }
    }
    // if it's an unsupported file format, we save the frame loaded into
    // unsupportedFrame as file
    else
    {
        // save bitmap
        if (unsupportedFrame->Save(copyFilePath.getFilename(),
#if _C4D_VERSION >= 120
                                   FILTER_PNG,
#else
                                   FILTER_TIF,
#endif
                                   NULL,
                                   SAVEBIT_ALPHA) != IMAGERESULT_OK)
        {
            ERRLOG_RETURN_VALUE(false, String("OEOBJConverter::prepareImageForOctane: Failed to save file '" + copyFilePath.getString() + "'"));
        }
#if _C4D_VERSION >= 105
        // copy time stamp - to avoid rewriting this image in the next export if
        // it hasn't changed
        LocalFileTime srcTime;
        if (!GeGetFileTime(srcFilePath.getFilename(), GE_FILETIME_MODIFIED, &srcTime) ||
            !GeSetFileTime(copyFilePath.getFilename(), GE_FILETIME_MODIFIED, &srcTime))
        {
            ERRLOG_RETURN_VALUE(false, String("OEOBJConverter::prepareImageForOctane: Failed to copy time stamp from '" + srcFilePath.getString() + "' to '" + copyFilePath.getString() + "'"));
        }
#endif
    }

    // store rename in list
    if (mWriteTempFiles)
    {
        mRenameList.pushBack(ToRename(copyFilePath, dstFilePath));
    }

    // update texture image filename, add image info to list and return
    texFilename = dstFilePath.getFilename();
    mImageInfos.set(img, texFilename);
    return true;
}


ULONG OEOBJConverter::determineCurrentFrame(const ImageInfo& img)
{
    // get document time and fps
    const BaseTime docTime = mDocument->GetTime();
    const LONG     docFps  = mDocument->GetFps();

    // the frame calculation will be done relative to the sequence start.
    // the start offset will be applied at the end.
    const LONG animFrameCount = img.mEndFrame - img.mStartFrame + 1;
    LONG       frame = 0;

    // if the timing mode is exact frame, align anim frame with document frames
    if (img.mTimingMode == BITMAPSHADER_TIMING_TIMING_FRAME)
    {
        frame = docTime.GetFrame(docFps);
    }
    // if the timing mode is exact frame, align anim time with document time
    else if (img.mTimingMode == BITMAPSHADER_TIMING_TIMING_SECOND)
    {
        frame = docTime.GetFrame(img.mFps);
    }
    // if the timing mode is range, we have to squeeze tex.mRangeLoops iterations
    // into the range tex.mRangeFrom - tex.mRangeTo
    else if (img.mTimingMode == BITMAPSHADER_TIMING_TIMING_AREA)
    {
        // special case 1: haven't reached range yet
        if (docTime <= img.mRangeFrom)
        {
            frame = 0;
        }
        // special case 2: have passed the range
        else if (docTime >= img.mRangeTo)
        {
            frame = (img.mPlaybackMode == BITMAPSHADER_TIMING_MODE_PINGPONG) ?
                    (animFrameCount - 1) * img.mRangeLoops :
                    animFrameCount * img.mRangeLoops - 1;
        }
        // handle normal case when document time is between rangeFrom and rangeTo
        else
        {
            const BaseTime sequTime = (img.mPlaybackMode == BITMAPSHADER_TIMING_MODE_PINGPONG) ?
                                      BaseTime(animFrameCount - 1, docFps) :
                                      BaseTime(animFrameCount, docFps);
            const BaseTime loops(img.mRangeLoops, 1);
            const BaseTime frameTime =
                (sequTime * loops * (docTime - img.mRangeFrom) /
                (img.mRangeTo - img.mRangeFrom));
            frame = frameTime.GetFrame(docFps);
        }
    }

    // now that we have the frame index in the animation sequence, we need to apply
    // the playback mode which maps that index into the range
    // tex.mStartFrame .. tex.mEndFrame

    // simple playback
    if (img.mPlaybackMode == BITMAPSHADER_TIMING_MODE_SIMPLE)
    {
        frame = clamp<LONG>(0, animFrameCount - 1, frame);
    }
    // loop playback
    else if (img.mPlaybackMode == BITMAPSHADER_TIMING_MODE_LOOP)
    {
        frame = frame % animFrameCount;
    }
    // ping-pong playback
    else if (img.mPlaybackMode == BITMAPSHADER_TIMING_MODE_PINGPONG)
    {
        if ((frame / (animFrameCount - 1)) & 1)
        {
            frame = (animFrameCount - 1) - (frame % (animFrameCount - 1));
        }
        else
        {
            frame = frame % (animFrameCount - 1);
        }
    }

    // apply animation offset and clamp at 0
    return (ULONG)std::max<LONG>(frame + img.mStartFrame, 0);
}


bool OEOBJConverter::supportedByOctane(const String& suffix)
{
    // All image formats supported by C4D and Octane. They are sorted by approx.
    // by popularity.
    static const char* cSupportedFormats[] =
        {
            ".jpg",
            ".jpeg",
            ".png",
            ".tif",
            ".tiff"
            ".tga",
            ".exr",
            ".hdr",
            ".bmp",
            ".gif",
            ".iff",
            ".lbm",
            ".pict",
        };
    static const size_t cSupportedFormatCount = ARRAY_LENGTH(cSupportedFormats);

    // now loop over all supported suffixes and compare them against the lower
    // case version of the input suffix
    const String lowerSuffix = suffix.ToLower();
    for (size_t i=0; i<cSupportedFormatCount; ++i)
    {
        if (lowerSuffix == cSupportedFormats[i]) { return true; }
    }

    return false;
}


BaseSelect* OEOBJConverter::getPolygonSelection(PolygonObject& object,
                                                const String&  name)
{
    for (BaseTag* tag=object.GetFirstTag(); tag; tag=tag->GetNext())
    {
        if (tag->IsInstanceOf(Tpolygonselection) && (tag->GetName() == name))
        {
            return ((SelectionTag*)tag)->GetBaseSelect();
        }
    }
    return 0;
}
