#ifndef _OE_OBJ_CONVERTER_H_
#define _OE_OBJ_CONVERTER_H_  1



#include <c4d.h>

#include "autoref.h"
#include "bufferedtextfile.h"
#include "dlist.h"
#include "dynarray1d.h"
#include "filepath.h"
#include "oe_uvw_texmap.h"
#include "rbtreemap.h"
#include "rbtreeset.h"
#include "stl_containers.h"



struct OEOBJMaterial;



///
class OEOBJConverter : public Hierarchy
{
public:

    struct ToRename
    {
        FilePath mFromPath;
        FilePath mToPath;

        ToRename();

        ToRename(const FilePath& fromPath,
                 const FilePath& toPath);
    };

    typedef DList<ToRename> RenameListT;

    OEOBJConverter();
    ~OEOBJConverter();

    bool convertScene(BaseDocument&   document,
                      const Real      scale,
                      const bool      useRelativePaths,
                      const Filename& projectDirectory,
                      const String&   projectName,
                      const bool      collectImages,
                      const bool      writeTempFiles,
                      FilePath        &objFilePath,
                      RenameListT     &renameList);

    //--- Implementations for Hierarchy ---

    virtual void* Alloc();
    virtual void Free(void* data);
    virtual void CopyTo(void* src,
                        void* dst);
    virtual Bool Do(void*         data,
                    BaseObject*   object,
                    const Matrix& globalMatrix,
                    Bool          controlObject);

private:

    struct HierarchyData;
    struct Texture;
    struct ImageInfo;
    struct Face;
    struct SimpleFace;

    typedef RBTreeSet<BaseList2D*>           ObjectsT;
    typedef StdSet<BaseMaterial*>            MaterialsT;
    typedef StdMap<String, OEOBJMaterial>    LightMaterialsT;
    typedef RBTreeMap<BaseMaterial*, String> MaterialNamesT;
    typedef RBTreeMap<String, ULONG>         NameCountsT;
    typedef RBTreeMap<ImageInfo, Filename>   ImageInfoT;
    typedef DynArray1D<Texture>              TexturesT;
    typedef DynArray1D<Vector>               VerticesT;
    typedef DynArray1D<SVector>              NormalsT;
    typedef DynArray1D<Face>                 FacesT;
    typedef DynArray1D<SimpleFace>           SimpleFacesT;

    BaseDocument     *mDocument;
    FilePath         mProjectDir;
    FilePath         mProjectFilesDir;
    bool             mCollectImages;
    bool             mWriteTempFiles;
    RenameListT      mRenameList;
    BufferedTextFile mOBJFile;
    BufferedTextFile mMTLFile;
    Real             mExportScale;
    BaseView*        mView;
    ObjectsT         mAreaLightObjects;
    MaterialsT       mUsedMaterials;
    LightMaterialsT  mUsedLightMaterials;
    MaterialNamesT   mMaterialNames;
    NameCountsT      mMaterialNameCounts;
    NameCountsT      mObjectNameCounts;
    TexturesT        mTextures;
    ImageInfoT       mImageInfos;
    NameCountsT      mImageNameCounts;
    FacesT           mFaces;
    /// The running vertex offset counter.
    ULONG            mVertexIxOffset;
    /// The UVW vertex offset counter.
    ULONG            mUVWIxOffset;
    /// The normal index offset counter.
    ULONG            mNormalIxOffset;


    void clearTemporaryData();
    ///
    static void findUniqueName(const String& name,
                               NameCountsT&  nameCountMap,
                               String&       uniqueName);

    /// Loops through the object hierarchy and collects all objects that are used for object
    /// area lights. Those objects will not be exported as part of the geometry export, but during
    /// the light export.
    void collectLightObjects();
    /// Determines unique names for all materials and collects them in a map. Those names will be
    /// used during the export.
    bool collectMaterials();
    /// Traverses the hierarchy exports the whole scene.
    bool exportObjects();
    /// Exports all used materials, sorted by name.
    bool exportMaterials();

    bool exportLightObject(BaseObject&   object,
                           const Matrix& globalMatrix);
    bool exportDiscAreaLight(BaseObject&   object,
                             const Matrix& globalMatrix);
    bool exportRectangleAreaLight(BaseObject&   object,
                                  const Matrix& globalMatrix);
    bool exportSphereAreaLight(BaseObject&   object,
                               const Matrix& globalMatrix);
    bool exportCylinderAreaLight(BaseObject&   object,
                                 const Matrix& globalMatrix,
                                 bool          perpendicular);
    bool exportCubeAreaLight(BaseObject&   object,
                             const Matrix& globalMatrix);
    bool exportHemisphereAreaLight(BaseObject&   object,
                                   const Matrix& globalMatrix);
    bool exportObjectAreaLight(BaseObject&   object);
    bool exportLightGeometry(BaseObject&         object,
                             const Matrix&       globalMatrix,
                             const VerticesT&    vertices,
                             const NormalsT&     normals,
                             const SimpleFacesT& faces);

    bool exportPolygonObject(PolygonObject& object,
                             const Matrix&  globalMatrix,
                             HierarchyData& hierarchyData);
    bool setupTextures(PolygonObject& object,
                       const Matrix&  globalMatrix,
                       HierarchyData& hierarchyData);
    bool exportMaterial(BaseMaterial* material);

    bool getTextureFilename(BaseMaterial& material,
                            LONG          linkId,
                            Filename&     texFilename);

    bool prepareImageForOctane(BaseList2D& bitmap,
                               Filename&   file);

    ULONG determineCurrentFrame(const ImageInfo& img);

    static bool supportedByOctane(const String& suffix);

    BaseSelect* getPolygonSelection(PolygonObject& object,
                                    const String&  name);
};



#endif  // #ifndef _OE_OBJ_CONVERTER_H_