#include "oe_uvw_texmap.h"



OEUVWTexMap::OEUVWTexMap(TextureTag*   textureTag,
                         const Matrix& globalMatrix)
: mTextureTag(textureTag),
  mTexGlobMat(globalMatrix),
  mUVWTagOwner(false),
  mUVWTag(0),
  mTagData(0),
  mScale(1.0),
  mShift(0.0)
{
  // determine scale and offset from tag
  if (textureTag) {
    mScale = Vector(getParameterReal(*textureTag, TEXTURETAG_TILESX, 1.0),
                    getParameterReal(*textureTag, TEXTURETAG_TILESY, 1.0),
                    1.0);
    mShift = Vector(getParameterReal(*textureTag, TEXTURETAG_OFFSETX, 0.0),
                    getParameterReal(*textureTag, TEXTURETAG_OFFSETY, 0.0),
                    0.0);
  }
}


OEUVWTexMap::~OEUVWTexMap()
{
  finish();
}


void OEUVWTexMap::init(PolygonObject& object,
                       const Matrix&  globalMatrix,
                       BaseView*      view)
{
  // clean up before we start again
  finish();

  // if the projection type is UVW, we have to find the UVW tag ourself:
  if (mTextureTag &&
      getParameterLong(*mTextureTag, TEXTURETAG_PROJECTION) == TEXTURETAG_PROJECTION_UVW)
  {
    mUVWTagOwner = false;
    // if the texture tag is assigned to the passed in polygon object, use the
    // next tag on the right relative to the texture tag. if there isn't one,
    // wrap around.
    if (mTextureTag && (mTextureTag->GetObject() == &object)) {
      for (BaseTag* tag=mTextureTag->GetNext();
           tag!=mTextureTag;
           tag=tag->GetNext())
      {
        if (!tag) {
          tag = object.GetFirstTag();
          if (tag == mTextureTag) { break; }
        }
        if (tag->IsInstanceOf(Tuvw)) {
          mUVWTag = (UVWTag*)tag;
          break;
        }
      }
    // if the texture tag is actually coming from a different object, use the
    // the left-most UVW tag we can find on the polygon object
    } else {
      for (BaseTag* tag=object.GetFirstTag(); tag; tag=tag->GetNext()) {
        if (tag->IsInstanceOf(Tuvw)) {
          mUVWTag = (UVWTag*)tag;
          break;
        }
      }
    }
  // if the projection type is not UVW, but something else, we let C4D create a
  // UVWTag from it
  } else if (mTextureTag) {
    mUVWTagOwner = true;
    mUVWTag = GenerateUVW(&object, globalMatrix, mTextureTag, mTexGlobMat, view);
  }

#if _C4D_VERSION>=115
  if (mUVWTag) {
    mTagData = mUVWTag->GetDataAddressR();
  }
#endif
}


void OEUVWTexMap::finish()
{
  if (mUVWTagOwner) {
    UVWTag::Free(mUVWTag);
    mUVWTagOwner = false;
  }
}


void OEUVWTexMap::getUVWs(ULONG  polyIx,
                          Vector uvws[4])
{
  UVWStruct uvwStruct;
  if (mUVWTag) {
#if _C4D_VERSION>=115
    mUVWTag->Get(mTagData, (LONG)polyIx, uvwStruct);
#else
    uvwStruct = mUVWTag->Get((LONG)polyIx);
#endif
    uvws[0] = (uvwStruct.a - mShift) ^ mScale;
    uvws[1] = (uvwStruct.b - mShift) ^ mScale;
    uvws[2] = (uvwStruct.c - mShift) ^ mScale;
    uvws[3] = (uvwStruct.d - mShift) ^ mScale;
  } else {
    uvws[0] = Vector(0.0);
    uvws[1] = Vector(0.0);
    uvws[2] = Vector(0.0);
    uvws[3] = Vector(0.0);
  }
}
