#ifndef _RBTREEMAP_H_
#define _RBTREEMAP_H_  1



#include <new>

#include <c4d.h>



/***************************************************************************//*!
 This class implments a map, based on the very cool "simplified"
 left-leaning red-black tree of Robert Sedgewick:
   http://www.cs.princeton.edu/~rs/

 The template key type K has to support the following functions:
  - Less operator     bool K::operator<(const K&)

 The template type has to support the following functions:
  - Copy constructor  T::T(const T&)
  - Copy operator     T::operator=(const T&)
*//****************************************************************************/

template <class K, class T, bool sUseMemoryPool = false>
class RBTreeMap
{
public:

  /// The key type.
  typedef K       KeyT;
  /// The value type.
  typedef T       ValueT;


  RBTreeMap();
  ~RBTreeMap();

  void erase();

  inline SizeT size() const;

  ValueT* set(const KeyT&   key,
              const ValueT& value,
              const bool    overwrite = true);

  const ValueT* get(const KeyT& key) const;
  ValueT*       get(const KeyT& key);


private:


  /// This helper structure implements a single node of a red-black tree.
  struct Node 
  {
    Node*  mLeft;
    Node*  mRight;
    bool   mIsRed;
    KeyT   mKey;
    ValueT mValue;


    /// Constructs a new node and sets its value.
    /// @param[in]  value
    ///   The value of which we store a copy in the new node.
    inline Node(const KeyT& key)
    : mLeft(0),
      mRight(0),
      mIsRed(true),
      mKey(key)
    {}

    /// Static helper function that checks if a pointer points to a black node.
    /// @param[in]  node
    ///   Pointer to a node, which can be NULL.
    /// @return 
    ///   true if the pointer is valid and is pointing to a black node.
    inline static bool isBlack(const Node* node)
    {
      return (node && !node->mIsRed);
    }

    /// Static helper function that checks if a pointer points to a red node.
    /// @param[in]  node
    ///   Pointer to a node, which can be NULL.
    /// @return 
    ///   true if the pointer is valid and is pointing to a red node.
    inline static bool isRed(const Node* node)
    {
      return (node && node->mIsRed);
    }
  };


  Node*       mRoot;
  SizeT       mSize;
  Node*       mNewNode;
  MemoryPool* mMemPool;


  // At the moment, we don't allow copying of RBTreeMaps.
  RBTreeMap(const RBTreeMap& other) {}
  RBTreeMap& operator==(const RBTreeMap& other)  {}

  inline static Node* rotateLeft(Node* node);
  inline static Node* rotateRight(Node* node);
  inline static void  colorFlip(Node* node);
  
  Node* insert(Node*       node,
               const KeyT& key);

  static void deleteRecursive(Node* node);
};


#include "rbtreemap_impl.h"



#endif  // #ifndef _RBTREEMAP_H_
