#ifndef _RBTREEMAP_IMPL_H_
#define _RBTREEMAP_IMPL_H_  1



/*******************************************************************************
 * Implementation of public member functions of template class RBTreeMap.
 *******************************************************************************/

/// Constructs a new (empty) red-black tree.
template <class K, class T, bool sUseMemoryPool>
RBTreeMap<K,T,sUseMemoryPool>::RBTreeMap()
: mRoot(0), mSize(0), mMemPool(0)
{
  if (sUseMemoryPool) {
    mMemPool = MemoryPool::Alloc(sizeof(Node) * 10240);
  }
}


/// Destroys a red-black tree and deallocates its resources.
template <class K, class T, bool sUseMemoryPool>
RBTreeMap<K,T,sUseMemoryPool>::~RBTreeMap()
{
  deleteRecursive(mRoot);
  if (sUseMemoryPool) {
    MemoryPool::Free(mMemPool);
  }
}


/// Deallocates all resources of a red-black tree. The tree will be empty
/// afterwards.
template <class K, class T, bool sUseMemoryPool>
void RBTreeMap<K,T,sUseMemoryPool>::erase()
{
  deleteRecursive(mRoot);
  if (sUseMemoryPool) {
    MemoryPool::Free(mMemPool);
    mMemPool = MemoryPool::Alloc(sizeof(Node) * 10240);
  }
  mRoot = 0;
  mSize = 0;
}


/// Returns the number of entries of red-black tree.
template <class K, class T, bool sUseMemoryPool>
inline SizeT RBTreeMap<K,T,sUseMemoryPool>::size() const
{
  return mSize;
}


/// Adds a new key/entry pair to the LLRB set. If there was already one with
/// the same key, we overwrite it.
///
/// @param[in]  key
///   The key of the value to add to the rb-tree.
/// @param[in]  value
///   The value to add.
/// @param[in]  overwrite  (optional)
///   If set to true, we replace already existing entries with the new one (default: true).
/// @return
///   NULL if we ran out of memory, otherwise the pointer to the stored value.
template <class K, class T, bool sUseMemoryPool>
typename RBTreeMap<K,T,sUseMemoryPool>::ValueT*
RBTreeMap<K,T,sUseMemoryPool>::set(const KeyT&   key,
                                   const ValueT& value,
                                   const bool    overwrite)
{
  SizeT oldSize = mSize;
  mNewNode = 0;
  mRoot = insert(mRoot, key);
  if (mNewNode) {
    if (overwrite || (mSize > oldSize)) { mNewNode->mValue = value; }
    return &mNewNode->mValue;
  }
  return 0;
}


/// Returns the (const) value for a specified key from the map.
///
/// @param[in]  key
///   The key for which the value will be looked up.
/// @return
///   Pointer to the found value, or NULL if there is no matching value.
template <class K, class T, bool sUseMemoryPool>
const typename RBTreeMap<K,T,sUseMemoryPool>::ValueT*
RBTreeMap<K,T,sUseMemoryPool>::get(const KeyT& key) const
{
  const Node* node = mRoot;
  while (node) {
    if (key < node->mKey) {
      node = node->mLeft;
    } else if (node->mKey < key) {
      node = node->mRight;
    } else {
      return &node->mValue;
    }
  }
  return 0;
}


/// Returns the (non-const) value for a specified key from the map.
///
/// @param[in]  key
///   The key for which the value will be looked up.
/// @return
///   Pointer to the found value, or NULL if there is no matching value.
template <class K, class T, bool sUseMemoryPool>
typename RBTreeMap<K,T,sUseMemoryPool>::ValueT*
RBTreeMap<K,T,sUseMemoryPool>::get(const KeyT& key)
{
  Node* node = mRoot;
  while (node) {
    if (key < node->mKey) {
      node = node->mLeft;
    } else if (node->mKey < key) {
      node = node->mRight;
    } else {
      return &node->mValue;
    }
  }
  return 0;
}


/*******************************************************************************
 * Implementation of private member functions of template class RBTreeSet.
 *******************************************************************************/

/// Rotates a sub-tree one step to the left without changing the order.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree (must not be NULL).
/// @return
///   Pointer to the new root node of the rotated sub-tree.
template <class K, class T, bool sUseMemoryPool>
inline typename RBTreeMap<K,T,sUseMemoryPool>::Node*
RBTreeMap<K,T,sUseMemoryPool>::rotateLeft(Node* node)
{
  Node* child   = node->mRight;
  node->mRight  = child->mLeft;
  child->mLeft  = node;
  child->mIsRed = node->mIsRed;
  node->mIsRed  = true;
  return child;
}


/// Rotates a sub-tree one step to the right without changing the order.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree (must not be NULL).
/// @return
///   Pointer to the new root node of the rotated sub-tree.
template <class K, class T, bool sUseMemoryPool>
inline typename RBTreeMap<K,T,sUseMemoryPool>::Node*
RBTreeMap<K,T,sUseMemoryPool>::rotateRight(Node* node)
{
  Node* child   = node->mLeft;
  node->mLeft   = child->mRight;
  child->mRight = node;
  child->mIsRed = node->mIsRed;
  node->mIsRed  = true;
  return child;
}


/// Flips the colors of a node and it's children. It's used to merge the
/// 2 red links of the children to one red link in the parent.
///
/// NOTE: This function can be called only, if the children of the node are
///       valid!
///
/// @param[in]  node
///   Pointer to the parent node (must nit be NULL).
template <class K, class T, bool sUseMemoryPool>
inline void RBTreeMap<K,T,sUseMemoryPool>::colorFlip(Node* node)
{
  node->mIsRed         ^= 1;
  node->mLeft->mIsRed  ^= 1;
  node->mRight->mIsRed ^= 1;
}


/// Recursive function that inserts a node into a LLRB sub-tree. If a node with
/// the same key already exists, it gets overwritten.
/// The properties of a LLRB tree will be preserved.
///
/// TODO: Make this function iterative using a stack.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree into which the new key/value pair
///   should be inserted (must not be NULL initially).
/// @param[in]  key
///   The key for which the value will be inserted.
/// @return
///   Pointer to the new root node of the sub-tree or NULL, if the allocation
///   of a new tree node failed.
template <class K, class T, bool sUseMemoryPool>
typename RBTreeMap<K,T,sUseMemoryPool>::Node*
RBTreeMap<K,T,sUseMemoryPool>::insert(Node*       node,
                                      const KeyT& key)
{
  // if we have found a leave, add new node to it and return
  if (!node) {
    // allocate memory
    void* mem;
    if (sUseMemoryPool) {
      mem = mMemPool->MemAlloc(sizeof(Node), false);
    } else {
      mem = GeAllocNC(sizeof(Node));
    }
    // construct Node using placement new
    if (mem) {
      mNewNode = new(mem) Node(key);
      ++mSize;
    } else {
      mNewNode = 0;
    }
    return mNewNode;
  }

  // traverse down the left sub-tree, if key is smaller than the key of the
  // current node
  if (key < node->mKey) {
    node->mLeft = insert(node->mLeft, key);
  // traverse down the right sub-tree, if key is larger than the key of the
  // current node
  } else if (node->mKey < key) {
    node->mRight = insert(node->mRight, key);
  // if we have found an already existing node with the same key, overwrite it
  } else {
    mNewNode = node;
  }

  // fix right-leaning leafs on the way up
  if (Node::isRed(node->mRight)) {
    node = rotateLeft(node);
  }

  // fix two reds in a row on the way up
  if (Node::isRed(node->mLeft) && Node::isRed(node->mLeft->mLeft)) {
    node = rotateRight(node);
  }

  // split 4-nodes on the way up
  if (Node::isRed(node->mLeft) && Node::isRed(node->mRight)) {
    colorFlip(node);
  }
  return node;
}


/// Deletes recursively all nodes of a sub-tree.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree (can be NULL).
template <class K, class T, bool sUseMemoryPool>
void RBTreeMap<K,T,sUseMemoryPool>::deleteRecursive(Node* node)
{
  if (!node) { return; }
  deleteRecursive(node->mLeft);
  deleteRecursive(node->mRight);
  node->Node::~Node();
  if (!sUseMemoryPool) { GeFree(node); }
}



#endif  // #ifndef _RBTREEMAP_IMPL_H_
