#ifndef _RBTREESET_H_
#define _RBTREESET_H_  1



/***************************************************************************//*!
 This class implments a set, based on the very cool "simplified"
 left-leaning red-black tree of Robert Sedgewick:
   http://www.cs.princeton.edu/~rs/

 The template type has to support the following functions:
  - Copy constructor  T::T(const T&)
  - Copy operator     T::operator=(const T&)
  - Less operator     bool T::operator<(const T&)
*//****************************************************************************/

template <class T, bool sUseMemoryPool = false>
class RBTreeSet
{
public:

  /// The value type.
  typedef T ValueT;


  RBTreeSet();
  ~RBTreeSet();

  void erase();

  inline SizeT size() const;

  const T* add(const T& value,
               bool     overwrite = true);

  const T* get(const T& value) const;
  T*       get(const T& value);


private:


  /// This helper structure implements a single node of a red-black tree.
  struct Node 
  {
    Node*  mLeft;
    Node*  mRight;
    bool   mIsRed;
    ValueT mValue;


    /// Constructs a new node and sets its value.
    /// @param[in]  value
    ///   The value of which we store a copy in the new node.
    inline Node(const ValueT& value)
    : mLeft(0),
      mRight(0),
      mIsRed(true),
      mValue(value)
    {}

    /// Static helper function that checks if a pointer points to a black node.
    /// @param[in]  node
    ///   Pointer to a node, which can be NULL.
    /// @return 
    ///   true if the pointer is valid and is pointing to a black node.
    inline static bool isBlack(const Node* node)
    {
      return (node && !node->mIsRed);
    }

    /// Static helper function that checks if a pointer points to a red node.
    /// @param[in]  node
    ///   Pointer to a node, which can be NULL.
    /// @return 
    ///   true if the pointer is valid and is pointing to a red node.
    inline static bool isRed(const Node* node)
    {
      return (node && node->mIsRed);
    }
  };


  Node*       mRoot;
  SizeT       mSize;
  Node*       mNewNode;
  MemoryPool* mMemPool;


  // At the moment, we don't allow copying of RBTreeMaps.
  RBTreeSet(const RBTreeSet& other) {}
  RBTreeSet& operator==(const RBTreeSet& other)  {}

  inline static Node* rotateLeft(Node* node);
  inline static Node* rotateRight(Node* node);
  inline static void  colorFlip(Node* node);
  
  Node* insert(Node*         node,
               const ValueT& value);

  static void deleteRecursive(Node* node);
};


#include "rbtreeset_impl.h"



#endif  // #ifndef _RBTREESET_H_
