#ifndef _RBTREESET_IMPL_H_
#define _RBTREESET_IMPL_H_  1



/*******************************************************************************
 * Implementation of public member functions of template class RBTreeSet.
 *******************************************************************************/

/// Constructs a new (empty) red-black tree.
template <class T, bool sUseMemoryPool>
RBTreeSet<T,sUseMemoryPool>::RBTreeSet()
: mRoot(0), mSize(0), mMemPool(0)
{
  if (sUseMemoryPool) {
    mMemPool = MemoryPool::Alloc(sizeof(Node) * 10240);
  }
}


/// Destroys a red-black tree and deallocates its resources.
template <class T, bool sUseMemoryPool>
RBTreeSet<T,sUseMemoryPool>::~RBTreeSet()
{
  deleteRecursive(mRoot);
  if (sUseMemoryPool) {
    MemoryPool::Free(mMemPool);
  }
}


/// Deallocates all resources of a red-black tree. The tree will be empty
/// afterwards.
template <class T, bool sUseMemoryPool>
void RBTreeSet<T,sUseMemoryPool>::erase()
{
  deleteRecursive(mRoot);
  if (sUseMemoryPool) {
    MemoryPool::Free(mMemPool);
    mMemPool = MemoryPool::Alloc(sizeof(Node) * 10240);
  }
  mRoot = 0;
  mSize = 0;
}


/// Returns the number of entries of red-black tree.
template <class T, bool sUseMemoryPool>
inline SizeT RBTreeSet<T,sUseMemoryPool>::size() const
{
  return mSize;
}


/// Adds a new entry to the LLRB set. If there was already one with the same
/// rang (/key), we overwrite it.
///
/// @param[in]  value
///   The value to add to the red-black tree.
/// @param[in]  overwrite  (optional)
///   If set to true, we replace already existing entries with the new one (default: true).
/// @return
///   NULL if we ran out of memory, otherwise the pointer to the stored entry.
template <class T, bool sUseMemoryPool>
const T* RBTreeSet<T,sUseMemoryPool>::add(const T& value,
                                          bool     overwrite)
{
  SizeT oldSize = mSize;
  mNewNode = 0;
  mRoot = insert(mRoot, value);
  if (mNewNode) {
    if (overwrite && (mSize == oldSize)) { mNewNode->mValue = value; }
    return &mNewNode->mValue;
  }
  return 0;
}


/// Returns the (const) value from the set with the same order as the passed in
/// value.
///
/// @param[in]  value
///   The value that defines the rang of the value to retrieve.
/// @return
///   Pointer to the found value, or NULL if there is no matching value.
template <class T, bool sUseMemoryPool>
const T* RBTreeSet<T,sUseMemoryPool>::get(const T& value) const
{
  const Node* node = mRoot;
  while (node) {
    if (value < node->mValue) {
      node = node->mLeft;
    } else if (node->mValue < value) {
      node = node->mRight;
    } else {
      return &node->mValue;
    }
  }
  return 0;
}


/// Returns the (non-const) value from the set with the same order as the passed in
/// value.
///
/// @param[in]  value
///   The value that defines the rang of the value to retrieve.
/// @return
///   Pointer to the found value, or NULL if there is no matching value.
template <class T, bool sUseMemoryPool>
T* RBTreeSet<T,sUseMemoryPool>::get(const T& value)
{
  Node* node = mRoot;
  while (node) {
    if (value < node->mValue) {
      node = node->mLeft;
    } else if (node->mValue < value) {
      node = node->mRight;
    } else {
      return &node->mValue;
    }
  }
  return 0;
}



/*******************************************************************************
 * Implementation of private member functions of template class RBTreeSet.
 *******************************************************************************/

/// Rotates a sub-tree one step to the left without changing the order.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree (must not be NULL).
/// @return
///   Pointer to the new root node of the rotated sub-tree.
template <class T, bool sUseMemoryPool>
inline typename RBTreeSet<T,sUseMemoryPool>::Node*
RBTreeSet<T,sUseMemoryPool>::rotateLeft(Node* node)
{
  Node* child   = node->mRight;
  node->mRight  = child->mLeft;
  child->mLeft  = node;
  child->mIsRed = node->mIsRed;
  node->mIsRed  = true;
  return child;
}


/// Rotates a sub-tree one step to the right without changing the order.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree (must not be NULL).
/// @return
///   Pointer to the new root node of the rotated sub-tree.
template <class T, bool sUseMemoryPool>
inline typename RBTreeSet<T,sUseMemoryPool>::Node*
RBTreeSet<T,sUseMemoryPool>::rotateRight(Node* node)
{
  Node* child   = node->mLeft;
  node->mLeft   = child->mRight;
  child->mRight = node;
  child->mIsRed = node->mIsRed;
  node->mIsRed  = true;
  return child;
}


/// Flips the colors of a node and it's children. It's used to merge the
/// 2 red links of the children to one red link in the parent.
///
/// NOTE: This function can be called only, if the children of the node are
///       valid!
///
/// @param[in]  node
///   Pointer to the parent node (must nit be NULL).
template <class T, bool sUseMemoryPool>
inline void RBTreeSet<T,sUseMemoryPool>::colorFlip(Node* node)
{
  node->mIsRed         ^= 1;
  node->mLeft->mIsRed  ^= 1;
  node->mRight->mIsRed ^= 1;
}


/// Recursive function that inserts a node into a LLRB sub-tree. If a value
/// with the same rang already exists, it gets overwritten.
/// The properties of a LLRB tree will be preserved.
///
/// TODO: Make this function iterative using a stack.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree into which the new value should
///   be inserted (must not be NULL initially).
/// @param[in]  value
///   The new value to insert (gets copied).
/// @return
///   Pointer to the new root node of the sub-tree or NULL, if the allocation
///   of a new tree node failed.
template <class T, bool sUseMemoryPool>
typename RBTreeSet<T,sUseMemoryPool>::Node*
RBTreeSet<T,sUseMemoryPool>::insert(Node*         node,
                                    const ValueT& value)
{
  // if we have found a leave, add new node to it and return
  if (!node) {
    // allocate memory
    void* mem;
    if (sUseMemoryPool) {
      mem = mMemPool->MemAlloc(sizeof(Node), false);
    } else {
      mem = GeAllocNC(sizeof(Node));
    }
    // construct Node using placement new
    if (mem) {
      mNewNode = ::new(mem) Node(value);
      ++mSize;
    } else {
      mNewNode = 0;
    }
    return mNewNode;
  }

  // traverse down the left sub-tree, if value is smaller than current node
  if (value < node->mValue) {
    node->mLeft = insert(node->mLeft, value);
  // traverse down the right sub-tree, if value is larger than current node
  } else if (node->mValue < value) {
    node->mRight = insert(node->mRight, value);
  // if we have found an already existing node with the same order, overwrite it
  } else {
    mNewNode = node;
  }

  // fix right-leaning leafs on the way up
  if (Node::isRed(node->mRight)) {
    node = rotateLeft(node);
  }

  // fix two reds in a row on the way up
  if (Node::isRed(node->mLeft) && Node::isRed(node->mLeft->mLeft)) {
    node = rotateRight(node);
  }

  // split 4-nodes on the way up
  if (Node::isRed(node->mLeft) && Node::isRed(node->mRight)) {
    colorFlip(node);
  }
  return node;
}


/// Deletes recursively all nodes of a sub-tree.
///
/// @param[in]  node
///   Pointer to the root node of the sub-tree (can be NULL).
template <class T, bool sUseMemoryPool>
void RBTreeSet<T,sUseMemoryPool>::deleteRecursive(Node* node)
{
  if (!node) { return; }
  deleteRecursive(node->mLeft);
  deleteRecursive(node->mRight);
  node->Node::~Node();
  if (!sUseMemoryPool) { GeFree(node); }
}



#endif  // #ifndef _RBTREESET_IMPL_H_