#include <cstdarg>
#include <cstdio>

#include "filepath.h"
#include "utilities.h"




#ifdef __PC
    const CHAR  gPathDelimiter = '\\';
    const CHAR* gPathDelimiterStr = "\\";
#else
    const CHAR gPathDelimiter = '/';
    const CHAR* gPathDelimiterStr = "/";
#endif



/// Works like printf, but writes the string to the debug console.
void debugLog(const CHAR* format, ...)
{
    static const SizeT cBufferSize = 2048;
    CHAR charBuffer[cBufferSize];

    va_list args;
    va_start(args, format);
    vsprintf(charBuffer, format, args);
    C4DOS.Ge->GeDebugOut("%s", charBuffer);
    va_end(args);
}


/// Writes a string to the debug console.
void debugLog(const String& msg)
{
    static const SizeT cBufferSize = 2048;
    CHAR charBuffer[cBufferSize];

    msg.GetCString(charBuffer, cBufferSize);
    charBuffer[cBufferSize-1] = '\0';
    C4DOS.Ge->GeDebugOut("%s", charBuffer);
}


/// Obtains a description parameter of type LONG of a C4DAtom.
///
/// @param[in]  atom
///   The atom of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  preset
///   The value the function returs, if the parameter doesn't exist or is not
///   of type LONG.
/// @return
///   The value of the parameter or the preset value, if we couldn't fetch it.
LONG getParameterLong(C4DAtom& atom,
                      LONG     paramID,
                      LONG     preset)
{
    GeData parameter;
    if (atom.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        return parameter.GetLong();
    }
    return preset;
}


/// Obtains a description parameter of type Real of a C4DAtom.
///
/// @param[in]  atom
///   The atom of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  preset
///   The value the function returs, if the parameter doesn't exist or is not
///   of type Real.
/// @return
///   The value of the parameter or the preset value, if we couldn't fetch it.
Real getParameterReal(C4DAtom& atom,
                      LONG     paramID,
                      Real     preset)
{
    GeData parameter;
    if (atom.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        return parameter.GetReal();
    }
    return preset;
}


/// Obtains a description parameter of type Vector of a C4DAtom.
///
/// @param[in]  atom
///   The atom of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  preset
///   The value the function returs, if the parameter doesn't exist or is not
///   of type Vector.
/// @return
///   The value of the parameter or the preset value, if we couldn't fetch it.
Vector getParameterVector(C4DAtom&      atom,
                          LONG          paramID,
                          const Vector& preset)
{
    GeData parameter;
    if (atom.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        return parameter.GetVector();
    }
    return preset;
}


/// Obtains a description parameter of type String of a C4DAtom.
///
/// @param[in]  atom
///   The atom of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  preset
///   The value the function returs, if the parameter doesn't exist or is not
///   of type String.
/// @return
///   The value of the parameter or the preset value, if we couldn't fetch it.
String getParameterString(C4DAtom&      atom,
                          LONG          paramID,
                          const String& preset)
{
    GeData parameter;
    if (atom.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        return parameter.GetString();
    }
    return preset;
}


/// Obtains a description parameter of type Filename of a C4DAtom.
///
/// @param[in]  atom
///   The atom of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  preset
///   The value the function returs, if the parameter doesn't exist or is not
///   of type Filename.
/// @return
///   The value of the parameter or the preset value, if we couldn't fetch it.
Filename getParameterFilename(C4DAtom&        atom,
                              LONG            paramID,
                              const Filename& preset)
{
    GeData parameter;
    if (atom.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        return parameter.GetFilename();
    }
    return preset;
}


/// Obtains a description parameter of type BaseTime of a C4DAtom.
///
/// @param[in]  atom
///   The atom of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  preset
///   The value the function returs, if the parameter doesn't exist or is not
///   of type BaseTime.
/// @return
///   The value of the parameter or the preset value, if we couldn't fetch it.
BaseTime getParameterTime(C4DAtom&        atom,
                          LONG            paramID,
                          const BaseTime& preset)
{
    GeData parameter;
    if (atom.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        return parameter.GetTime();
    }
    return preset;
}


/// Obtains a description parameter of type LINK from an object.
///
/// @param[in]  node
///   The atom node of which we want to read out the parameter.
/// @param[in]  paramID
///   The ID of the description parameter.
/// @param[in]  instanceOf
///   The object type that is allowed in the link.
/// @return
///   The link parameter (BaseList2D*) or o if we couldn't fetch it.
BaseList2D* getParameterLink(GeListNode& node,
                             LONG        paramID,
                             LONG        instanceOf)
{
    GeData parameter;
    if (node.GetParameter(DescLevel(paramID), parameter, DESCFLAGS_GET_0))
    {
        BaseLink* link = parameter.GetBaseLink();
        if (link)
        {
            return link->GetLink(node.GetDocument(), instanceOf);
        }
    }
    return 0;
}


/// Traverses the generated hierarchy of a parametric object upwards
/// and looks for the first occurence of a tag of a specific type.
///
/// @param[in]  object
///   The object of which the generator hierarchy will we traversed.
/// @param[in]  tagId
///   The ID of the tag to look for.
/// @return
///   The pointer to the first found tag of the specified type or NULL
///   if we couldn't find it.
BaseTag* findTagForParamObject(BaseObject* object,
                               LONG        tagId)
{
    BaseTag* tag;
    for (; object; object=object->GetCacheParent())
    {
        tag = object->GetTag(tagId);
        if (tag) { return tag; }
    }
    return 0;
}


/// Replacement of SVector::Normalize() as it was added later to Vector and is
/// not available in R9.6.
///
/// @param[in,out]  v
///   Reference to the vector that will be normalised.
/// @return
///   A reference to the input vector, which has been normalised, so that you
///   can use this function in more elaborate expressions.
SVector& normalize(SVector& v)
{
    SReal vlen = Len(v);
    if (vlen != 0)
    {
        v *= 1 / vlen;
    }
    else
    {
        v.x = 1;
        v.y = 0;
        v.z = 0;
    }
    return v;
}


/// Replacement of LVector::Normalize() as it was added later to Vector and is
/// not available in R9.6.
///
/// @param[in,out]  v
///   Reference to the vector that will be normalised.
/// @return
///   A reference to the input vector, which has been normalised, so that you
///   can use this function in more elaborate expressions.
LVector& normalize(LVector& v)
{
    LReal vlen = Len(v);
    if (vlen != 0.0)
    {
        v *= 1.0 / vlen;
    }
    else
    {
        v.x = 1.0;
        v.y = 0.0;
        v.z = 0.0;
    }
    return v;
}


SVector calculateNormal(const SVector& a,
                        const SVector& b,
                        const SVector& c)
{
    SVector cross = (b-a) % (c-a);
    return normalize(cross);
}


LVector calculateNormal(const LVector& a,
                        const LVector& b,
                        const LVector& c)
{
    LVector cross = (b-a) % (c-a);
    return normalize(cross);
}


SVector calculateNormal(const SVector& a,
                        const SVector& b,
                        const SVector& c,
                        const SVector& d)
{
    SVector sum = (b-a) % (c-a) + (c-a) % (d-a);
    return normalize(sum);
}


LVector calculateNormal(const LVector& a,
                        const LVector& b,
                        const LVector& c,
                        const LVector& d)
{
    LVector sum = (b-a) % (c-a) + (c-a) % (d-a);
    return normalize(sum);
}


Vector slerp(Vector a,
             Vector b,
             Real   t)
{
    static const Real cEpsilon = 0.00001;

    // normalise parameters
    t = Clamp(0.0, 1.0, t);
    normalize(a);
    normalize(b);

    // calculate angle between the two vectors
    Real angle = acos(a * b);

    // trivial case: a and b are equal -> no interpolation necessary
    if (angle == 0.0) { return a; }

    Vector result;

    // if a is almost b: do a linear interpolation, as the error would be minimal
    if (angle < cEpsilon)
    {
        result = a + (b - a) * t;
    }
    // unsolvable case: a and b are exactly opposite -> no slerp possible
    else if (angle + cEpsilon >= pi)
    {
        return Vector(1.0, 0.0, 0.0);
    }
    else
    {
        result = (sin((1.0-t)*angle) / sin(angle)) * a +
                 (sin(      t*angle) / sin(angle)) * b;
    }

    // return normalize result
    return normalize(result);
}


void collectAllObjects(BaseDocument& doc,
                       LONG          type,
                       AtomArray&    objects)
{
    // start with an empty container
    objects.Flush();

    // loop over all objects of the scene and store all camera objects in container
    BaseObject* next;
    for (BaseObject* obj=doc.GetFirstObject(); obj;)
    {
        if (obj->IsInstanceOf(type))
        {
            objects.Append(obj);
        }
        next = obj->GetDown();
        if (!next)
        {
            next = obj->GetNext();
            while (!next)
            {
                obj = obj->GetUp();
                if (obj)
                {
                    next = obj->GetNext();
                }
                else
                {
                    break;
                }
            }
        }
        obj = next;
    }
}


LONG getObjectIx(BaseDocument&     doc,
                 const BaseObject* object,
                 LONG              type)
{
    // loop over all objects of the scene and count all objects of the same type
    // and try to find the 
    LONG index = 0;
    BaseObject* next;
    for (BaseObject* obj=doc.GetFirstObject(); obj;)
    {
        if (obj->IsInstanceOf(type))
        {
            if (obj == object) { return index; }
            ++index;
        }
        next = obj->GetDown();
        if (!next)
        {
            next = obj->GetNext();
            while (!next)
            {
                obj = obj->GetUp();
                if (obj)
                {
                    next = obj->GetNext();
                }
                else
                {
                    break;
                }
            }
        }
        obj = next;
    }
    return -1;
}


BaseObject* findObject(BaseDocument& doc,
                       ULONG         index,
                       LONG          objectType,
                       LONG          tagType)
{
    ULONG count = 0;
    BaseObject* next;
    for (BaseObject* obj=doc.GetFirstObject(); obj;)
    {
        if (obj->IsInstanceOf(objectType) &&
            (!tagType || obj->GetTag(tagType)))
        {
            if (count == index) { return obj; }
            ++count;
        }
        next = obj->GetDown();
        if (!next)
        {
            next = obj->GetNext();
            while (!next)
            {
                obj = obj->GetUp();
                if (obj)
                {
                    next = obj->GetNext();
                }
                else
                {
                    break;
                }
            }
        }
        obj = next;
    }
    return 0;
}


void mangleName(String& name)
{
    if (name.Content())
    {
        for (LONG pos=0; name.FindFirst(' ', &pos, pos);) { name[pos] = '_'; }
    }
    else
    {
        name = "???";
    }
}


void dumpContainer(
    const BaseContainer &container,
    const int           indentLevel)
{
    GeData *data;
    debugLog("%*s%p", indentLevel, "", &container);
    for (int i=0; (data = container.GetIndexData((LONG)i)) != 0; ++i)
    {
        const int  id = container.GetIndexId(i);
        const LONG dataType = data->GetType();
        switch (dataType)
        {
            case DA_NIL:
                debugLog("%*s%-3d:   id = %6d  type = NIL",
                         indentLevel, "", i, id);
                break;
            case DA_VOID:
                debugLog("%*s%-3d:   id = %6d  type = VOID     value = %p",
                         indentLevel, "", i, id,
                         data->GetVoid());
                break;
            case DA_LONG:
                debugLog("%*s%-3d:   id = %6d  type = LONG     value = %d",
                         indentLevel, "", i, id,
                         (int)data->GetLong());
                break;
            case DA_REAL:
                debugLog("%*s%-3d:   id = %6d  type = REAL     value = %f",
                         indentLevel, "", i, id,
                         data->GetReal());
            case DA_TIME:
                debugLog("%*s%-3d:   id = %6d  type = TIME     value = %f",
                         indentLevel, "", i, id,
                         data->GetTime().Get());
                break;
            case DA_VECTOR:
                debugLog("%*s%-3d:   id = %6d  type = VECTOR   value = %f,%f,%f",
                         indentLevel, "", i, id,
                         data->GetVector().x, data->GetVector().y, data->GetVector().z);
                break;
            case DA_MATRIX:
                debugLog("%*s%-3d:   id = %6d  type = MATRIX   v1 = %f,%f,%f  v2 = %f,%f,%f  v3 = %f,%f,%f  off = %f,%f,%f",
                         indentLevel, "", i, id,
                         data->GetMatrix().v1.x, data->GetMatrix().v1.y, data->GetMatrix().v1.z,
                         data->GetMatrix().v2.x, data->GetMatrix().v2.y, data->GetMatrix().v2.z,
                         data->GetMatrix().v3.x, data->GetMatrix().v3.y, data->GetMatrix().v3.z,
                         data->GetMatrix().off.x, data->GetMatrix().off.y, data->GetMatrix().off.z);
                break;
            case DA_LLONG:
                debugLog("%*s%-3d:   id = %6d  type = LLONG    value = %ld",
                         indentLevel, "", i, id,
                         (long int)data->GetLLong());
                break;
            case DA_BYTEARRAY:
                debugLog("%*s%-3d:   id = %6d  type = BYTEARRAY",
                         indentLevel, "", i, id);
                break;
            case DA_STRING:
                {
                    CHAR *dataStr = data->GetString().GetCStringCopy();
                    debugLog("%*s%-3d:   id = %6d  type = STRING   value = '%s'",
                             indentLevel, "", i, id,
                             dataStr);
                    bDelete(dataStr);
                }
                break;
            case DA_FILENAME:
                {
                    CHAR *dataStr = data->GetFilename().GetString().GetCStringCopy();
                    debugLog("%*s%-3d:   id = %6d  type = STRING   value = '%s'",
                             indentLevel, "", i, id,
                             dataStr);
                    bDelete(dataStr);
                }
                break;
            case DA_CONTAINER:
                debugLog("%*s%-3d:   id = %6d  type = CONTAINER   content =",
                         indentLevel, "", i, id);
                dumpContainer(*data->GetContainer(), indentLevel + 2);
                break;
            case DA_ALIASLINK:
                debugLog("%*s%-3d:   id = %6d  type = ALIASLINK",
                         indentLevel, "", i, id);
                break;
            case DA_MARKER:
                debugLog("%*s%-3d:   id = %6d  type = MARKER",
                         indentLevel, "", i, id);
                break;
            case DA_MISSINGPLUG:
                debugLog("%*s%-3d:   id = %6d  type = MISSINGPLUG",
                         indentLevel, "", i, id);
                break;
            default:
                if (dataType >= DA_CUSTOMDATATYPE)
                {
                    debugLog("%*s%-3d:   id = %6d  type = CUSTOM",
                             indentLevel, "", i, id);
                }
                else
                {
                    debugLog("%*s%-3d:   id = %6d  type = UNKNOWN",
                             indentLevel, "", i, id);
                }
                break;
        }
    }
}



#if defined __PC

ProcessHandle _runProgramInternal(const wchar_t* appPath,
                                  wchar_t*       cmdLine,
                                  const wchar_t* workingDir,
                                  bool           runInBackground);

ProcessHandle runProgram(const FilePath& appPath,
                         const String*   args,
                         SizeT           argCount,
                         const String*   workingDir,
                         bool            runInBackground)
{
    // make sure that the assumption is correct that wchar_t == UWORD
    GeAssert(sizeof(wchar_t) == sizeof(UWORD));

    wchar_t appPathCStr[1024];
    wchar_t cmdLineCStr[8192];
    wchar_t workDirCStr[1024];

    String cmdLine = appPath.getString();
    cmdLine.GetUcBlockNull((UWORD*)appPathCStr, ARRAY_LENGTH(appPathCStr));
    cmdLine = "\"" + cmdLine + "\"";
    for (SizeT argIx=0; argIx<argCount; ++argIx)
    {
        cmdLine += " " + args[argIx];
    }
    ERRLOG(cmdLine);
    debugLog(cmdLine);
    cmdLine.GetUcBlockNull((UWORD*)cmdLineCStr, ARRAY_LENGTH(cmdLineCStr));
    if (workingDir && workingDir->Content())
    {
        workingDir->GetUcBlockNull((UWORD*)workDirCStr, ARRAY_LENGTH(workDirCStr));
        return _runProgramInternal(appPathCStr,
                                   cmdLineCStr,
                                   workDirCStr,
                                   (runInBackground!=false));
    }
    else
    {
        return _runProgramInternal(appPathCStr,
                                   cmdLineCStr,
                                   NULL,
                                   (runInBackground!=false));
    }
}


#elif defined __MAC


ProcessSerialNumber _runProgramInternal(const char* appPath,
                                        char *const args[],
                                        bool        runInBackground);

ProcessHandle runProgram(const FilePath& appPath,
                         const String    *args,
                         const SizeT     argCount,
                         const String*   workingDir,
                         const bool      runInBackground)
{
    ProcessHandle handle = {0, 0};
    char**        argCStrs;
    char**        argCStr;
    LONG          len;
    String        appPathStr;

    // allocate argument array
    argCStrs = bNew char*[argCount+2];
    if (!argCStrs) { goto CLEANUP_RETURN; }
    argCStr = argCStrs;

    // convert application path to plain C string and store itas first entry in array
    appPathStr = appPath.getString();
    len = appPathStr.GetCStringLen(STRINGENCODING_UTF8) + 1;
    *argCStr = bNew char[len];
    if (!*argCStr) { goto CLEANUP_RETURN; }
    appPathStr.GetCString(*argCStr, len, STRINGENCODING_UTF8);
    (*argCStr)[len-1] = '\0';
    ++argCStr;

    // loop over all arguments, convert them to C strings and store them in array
    for (SizeT i=0; i<argCount; ++i)
    {
        const String& arg(args[i]);
        if (arg.Content())
        {
            LONG len = arg.GetCStringLen(STRINGENCODING_UTF8) + 1;
            *argCStr = bNew char[len];
            if (!*argCStr) { goto CLEANUP_RETURN; }
            arg.GetCString(*argCStr, len, STRINGENCODING_UTF8);
            (*argCStr)[len-1] = '\0';
            ++argCStr;
        }
    }

    // run program
    handle = _runProgramInternal(argCStrs[0],
                                 argCStrs,
                                 (runInBackground!=false));

CLEANUP_RETURN:
    if (argCStrs)
    {
        for (argCStr=argCStrs; *argCStr; ++argCStr)
        {
            bDelete(*argCStr);
        }
        bDelete(argCStrs);
    }
    return handle;
}


#endif


bool _isValidInternal(const ProcessHandle& process);

bool isValid(const ProcessHandle& process)
{
    return _isValidInternal(process);
}


bool _programIsRunningInternal(const ProcessHandle& process);

bool programIsRunning(const ProcessHandle& process)
{
    return _programIsRunningInternal(process);
}


void _stopProgramInternal(const ProcessHandle& process);

void stopProgram(const ProcessHandle& process)
{
    _stopProgramInternal(process);
}
