#ifndef _UTILITIES_H_
#define _UTILITIES_H_  1



#include <c4d.h>

#include "common.h"



/*****************************************************************************
 * Macros
 *****************************************************************************/

/// Logs/prints an error message.
#define ERRLOG(msg)                                                           \
  { GePrint(msg); }

/// Logs/prints an error message and returns without a return value.
#define ERRLOG_RETURN(msg)                                                    \
  { ERRLOG(msg);  return; }

/// Logs/prints an error message and returns with the specified return value.
#define ERRLOG_RETURN_VALUE(returnValue,msg)                                  \
  { ERRLOG(msg);  return returnValue; }

/// Logs/prints an error message, defined by an ID and returns with the
/// specified return value.
#define ERRLOG_ID_RETURN_VALUE(returnValue,id,msg)                            \
  { mErrorStringID=(id);  ERRLOG(msg);  return returnValue; }

/// Evaluates the length of an array (should be calculated during compile time).
#define ARRAY_LENGTH(a)                                                       \
  ( sizeof(a) / sizeof(*(a)) )


/*****************************************************************************
 * Predeclarations of classes
 *****************************************************************************/

class FilePath;



/*****************************************************************************
 * Common types
 *****************************************************************************/

#if defined __PC

typedef void* ProcessHandle; 
static const ProcessHandle gEmptyHandle = NULL;

#elif defined __MAC

typedef struct ProcessSerialNumber ProcessHandle;
static const ProcessHandle gEmptyHandle = {0, 0};

#endif



/*****************************************************************************
 * Common constants
 *****************************************************************************/

extern const CHAR  gPathDelimiter;
extern const CHAR* gPathDelimiterStr;



/*****************************************************************************
 * Functions
 *****************************************************************************/

void debugLog(const CHAR* format, ...);

void debugLog(const String& msg);


LONG getParameterLong(C4DAtom& object,
                      LONG     paramID,
                      LONG     preset = 0);

Real getParameterReal(C4DAtom& object,
                      LONG     paramID,
                      Real     preset = 0.0);

Vector getParameterVector(C4DAtom&      object,
                          LONG          paramID,
                          const Vector& preset = Vector());

String getParameterString(C4DAtom&      atom,
                          LONG          paramID,
                          const String& preset = String());

Filename getParameterFilename(C4DAtom&        atom,
                              LONG            paramID,
                              const Filename& preset = Filename());

BaseTime getParameterTime(C4DAtom&        atom,
                          LONG            paramID,
                          const BaseTime& preset = BaseTime());

BaseList2D* getParameterLink(GeListNode& node,
                             LONG        paramID,
                             LONG        instanceOf = 0);


BaseTag* findTagForParamObject(BaseObject* object,
                               LONG        tagId);


SVector& normalize(SVector& v);

LVector& normalize(LVector& v);


SVector calculateNormal(const SVector& a,
                        const SVector& b,
                        const SVector& c);

LVector calculateNormal(const LVector& a,
                        const LVector& b,
                        const LVector& c);

SVector calculateNormal(const SVector& a,
                        const SVector& b,
                        const SVector& c,
                        const SVector& d);

LVector calculateNormal(const LVector& a,
                        const LVector& b,
                        const LVector& c,
                        const LVector& d);


Vector slerp(Vector a,
             Vector b,
             Real   t);

void collectAllObjects(BaseDocument& doc,
                       LONG          type,
                       AtomArray&    objects);

LONG getObjectIx(BaseDocument&     doc,
                 const BaseObject* object,
                 LONG              type);

BaseObject* findObject(BaseDocument& doc,
                       ULONG         index,
                       LONG          objectType,
                       LONG          tagType = 0);


void mangleName(String& name);

void dumpContainer(
    const BaseContainer &container,
    const int           indentLevel = 0);


ProcessHandle runProgram(const FilePath& programFilename,
                         const String*   args,
                         const SizeT     argCount,
                         const String*   workingDir,
                         const bool      runInBackground);

bool isValid(const ProcessHandle& process);

bool programIsRunning(const ProcessHandle& process);

void stopProgram(const ProcessHandle& process);



/*****************************************************************************
 * Inlined functions
 *****************************************************************************/

/// Wrapper which returns the point positions of a PointObject. This is needed
/// as with R10 the function name changed.
///
/// @param[in]  object
///   Reference to the object of which the point array will be obtained.
/// @return
///   The array of point positions of the object.
inline const Vector* getPoints(PointObject& object)
{
  return object.GetPointR();
}


/// Wrapper which returns the polygons of a PointObject. This is needed
/// as with R10 the function name changed.
///
/// @param[in]  object
///   Reference to the object of which the polygon array will be obtained.
/// @return
///   The array of polygons of the object.
inline const CPolygon* getPolygons(PolygonObject& object)
{
  return object.GetPolygonR();
}


///
template <typename T>
inline T clamp(const T lower,
               const T upper,
               const T value)
{
    if (value < lower) { return lower; }
    if (value > upper) { return upper; }
    return value;
}



#endif  // #ifndef _UTILITIES_H_
